/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.database.data;

import java.io.Serializable;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Table;
import javax.persistence.UniqueConstraint;

import org.apache.commons.lang3.StringUtils;

@Entity( name = "ServerData" )
@Table( name = "dv_server", uniqueConstraints = { @UniqueConstraint( columnNames = { "name" } ) } )
public class ServerData
    implements Serializable
{

    private static final long serialVersionUID = -2506326764170084191L;

    public static final int DEFAULT_MAX_USERS = 10;

    public static final int DEFAULT_SSH_PORT = 22;

    public static final int DEFAULT_XPRA_MAX_PORT = 65535;

    public static final int DEFAULT_XPRA_MIN_PORT = 30000;

    public static final long NEW_SERVER = 0;

    public static final long UNDEFINED_SERVER = -1;

    public static final Boolean DEFAULT_ENABLED = Boolean.TRUE;

    public static final String NEW_DATA = StringUtils.EMPTY;

    @Id
    @GeneratedValue( strategy = GenerationType.AUTO )
    private Long id;

    @Column( name = "address", length = 100, nullable = false )
    private String address;

    @Column( name = "name", length = 100, nullable = false )
    private String name;

    @Column( name = "note", length = 200, nullable = true )
    private String note;

    @Column( name = "ssh_port", nullable = false )
    private Integer sshPort;

    @Column( name = "xpra_min_port", nullable = false )
    private Integer xpraMinPort;

    @Column( name = "xpra_max_port", nullable = false )
    private Integer xpraMaxPort;

    @Column( name = "xpra_version", length = 20, nullable = true )
    private String xpraVersion;

    @Column( name = "max_users", nullable = false )
    private Integer maxUsers;

    @Column( name = "enabled", nullable = false )
    private Boolean enabled;

    public ServerData()
    {
        setEnabled( DEFAULT_ENABLED );
        setMaxUsers( DEFAULT_MAX_USERS );
        setSshPort( DEFAULT_SSH_PORT );
        setXpraMaxPort( DEFAULT_XPRA_MAX_PORT );
        setXpraMinPort( DEFAULT_XPRA_MIN_PORT );
    }

    public ServerData( long id, String name, String note, String address, int sshPort, int xpraMinPort, int xpraMaxPort, String xpraVersion, int maxUsers, boolean enabled )
    {
        setId( id );
        setName( name );
        setNote( note );
        setAddress( address );
        setSshPort( sshPort );
        setXpraMinPort( xpraMinPort );
        setXpraMaxPort( xpraMaxPort );
        setXpraVersion( xpraVersion );
        setMaxUsers( maxUsers );
        setEnabled( enabled );
    }

    public Long getId()
    {
        return id;
    }

    public void setId( Long id )
    {
        this.id = id;
    }

    public String getName()
    {
        return name;
    }

    public void setName( String name )
    {
        this.name = StringUtils.stripToEmpty( name );
    }

    public String getNote()
    {
        return note;
    }

    public void setNote( String note )
    {
        this.note = StringUtils.stripToEmpty( note );
    }

    public String getAddress()
    {
        return address;
    }

    public void setAddress( String address )
    {
        this.address = StringUtils.stripToEmpty( address );
    }

    public Integer getSshPort()
    {
        return sshPort;
    }

    public void setSshPort( Integer sshPort )
    {
        this.sshPort = sshPort;
    }

    public Integer getXpraMinPort()
    {
        return xpraMinPort;
    }

    public void setXpraMinPort( Integer xpraMinPort )
    {
        this.xpraMinPort = xpraMinPort;
    }

    public Integer getXpraMaxPort()
    {
        return xpraMaxPort;
    }

    public void setXpraMaxPort( Integer xpraMaxPort )
    {
        this.xpraMaxPort = xpraMaxPort;
    }

    public String getXpraVersion()
    {
        return xpraVersion;
    }

    public void setXpraVersion( String xpraVersion )
    {
        this.xpraVersion = StringUtils.stripToEmpty( xpraVersion );
    }

    public Integer getMaxUsers()
    {
        return maxUsers;
    }

    public void setMaxUsers( Integer maxUsers )
    {
        this.maxUsers = maxUsers;
    }

    public Boolean isEnabled()
    {
        return enabled;
    }

    public void setEnabled( Boolean enabled )
    {
        this.enabled = enabled;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ( ( address == null ) ? 0 : address.hashCode() );
        result = prime * result + ( ( enabled == null ) ? 0 : enabled.hashCode() );
        result = prime * result + ( ( id == null ) ? 0 : id.hashCode() );
        result = prime * result + ( ( maxUsers == null ) ? 0 : maxUsers.hashCode() );
        result = prime * result + ( ( name == null ) ? 0 : name.hashCode() );
        result = prime * result + ( ( note == null ) ? 0 : note.hashCode() );
        result = prime * result + ( ( sshPort == null ) ? 0 : sshPort.hashCode() );
        result = prime * result + ( ( xpraMaxPort == null ) ? 0 : xpraMaxPort.hashCode() );
        result = prime * result + ( ( xpraMinPort == null ) ? 0 : xpraMinPort.hashCode() );
        result = prime * result + ( ( xpraVersion == null ) ? 0 : xpraVersion.hashCode() );
        return result;
    }

    @Override
    public boolean equals( Object obj )
    {
        if ( this == obj )
        {
            return true;
        }
        if ( obj == null )
        {
            return false;
        }
        if ( !( obj instanceof ServerData ) )
        {
            return false;
        }
        ServerData other = (ServerData) obj;
        if ( address == null )
        {
            if ( other.address != null )
            {
                return false;
            }
        }
        else if ( !address.equals( other.address ) )
        {
            return false;
        }
        if ( enabled == null )
        {
            if ( other.enabled != null )
            {
                return false;
            }
        }
        else if ( !enabled.equals( other.enabled ) )
        {
            return false;
        }
        if ( id == null )
        {
            if ( other.id != null )
            {
                return false;
            }
        }
        else if ( !id.equals( other.id ) )
        {
            return false;
        }
        if ( maxUsers == null )
        {
            if ( other.maxUsers != null )
            {
                return false;
            }
        }
        else if ( !maxUsers.equals( other.maxUsers ) )
        {
            return false;
        }
        if ( name == null )
        {
            if ( other.name != null )
            {
                return false;
            }
        }
        else if ( !name.equals( other.name ) )
        {
            return false;
        }
        if ( note == null )
        {
            if ( other.note != null )
            {
                return false;
            }
        }
        else if ( !note.equals( other.note ) )
        {
            return false;
        }
        if ( sshPort == null )
        {
            if ( other.sshPort != null )
            {
                return false;
            }
        }
        else if ( !sshPort.equals( other.sshPort ) )
        {
            return false;
        }
        if ( xpraMaxPort == null )
        {
            if ( other.xpraMaxPort != null )
            {
                return false;
            }
        }
        else if ( !xpraMaxPort.equals( other.xpraMaxPort ) )
        {
            return false;
        }
        if ( xpraMinPort == null )
        {
            if ( other.xpraMinPort != null )
            {
                return false;
            }
        }
        else if ( !xpraMinPort.equals( other.xpraMinPort ) )
        {
            return false;
        }
        if ( xpraVersion == null )
        {
            if ( other.xpraVersion != null )
            {
                return false;
            }
        }
        else if ( !xpraVersion.equals( other.xpraVersion ) )
        {
            return false;
        }
        return true;
    }

}
