/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.database.filters;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.commons.collections4.MapUtils;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData_;

public class ApplicationFilter
    implements Serializable
{

    private static final long serialVersionUID = -7963334568010547996L;

    private static final String FILTER = "%%%s%%";

    public static final <T> CriteriaQuery<T> addFilter( CriteriaBuilder criteriaBuilder, CriteriaQuery<T> criteriaQuery, Root<ApplicationData> application, Map<ApplicationColumns, String> filterMap )
    {
        if ( MapUtils.isNotEmpty( filterMap ) )
        {
            List<Predicate> predicates = new ArrayList<>();
            for ( Map.Entry<ApplicationColumns, String> filter : filterMap.entrySet() )
            {
                switch ( filter.getKey() )
                {
                    case COMMAND:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( application.get( ApplicationData_.command ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                                 .toLowerCase() ) ) );
                        break;
                    case COMMAND_OPTIONS:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( application.get( ApplicationData_.commandOptions ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                                        .toLowerCase() ) ) );
                        break;
                    case ENABLED:
                        Boolean enabled = Boolean.parseBoolean( filter.getValue() );
                        if ( enabled )
                        {
                            predicates.add( criteriaBuilder.isTrue( application.get( ApplicationData_.enabled ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( application.get( ApplicationData_.enabled ) ) );
                        }
                        break;
                    case NAME:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( application.get( ApplicationData_.name ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                              .toLowerCase() ) ) );
                        break;
                    case NOTE:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( application.get( ApplicationData_.note ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                              .toLowerCase() ) ) );
                        break;
                    case ONE_INSTANCE:
                        Boolean oneInstance = Boolean.parseBoolean( filter.getValue() );
                        if ( oneInstance )
                        {
                            predicates.add( criteriaBuilder.lt( application.get( ApplicationData_.oneInstance ), ApplicationData.ONE_INSTANCE_DISABLED ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.equal( application.get( ApplicationData_.oneInstance ), ApplicationData.ONE_INSTANCE_DISABLED ) );
                        }
                        break;
                    case SERVER_AUTO_SELECT:
                        Boolean serverAutoSelect = Boolean.parseBoolean( filter.getValue() );
                        if ( serverAutoSelect )
                        {
                            predicates.add( criteriaBuilder.isTrue( application.get( ApplicationData_.serverAutoSelect ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( application.get( ApplicationData_.serverAutoSelect ) ) );
                        }
                        break;
                    case SERVERS_GROUP:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( application.join( ApplicationData_.serversGroupDatas )
                                                                                                .get( ServersGroupData_.name ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    default:
                        // nothing to do
                        break;
                }
            }
            return criteriaQuery.where( criteriaBuilder.and( predicates.toArray( Predicate[]::new ) ) );
        }
        return criteriaQuery;
    }

}
