/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.database.filters;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.commons.collections4.MapUtils;
import org.dussan.vaadin.dvirtual.database.columns.UserColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData_;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserData_;

public class UserFilter
    implements Serializable
{

    private static final long serialVersionUID = 2927987717129223023L;

    private static final String FILTER = "%%%s%%";

    public static final <T> CriteriaQuery<T> addFilter( CriteriaBuilder criteriaBuilder, CriteriaQuery<T> criteriaQuery, Root<UserData> user, Map<UserColumns, String> filterMap )
    {
        if ( MapUtils.isNotEmpty( filterMap ) )
        {
            List<Predicate> predicates = new ArrayList<>();
            for ( Map.Entry<UserColumns, String> filter : filterMap.entrySet() )
            {
                switch ( filter.getKey() )
                {
                    case ADMINISTRATOR:
                        Boolean administrator = Boolean.parseBoolean( filter.getValue() );
                        if ( administrator )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.administrator ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.administrator ) ) );
                        }
                        break;
                    case APPLICATION:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( user.join( UserData_.applicationDatas )
                                                                                         .get( ApplicationData_.name ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    case APPLICATION_MAX:
                        predicates.add( criteriaBuilder.like( user.get( UserData_.maxApplications )
                                                                  .as( String.class ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    case APPLICATIONS_GROUP:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( user.join( UserData_.applicationsGroupDatas )
                                                                                         .get( ApplicationsGroupData_.name ) ),
                                                              String.format( FILTER, filter.getValue()
                                                                                           .toLowerCase() ) ) );
                        break;
                    case DOWNLOAD:
                        Boolean download = Boolean.parseBoolean( filter.getValue() );
                        if ( download )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.download ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.download ) ) );
                        }
                        break;
                    case ENABLED:
                        Boolean enabled = Boolean.parseBoolean( filter.getValue() );
                        if ( enabled )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.enabled ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.enabled ) ) );
                        }
                        break;
                    case FULLNAME:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( user.get( UserData_.fullName ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                    .toLowerCase() ) ) );
                        break;
                    case FULL_SCREEN:
                        Boolean fullSceen = Boolean.parseBoolean( filter.getValue() );
                        if ( fullSceen )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.fullScreen ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.fullScreen ) ) );
                        }
                        break;
                    case SCREENSHOT:
                        Boolean sceenshot = Boolean.parseBoolean( filter.getValue() );
                        if ( sceenshot )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.screenshot ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.screenshot ) ) );
                        }
                        break;
                    case UPLOAD:
                        Boolean upload = Boolean.parseBoolean( filter.getValue() );
                        if ( upload )
                        {
                            predicates.add( criteriaBuilder.isTrue( user.get( UserData_.upload ) ) );
                        }
                        else
                        {
                            predicates.add( criteriaBuilder.isFalse( user.get( UserData_.upload ) ) );
                        }
                        break;
                    case USERNAME:
                        predicates.add( criteriaBuilder.like( criteriaBuilder.lower( user.get( UserData_.userName ) ), String.format( FILTER, filter.getValue()
                                                                                                                                                    .toLowerCase() ) ) );
                        break;
                    default:
                        // nothing to do
                        break;
                }
            }
            return criteriaQuery.where( criteriaBuilder.and( predicates.toArray( Predicate[]::new ) ) );
        }
        return criteriaQuery;
    }

}
