/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.info;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;

public class KeyboardCharcodeName
    implements Serializable
{

    private static final long serialVersionUID = 34651375923406109L;

    private Map<Integer, String> charcodeName = new HashMap<>();

    public KeyboardCharcodeName()
    {
        charcodeName.put( 8, "BackSpace" );
        charcodeName.put( 9, "Tab" );
        charcodeName.put( 12, "KP_Begin" );
        charcodeName.put( 13, "Return" );
        charcodeName.put( 16, "Shift_L" );
        charcodeName.put( 17, "Control_L" );
        charcodeName.put( 18, "Alt_L" );
        charcodeName.put( 19, "Pause" ); // pause/break
        charcodeName.put( 20, "Caps_Lock" );
        charcodeName.put( 27, "Escape" );
        charcodeName.put( 31, "Mode_switch" );
        charcodeName.put( 32, "space" );
        charcodeName.put( 33, "Prior" ); // Page Up
        charcodeName.put( 34, "Next" ); // Page Down
        charcodeName.put( 35, "End" );
        charcodeName.put( 36, "Home" );
        charcodeName.put( 37, "Left" );
        charcodeName.put( 38, "Up" );
        charcodeName.put( 39, "Right" );
        charcodeName.put( 40, "Down" );
        charcodeName.put( 42, "Print" );
        charcodeName.put( 45, "Insert" );
        charcodeName.put( 46, "Delete" );
        charcodeName.put( 58, "colon" );
        charcodeName.put( 59, "semicolon" );
        charcodeName.put( 60, "less" );
        charcodeName.put( 61, "equal" );
        charcodeName.put( 62, "greater" );
        charcodeName.put( 63, "question" );
        charcodeName.put( 64, "at" );
        charcodeName.put( 91, "Menu" ); // Left Window Key
        charcodeName.put( 92, "Menu" ); // Right Window Key
        charcodeName.put( 93, "KP_Enter" ); // "select key"?
        charcodeName.put( 106, "KP_Multiply" );
        charcodeName.put( 107, "KP_Add" );
        charcodeName.put( 109, "KP_Subtract" );
        charcodeName.put( 110, "KP_Delete" );
        charcodeName.put( 111, "KP_Divide" );
        charcodeName.put( 144, "Num_Lock" );
        charcodeName.put( 145, "Scroll_Lock" );
        charcodeName.put( 160, "dead_circumflex" );
        charcodeName.put( 161, "exclam" );
        charcodeName.put( 162, "quotedbl" );
        charcodeName.put( 163, "numbersign" );
        charcodeName.put( 164, "dollar" );
        charcodeName.put( 165, "percent" );
        charcodeName.put( 166, "ampersand" );
        charcodeName.put( 167, "underscore" );
        charcodeName.put( 168, "parenleft" );
        charcodeName.put( 169, "parenright" );
        charcodeName.put( 170, "asterisk" );
        charcodeName.put( 171, "plus" );
        charcodeName.put( 172, "bar" );
        charcodeName.put( 173, "minus" );
        charcodeName.put( 174, "braceleft" );
        charcodeName.put( 175, "braceright" );
        charcodeName.put( 176, "asciitilde" );
        charcodeName.put( 186, "semicolon" );
        charcodeName.put( 187, "equal" );
        charcodeName.put( 188, "comma" );
        charcodeName.put( 189, "minus" );
        charcodeName.put( 190, "period" );
        charcodeName.put( 191, "slash" );
        charcodeName.put( 192, "grave" );
        charcodeName.put( 219, "bracketleft" );
        charcodeName.put( 220, "backslash" );
        charcodeName.put( 221, "bracketright" );
        charcodeName.put( 222, "apostrophe" );

        for ( int index = 0; index < 26; index++ )
        {
            charcodeName.put( 65 + index, StringUtils.mid( "abcdefghijklmnopqrstuvwxyz", index, 1 ) );
        }

        for ( int index = 0; index < 10; index++ )
        {
            charcodeName.put( 48 + index, Integer.toString( index ) );
            charcodeName.put( 96 + index, Integer.toString( index ) );
        }

        for ( int index = 1; index <= 24; index++ )
        {
            charcodeName.put( 111 + index, StringUtils.join( "F", index ) );
        }
    }

    public boolean contains( int keyCode )
    {
        return Objects.nonNull( MapUtils.getString( charcodeName, keyCode ) );
    }

    public String get( int keyCode )
    {
        return MapUtils.getString( charcodeName, keyCode );
    }

}
