/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.client.Command.XPRA;
import org.dussan.vaadin.dvirtual.client.VirtualClient;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;

public class ChildApplicationRunner
    implements Runnable
{

    private static final double SLEEP_INTERVAL_FACTOR = 10d; // milliseconds

    private static final int BYTES_LENGTH = 8192;

    private static final int NO_STREAM_VALUE = -1;

    private static final long MAX_WAIT_INTERVAL = 30000; // milliseconds

    private static final long MAX_SLEEP_INTERVAL = VirtualClientRunner.MAX_WAIT_CHILD_WINDOW_CREATED_INTERVAL - 1000; // milliseconds

    private static final long SLEEP_INTERVAL = 50; // milliseconds

    private static final long START_RECREATE_WINDOW_INTERVAL = VirtualClientRunner.MAX_WAIT_CHILD_WINDOW_CREATED_INTERVAL + 500; // milliseconds

    private static final long START_SLEEP_INTERVAL = 2000; // milliseconds

    private static final String STRING_MINUS = "-";

    private static final String TEMP_DIR = System.getProperty( "java.io.tmpdir" );

    private static final long WAIT_INTERVAL = 500; // milliseconds

    boolean childApplicationInstalled = false;

    boolean childApplicationStarted = false;

    private int sshPort = UserSessionData.NEW_PORT;

    private int xpraPort = UserSessionData.NEW_PORT;

    private long timer = 0;

    private String userName = null;

    private String command = null;

    private String xpraAddress = null;

    private VirtualClient virtualClient = null;

    private VirtualUtil virtualUtil = null;

    public ChildApplicationRunner( UserSessionData userSessionData, String command, VirtualClient virtualClient, VirtualUtil virtualUtil )
    {
        this( userSessionData.getUserData()
                             .getUserName(),
              command, userSessionData.getServerData()
                                      .getAddress(),
              userSessionData.getPort(), userSessionData.getServerData()
                                                        .getSshPort(),
              virtualClient, virtualUtil );
    }

    public ChildApplicationRunner( String userName, String command, String xpraAddress, int xpraPort, int sshPort, VirtualClient virtualClient, VirtualUtil virtualUtil )
    {
        this.userName = userName;
        this.command = command;
        this.xpraAddress = xpraAddress;
        this.xpraPort = xpraPort;
        this.sshPort = sshPort;
        this.virtualClient = virtualClient;
        this.virtualUtil = virtualUtil;
    }

    public void applicationStarted()
    {
        // nothing to do
    }

    public void connectionProgress( double progress )
    {
        // nothing to do
    }

    public void startError()
    {
        // nothing to do
    }

    @Override
    public void run()
    {
        // start application on the XPRA server
        Channel channel = null;
        Session session = null;
        try
        {
            // start SSH session
            session = SshUtil.createSession( userName, EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ), xpraAddress, sshPort );
            session.connect();

            // find application
            if ( timer == 0 )
            {
                InputStream inputStream = null;
                byte[] tempInputStreamValue = null;
                for ( ; timer <= MAX_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    boolean applicationWithPath = StringUtils.contains( command, File.separator );
                    tempInputStreamValue = new byte[BYTES_LENGTH];
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION, command ) );
                    inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }

                    int length = NO_STREAM_VALUE;
                    String inputStreamValue = null;
                    while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                    {
                        if ( length > 0 )
                        {
                            inputStreamValue = new String( tempInputStreamValue, 0, length );
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();

                    // progress
                    if ( timer >= START_SLEEP_INTERVAL && ( timer % ( SLEEP_INTERVAL_FACTOR * SLEEP_INTERVAL ) ) == 0 )
                    {
                        connectionProgress( SLEEP_INTERVAL_FACTOR * timer / MAX_SLEEP_INTERVAL );
                    }

                    if ( StringUtils.isNotBlank( inputStreamValue ) )
                    {
                        childApplicationInstalled = true;
                        break;
                    }
                }
            }

            // start child application
            if ( childApplicationInstalled )
            {
                ApplicationLocale applicationLocale = virtualUtil.getApplicationLocale();
                String tmpDir = String.join( StringUtils.EMPTY, Path.of( TEMP_DIR, XPRA.XPRA_APPLICATION )
                                                                    .toString(),
                                             STRING_MINUS, virtualUtil.getUserName() );
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.START_CHILD, xpraPort, //virtualUtil.getBandwidthLimit(),
                                                                     applicationLocale.getLocale(), tmpDir, command ) );
                timer = virtualUtil.isRecreateNewWindowEnabled() ? START_RECREATE_WINDOW_INTERVAL : timer;
                channel.connect();
                for ( ; timer <= MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );

                    // progress
                    if ( timer >= START_SLEEP_INTERVAL && ( timer % ( SLEEP_INTERVAL_FACTOR * SLEEP_INTERVAL ) ) == 0 )
                    {
                        connectionProgress( SLEEP_INTERVAL_FACTOR * timer / MAX_WAIT_INTERVAL );
                    }

                    if ( channel.isClosed() && ( virtualUtil.isDrawingStarted() || ( Objects.nonNull( virtualClient ) && virtualClient.isDrawingStartedInAllDialogOrNormalWindows() ) ) )
                    {
                        virtualUtil.setChildApplicationCanStarted( false );
                        childApplicationStarted = true;
                        break;
                    }
                }
                channel.disconnect();

                if ( childApplicationStarted )
                {
                    applicationStarted();
                }
                else
                {
                    TimeUnit.MILLISECONDS.sleep( WAIT_INTERVAL );
                    startError();
                }
            }
            else
            {
                timer = MAX_WAIT_INTERVAL;
                TimeUnit.MILLISECONDS.sleep( WAIT_INTERVAL );
                startError();
            }
            session.disconnect();
        }
        catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }
    }

}
