/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Objects;
import java.util.UUID;
import java.util.concurrent.ScheduledFuture;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.slf4j.Logger;

import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;
import com.vaadin.flow.server.StreamResource;

public abstract class FileDownloadRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String FILE_SEPARATOR = File.separator;

    private static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

    private static final String TEMPORARY_FILE_EXTENSION = ".dv_";

    private static final String TEMPORARY_FOLDER = System.getProperty( "java.io.tmpdir" );

    private String downloadedFileName = null;

    private ScheduledFuture<?> scheduledFuture = null;

    private UserSessionData userSessionData = null;

    private VirtualUtil virtualUtil = null;

    public FileDownloadRunner( VirtualUtil virtualUtil, UserSessionData userSessionData )
    {
        this.virtualUtil = virtualUtil;
        this.userSessionData = userSessionData;
    }

    public abstract void downloadingFile( StreamResource fileStreamResource, String fileName );

    public abstract void startDownloadingFile();

    public abstract void stopDownloadingFile();

    private StreamResource getStreamResource( String fileName, File file )
    {
        return new StreamResource( fileName, () -> {
            try
            {
                return new BufferedInputStream( new FileInputStream( file ) );
            }
            catch ( IOException exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
            return null;
        } );
    }

    @Override
    public void run()
    {
        ChannelSftp sftpChannel = null;
        Session session = null;

        if ( Objects.nonNull( virtualUtil ) && StringUtils.isNotBlank( downloadedFileName ) )
        {
            // start downloading file
            startDownloadingFile();

            // stop the scheduler
            scheduledFuture.cancel( false );
            scheduledFuture = null;

            // send the file from the server to the client
            try
            {
                downloadedFileName = StringUtils.remove( downloadedFileName, LINE_SEPARATOR );
                String tmpDownloadedFileName = StringUtils.join( TEMPORARY_FOLDER, FILE_SEPARATOR, UUID.randomUUID(), TEMPORARY_FILE_EXTENSION );
                String[] filepath = StringUtils.split( downloadedFileName, FILE_SEPARATOR );
                String fileName = filepath[filepath.length - 1];

                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 userSessionData.getServerData()
                                                                .getAddress(),
                                                 userSessionData.getServerData()
                                                                .getSshPort() );
                session.connect();

                sftpChannel = (ChannelSftp) session.openChannel( Command.SSH.SFTP );
                sftpChannel.connect();
                sftpChannel.get( downloadedFileName, tmpDownloadedFileName );

                File file = new File( tmpDownloadedFileName );
                file.deleteOnExit();
                StreamResource fileStreamResource = getStreamResource( file.getName(), file );
                downloadingFile( fileStreamResource, fileName );

                sftpChannel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | JSchException | SftpException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( sftpChannel ) && sftpChannel.isConnected() )
                {
                    sftpChannel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }

        // stop downloading file
        stopDownloadingFile();
    }

    public void setScheduledFuture( ScheduledFuture<?> scheduledFuture )
    {
        this.scheduledFuture = scheduledFuture;
    }

    public String getDownloadedFileName()
    {
        return downloadedFileName;
    }

    public void setDownloadedFileName( String downloadedFileName )
    {
        this.downloadedFileName = downloadedFileName;
    }

}
