/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ScheduledFuture;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.PredicateUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.config.ApplicationConfig;
import org.dussan.vaadin.dvirtual.config.ApplicationsGroupConfig;
import org.dussan.vaadin.dvirtual.config.ServerConfig;
import org.dussan.vaadin.dvirtual.config.ServersGroupConfig;
import org.dussan.vaadin.dvirtual.config.UserConfig;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;

public abstract class ImportDataRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private int importedDataCounter = 0;

    private String fileName = null;

    private String fileType = null;

    private String folder = null;

    private Database database = null;

    private ScheduledFuture<?> scheduledFuture = null;

    private VirtualUtil virtualUtil = null;

    public ImportDataRunner( Database database, VirtualUtil virtualUtil, String folder, String fileName, String fileType )
    {
        this.database = database;
        this.virtualUtil = virtualUtil;
        this.folder = folder;
        this.fileName = fileName;
        this.fileType = fileType;
    }

    @Override
    public void run()
    {
        // stop the scheduler
        scheduledFuture.cancel( false );
        scheduledFuture = null;

        String dataImportedMessage = null;
        if ( Objects.nonNull( database ) && StringUtils.isNotBlank( folder ) && StringUtils.isNotBlank( fileName ) )
        {
            try
            {
                if ( StringUtils.equals( fileType, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS ) ) )
                {
                    List<ApplicationData> datas = new ApplicationConfig( folder, fileName ).getApplicationDatas();
                    if ( CollectionUtils.isNotEmpty( datas ) )
                    {
                        datas.forEach( data -> {
                            Long[] ids = data.getServersGroupDatas()
                                             .stream()
                                             .map( ServersGroupData::getId )
                                             .filter( Objects::nonNull )
                                             .toArray( Long[]::new );
                            String[] names = data.getServersGroupDatas()
                                                 .stream()
                                                 .map( ServersGroupData::getName )
                                                 .filter( Objects::nonNull )
                                                 .toArray( String[]::new );
                            Set<ServersGroupData> d = new HashSet<>();
                            if ( ArrayUtils.isNotEmpty( ids ) )
                            {
                                d.addAll( database.getServerGroups( ids ) );
                            }
                            if ( ArrayUtils.isNotEmpty( names ) )
                            {
                                d.addAll( database.getServerGroups( names ) );
                            }
                            CollectionUtils.filter( d, PredicateUtils.notNullPredicate() );
                            data.setServersGroupDatas( new ArrayList<>( d ) );
                            if ( database.addApplication( data ) != ApplicationData.UNDEFINED_APPLICATION )
                            {
                                importedDataCounter += 1;
                            }
                        } );
                        dataImportedMessage = String.format( virtualUtil.getTranslation( Translation.DATA_IMPORT_OK, importedDataCounter, datas.size() ) );
                    }
                    else
                    {
                        dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                    }
                }
                else if ( StringUtils.equals( fileType, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS ) ) )
                {
                    List<ApplicationsGroupData> datas = new ApplicationsGroupConfig( folder, fileName ).getApplicationsGroupDatas();
                    if ( CollectionUtils.isNotEmpty( datas ) )
                    {
                        datas.forEach( data -> {
                            Long[] ids = data.getApplicationDatas()
                                             .stream()
                                             .map( ApplicationData::getId )
                                             .filter( Objects::nonNull )
                                             .toArray( Long[]::new );
                            String[] names = data.getApplicationDatas()
                                                 .stream()
                                                 .map( ApplicationData::getName )
                                                 .filter( Objects::nonNull )
                                                 .toArray( String[]::new );
                            Set<ApplicationData> d = new HashSet<>();
                            if ( ArrayUtils.isNotEmpty( ids ) )
                            {
                                d.addAll( database.getApplications( ids ) );
                            }
                            if ( ArrayUtils.isNotEmpty( names ) )
                            {
                                d.addAll( database.getApplications( names ) );
                            }
                            CollectionUtils.filter( d, PredicateUtils.notNullPredicate() );
                            data.setApplicationDatas( new ArrayList<>( d ) );
                            if ( database.addApplicationsGroup( data ) != ApplicationsGroupData.UNDEFINED_APPLICATIONS_GROUP )
                            {
                                importedDataCounter += 1;
                            }
                        } );
                        dataImportedMessage = String.format( virtualUtil.getTranslation( Translation.DATA_IMPORT_OK, importedDataCounter, datas.size() ) );
                    }
                    else
                    {
                        dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                    }
                }
                else if ( StringUtils.equals( fileType, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS ) ) )
                {
                    List<ServersGroupData> datas = new ServersGroupConfig( folder, fileName ).getServersGroupDatas();
                    if ( CollectionUtils.isNotEmpty( datas ) )
                    {
                        datas.forEach( data -> {
                            Long[] ids = data.getServerDatas()
                                             .stream()
                                             .map( ServerData::getId )
                                             .filter( Objects::nonNull )
                                             .toArray( Long[]::new );
                            String[] names = data.getServerDatas()
                                                 .stream()
                                                 .map( ServerData::getName )
                                                 .filter( Objects::nonNull )
                                                 .toArray( String[]::new );
                            Set<ServerData> d = new HashSet<>();
                            if ( ArrayUtils.isNotEmpty( ids ) )
                            {
                                d.addAll( database.getServers( ids ) );
                            }
                            if ( ArrayUtils.isNotEmpty( names ) )
                            {
                                d.addAll( database.getServers( names ) );
                            }
                            CollectionUtils.filter( d, PredicateUtils.notNullPredicate() );
                            data.setServerDatas( new ArrayList<>( d ) );
                            if ( database.addServersGroup( data ) != ServersGroupData.UNDEFINED_SERVERS_GROUP )
                            {
                                importedDataCounter += 1;
                            }
                        } );
                        dataImportedMessage = String.format( virtualUtil.getTranslation( Translation.DATA_IMPORT_OK, importedDataCounter, datas.size() ) );
                    }
                    else
                    {
                        dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                    }
                }
                else if ( StringUtils.equals( fileType, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ) ) )
                {
                    List<ServerData> datas = new ServerConfig( folder, fileName ).getServerDatas();
                    if ( CollectionUtils.isNotEmpty( datas ) )
                    {
                        datas.forEach( data -> {
                            if ( database.addServer( data ) != ServerData.UNDEFINED_SERVER )
                            {
                                importedDataCounter += 1;
                            }
                        } );
                        dataImportedMessage = String.format( virtualUtil.getTranslation( Translation.DATA_IMPORT_OK, importedDataCounter, datas.size() ) );
                    }
                    else
                    {
                        dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                    }
                }
                else if ( StringUtils.equals( fileType, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS ) ) )
                {
                    UserConfig userConfig = new UserConfig( folder, fileName );
                    List<UserData> datas = new UserConfig( folder, fileName ).getUserDatas();
                    if ( CollectionUtils.isNotEmpty( datas ) )
                    {
                        datas.forEach( data -> {
                            Long[] applicationIds = userConfig.getUserApplicationDatas()
                                                              .stream()
                                                              .filter( d -> StringUtils.equals( d.getUserData()
                                                                                                 .getUserName(),
                                                                                                data.getUserName() ) )
                                                              .map( d -> d.getApplicationData()
                                                                          .getId() )
                                                              .filter( Objects::nonNull )
                                                              .toArray( Long[]::new );
                            String[] applicationNames = userConfig.getUserApplicationDatas()
                                                                  .stream()
                                                                  .filter( d -> StringUtils.equals( d.getUserData()
                                                                                                     .getUserName(),
                                                                                                    data.getUserName() ) )
                                                                  .map( d -> d.getApplicationData()
                                                                              .getName() )
                                                                  .filter( Objects::nonNull )
                                                                  .toArray( String[]::new );
                            Set<ApplicationData> applicationDatas = new HashSet<>();
                            if ( ArrayUtils.isNotEmpty( applicationIds ) )
                            {
                                applicationDatas.addAll( database.getApplications( applicationIds ) );
                            }
                            if ( ArrayUtils.isNotEmpty( applicationNames ) )
                            {
                                applicationDatas.addAll( database.getApplications( applicationNames ) );
                            }
                            CollectionUtils.filter( applicationDatas, PredicateUtils.notNullPredicate() );

                            Long[] applicationsGroupIds = userConfig.getUserApplicationsGroupDatas()
                                                                    .stream()
                                                                    .filter( d -> StringUtils.equals( d.getUserData()
                                                                                                       .getUserName(),
                                                                                                      data.getUserName() ) )
                                                                    .map( d -> d.getApplicationsGroupData()
                                                                                .getId() )
                                                                    .filter( Objects::nonNull )
                                                                    .toArray( Long[]::new );
                            String[] applicationsGroupNames = userConfig.getUserApplicationsGroupDatas()
                                                                        .stream()
                                                                        .filter( d -> StringUtils.equals( d.getUserData()
                                                                                                           .getUserName(),
                                                                                                          data.getUserName() ) )
                                                                        .map( d -> d.getApplicationsGroupData()
                                                                                    .getName() )
                                                                        .filter( Objects::nonNull )
                                                                        .toArray( String[]::new );
                            Set<ApplicationsGroupData> applicationsGroupDatas = new HashSet<>();
                            if ( ArrayUtils.isNotEmpty( applicationsGroupIds ) )
                            {
                                applicationsGroupDatas.addAll( database.getApplicationGroups( applicationsGroupIds ) );
                            }
                            if ( ArrayUtils.isNotEmpty( applicationsGroupNames ) )
                            {
                                applicationsGroupDatas.addAll( database.getApplicationGroups( applicationsGroupNames ) );
                            }
                            CollectionUtils.filter( applicationsGroupDatas, PredicateUtils.notNullPredicate() );

                            long userId = database.addUser( data );
                            if ( userId != UserData.UNDEFINED_USER )
                            {
                                UserData userData = database.getUser( userId );
                                applicationDatas.forEach( d -> database.addUserApplication( userData, d ) );
                                applicationsGroupDatas.forEach( d -> database.addUserApplicationsGroup( userData, d ) );
                                importedDataCounter += 1;
                            }
                        } );
                        dataImportedMessage = String.format( virtualUtil.getTranslation( Translation.DATA_IMPORT_OK, importedDataCounter, datas.size() ) );
                    }
                    else
                    {
                        dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                    }
                }
                else
                {
                    dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
                }
            }
            catch ( IllegalThreadStateException exception )
            {
                dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_ERROR_IMPORT );
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                try
                {
                    // delete file (srcUploadedFileName) from temporary folder
                    Files.deleteIfExists( Paths.get( folder, fileName ) );
                }
                catch ( IOException e )
                {
                    LOGGER.error( e.getMessage() );
                }
            }
        }
        else
        {
            dataImportedMessage = virtualUtil.getTranslation( Translation.DATA_IMPORT_NO );
        }
        dataImported( dataImportedMessage );
    }

    public abstract void dataImported( String message );

    public void setScheduledFuture( ScheduledFuture<?> scheduledFuture )
    {
        this.scheduledFuture = scheduledFuture;
    }

}
