/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.client.Command.XPRA;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.NotificationUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogConnectingUtil;
import org.dussan.vaadin.dvirtual.xpra.packets.base.WindowType;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.UI;

public abstract class TestApplicationRunner
    implements Runnable
{

    private static final int BYTES_LENGTH = 8192;

    private static final int COMPLETE_TESTING_SLEEP_INTERVAL = 3000; // milliseconds

    private static final int FREE_PORT_INTERVAL = 1000;

    private static final int MAX_WAIT_INTERVAL = 10000; // milliseconds

    private static final int NO_WINDOWS = 0;

    private static final int ONLY_ONE_WINDOW = 1;

    private static final int PING_INTERVAL = 5000; // milliseconds

    private static final int SLEEP_INTERVAL = 100; // milliseconds

    private static final int SSH_SLEEP_INTERVAL = 4000; // milliseconds

    private static final int STATUS_SLEEP_INTERVAL = 1000; // milliseconds

    private static final int XPRA_FIRST_INSTANCE_SLEEP_INTERVAL = 2000; // milliseconds

    private static final String FILE_SEPARATOR = File.separator;

    private static final String STRING_COMMA = ", ";

    private static final String STRING_MINUS = "-";

    private static final String STRING_REGEX_NUMBER_VALUE = "[^\\d]";

    private static final String STRING_REGEX_STRING_VALUE = "[^A-Za-z0-9]";

    private static final String STRING_SPACE = " ";

    private static final String STRING_UNDERLINE = "_";

    private static final String TEMP_DIR = System.getProperty( "java.io.tmpdir" );

    boolean oneInstanceLocked = true;

    boolean sameInstance = false;

    private int freeXpraPortForFirstInstance = UserSessionData.NEW_PORT;

    private int freeXpraPortForSecondInstance = UserSessionData.NEW_PORT;

    private String command = null;

    private String commandOptions = null;

    private List<ServerData> serverDatas = null;

    private VirtualUtil virtualUtil = null;

    private NotificationUtil notificationUtil = null;

    private ServerData xpraServerData = null;

    private UI ui = null;

    public TestApplicationRunner( UI ui, String command, String commandOptions, List<ServerData> serverDatas, NotificationUtil notificationUtil, VirtualUtil virtualUtil )
    {
        this.ui = ui;
        this.command = command;
        this.commandOptions = commandOptions;
        this.serverDatas = serverDatas;
        this.notificationUtil = notificationUtil;
        this.virtualUtil = virtualUtil;
    }

    public abstract void applicationStarted( ServerData serverData, int xpraPort );

    public abstract void applicationTestingStarted();

    public abstract void applicationTestingStopped();

    public abstract void updatingApplicationData( String iconName, boolean oneInstanceLocked, boolean sameInstance );

    private int getNumberOfWindows( int xpraPort )
    {
        int numberOfWindows = 0;
        Channel channel = null;
        Session session = null;
        InputStream inputStream = null;
        byte[] tempInputStreamValue = null;

        try
        {
            // start SSH session
            session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                             xpraServerData.getAddress(), xpraServerData.getSshPort() );
            session.connect();

            tempInputStreamValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.INFO_NUNMBER_OF_WINDOWS, xpraPort ) );
            inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            while ( inputStream.available() > 0 )
            {
                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                if ( length > 0 )
                {
                    String value = new String( tempInputStreamValue, 0, length );
                    if ( StringUtils.isNotBlank( value ) )
                    {
                        numberOfWindows = Integer.parseInt( value.replaceAll( STRING_REGEX_NUMBER_VALUE, StringUtils.EMPTY ) );
                        break;
                    }
                }
            }
            inputStream.close();
            channel.disconnect();
            session.disconnect();
        }
        catch ( InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        return numberOfWindows;
    }

    private String getWindowType( int xpraPort )
    {
        String windowType = StringUtils.EMPTY;
        Channel channel = null;
        Session session = null;
        InputStream inputStream = null;
        byte[] tempInputStreamValue = null;

        try
        {
            // start SSH session
            session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                             xpraServerData.getAddress(), xpraServerData.getSshPort() );
            session.connect();

            tempInputStreamValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.INFO_OF_PARENT_WINDOW, xpraPort ) );
            inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            while ( inputStream.available() > 0 )
            {
                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                if ( length > 0 )
                {
                    String value = new String( tempInputStreamValue, 0, length );
                    if ( StringUtils.isNotBlank( value ) )
                    {
                        windowType = value;
                        break;
                    }
                }
            }
            inputStream.close();
            channel.disconnect();
            session.disconnect();
        }
        catch ( InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        return windowType;
    }

    private void fatalError()
    {
        virtualUtil.setApplicationTestingError( true );
        ui.getSession()
          .lock();
        ui.access( () -> {
            if ( freeXpraPortForFirstInstance != UserSessionData.NEW_PORT )
            {
                StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForFirstInstance, virtualUtil );
                virtualUtil.startRunnerService( stopServerRunner );
            }
            if ( freeXpraPortForSecondInstance != UserSessionData.NEW_PORT )
            {
                StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForSecondInstance, virtualUtil );
                virtualUtil.startRunnerService( stopServerRunner );
            }
            notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_FATAL ) );
            notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
            notificationUtil.setProgressBarVisible( false );
            notificationUtil.setStatusVisible( false );
            ui.push();
        } );
        ui.getSession()
          .unlock();
    }

    @Override
    public void run()
    {
        // application testing started
        applicationTestingStarted();

        // available servers
        List<ServerData> availableServerDatas = serverDatas.stream()
                                                           .distinct()
                                                           .filter( Objects::nonNull )
                                                           .filter( ServerData::isEnabled )
                                                           .collect( Collectors.toList() );
        // no available servers
        if ( CollectionUtils.isEmpty( availableServerDatas ) )
        {
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_SERVERS ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // SSH servers, application, free Xpra port
        List<String> missingApplicationsDatas = new ArrayList<>();
        List<String> missingLocaleDatas = new ArrayList<>();
        List<String> missingXpraApplicationDatas = new ArrayList<>();
        List<String> sshServerDatas = new ArrayList<>();
        for ( ServerData serverData : availableServerDatas )
        {
            int index = serverDatas.indexOf( serverData ) + 1;
            double serversProgress = ( (double) index / serverDatas.size() ) * DialogConnectingUtil.PROGRESS_20PERCENT;
            if ( index == 1 )
            {
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_START );
                    notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS ) );
                    notificationUtil.setProgressBarVisible( true );
                    notificationUtil.setStatusVisible( true );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                try
                {
                    TimeUnit.MILLISECONDS.sleep( STATUS_SLEEP_INTERVAL );
                }
                catch ( InterruptedException exception )
                {
                    Thread.currentThread()
                          .interrupt();
                }
            }

            // progress
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setProgressValue( serversProgress );
                notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_RUNNING, index, serverDatas.size() ) );
                ui.push();
            } );
            ui.getSession()
              .unlock();

            Channel channel = null;
            Session session = null;
            InputStream inputStream = null;
            byte[] tempInputStreamValue = null;
            try (Socket socket = new Socket())
            {
                InetAddress host = InetAddress.getByName( serverData.getAddress() );
                boolean reachableHost = host.isReachable( PING_INTERVAL );
                if ( reachableHost )
                {
                    socket.setReuseAddress( true );
                    SocketAddress socketAddress = new InetSocketAddress( serverData.getAddress(), serverData.getSshPort() );
                    socket.connect( socketAddress, PING_INTERVAL );
                    if ( socket.isConnected() )
                    {
                        // SSH server
                        sshServerDatas.add( serverData.getName() );

                        // start SSH session
                        session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                         serverData.getAddress(), serverData.getSshPort() );
                        session.connect();

                        // find application
                        String inputStreamValue = null;
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL && notificationUtil.isNotStopped(); timer += SLEEP_INTERVAL )
                        {
                            boolean applicationWithPath = StringUtils.contains( command, FILE_SEPARATOR );
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION, command ) );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL && notificationUtil.isNotStopped(); t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }
                            while ( inputStream.available() > 0 )
                            {
                                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                                if ( length > 0 )
                                {
                                    inputStreamValue = new String( tempInputStreamValue, 0, length );
                                    if ( StringUtils.isNotBlank( inputStreamValue ) )
                                    {
                                        break;
                                    }
                                }
                            }
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                break;
                            }
                        }
                        if ( StringUtils.isBlank( inputStreamValue ) )
                        {
                            missingApplicationsDatas.add( serverData.getName() );
                        }

                        // find locale
                        inputStreamValue = null;
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL && notificationUtil.isNotStopped(); timer += SLEEP_INTERVAL )
                        {
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( Command.SSH.FIND_LOCALE );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL && notificationUtil.isNotStopped(); t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }
                            StringBuilder stringBuilder = new StringBuilder();
                            while ( inputStream.available() > 0 )
                            {
                                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                                if ( length > 0 )
                                {
                                    stringBuilder.append( new String( tempInputStreamValue, 0, length ) );
                                }
                            }
                            inputStreamValue = stringBuilder.toString();
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                break;
                            }
                        }
                        ApplicationLocale applicationLocale = virtualUtil.getApplicationLocale();
                        if ( StringUtils.isBlank( inputStreamValue ) || ( !StringUtils.containsIgnoreCase( inputStreamValue, applicationLocale.getLocale() )
                            && !StringUtils.containsIgnoreCase( inputStreamValue, StringUtils.join( applicationLocale.name(), ApplicationLocale.UTF8 ) ) ) )
                        {
                            missingLocaleDatas.add( serverData.getName() );
                        }

                        // find XPRA
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL && notificationUtil.isNotStopped(); timer += SLEEP_INTERVAL )
                        {
                            boolean applicationWithPath = StringUtils.contains( Command.XPRA.XPRA_APPLICATION, FILE_SEPARATOR );
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION,
                                                                                 Command.XPRA.XPRA_APPLICATION ) );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL && notificationUtil.isNotStopped(); t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }
                            while ( inputStream.available() > 0 )
                            {
                                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                                if ( length > 0 )
                                {
                                    inputStreamValue = new String( tempInputStreamValue, 0, length );
                                    if ( StringUtils.isNotBlank( inputStreamValue ) )
                                    {
                                        break;
                                    }
                                }
                            }
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                break;
                            }
                        }
                        if ( StringUtils.isBlank( inputStreamValue ) )
                        {
                            missingXpraApplicationDatas.add( serverData.getName() );
                        }
                        inputStreamValue = null;
                        session.disconnect();
                    }
                }
            }
            catch ( InterruptedException | IOException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }

            if ( notificationUtil.isStopped() )
            {
                break;
            }
        }

        // no SSH servers
        if ( notificationUtil.isNotStopped() && CollectionUtils.isEmpty( sshServerDatas ) )
        {
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_SERVERS_SSH ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // progress
        ui.getSession()
          .lock();
        ui.access( () -> {
            notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_20PERCENT );
            ui.push();
        } );
        ui.getSession()
          .unlock();

        // missing application on some servers
        if ( notificationUtil.isNotStopped() && CollectionUtils.isNotEmpty( missingApplicationsDatas ) )
        {
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                Collections.sort( missingApplicationsDatas );
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_INSTALL, StringUtils.join( missingApplicationsDatas, STRING_COMMA ) ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // missing locale on some servers
        if ( notificationUtil.isNotStopped() && CollectionUtils.isNotEmpty( missingLocaleDatas ) )
        {
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                Collections.sort( missingLocaleDatas );
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_LOCALE, virtualUtil.getApplicationLocale()
                                                                                                                               .getLocale(),
                                                                         StringUtils.join( missingLocaleDatas, STRING_COMMA ) ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // missing Xpra application on some servers
        if ( notificationUtil.isNotStopped() && CollectionUtils.isNotEmpty( missingXpraApplicationDatas ) )
        {
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                Collections.sort( missingXpraApplicationDatas );
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_XPRA, StringUtils.join( missingXpraApplicationDatas, STRING_COMMA ) ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // testing has stopped
        if ( notificationUtil.isStopped() )
        {
            // application testing stopped
            applicationTestingStopped();
            return;
        }

        // progress
        ui.getSession()
          .lock();
        ui.access( () -> {
            notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_30PERCENT );
            notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_STARTING ) );
            notificationUtil.getButton()
                            .setEnabled( false );
            notificationUtil.getButton()
                            .setText( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
            ui.push();
        } );
        ui.getSession()
          .unlock();

        // start application on the XPRA server
        Channel channel = null;
        Session session = null;
        boolean xpraFirstInstanceStarted = false;
        boolean xpraSecondInstanceStarted = false;
        ApplicationLocale applicationLocale = virtualUtil.getApplicationLocale();
        try
        {
            // xpra server
            Collections.shuffle( availableServerDatas );
            xpraServerData = availableServerDatas.get( 0 );

            // start the first instance on the XPRA server
            String tmpDir = StringUtils.join( Path.of( TEMP_DIR, XPRA.XPRA_APPLICATION ), STRING_MINUS, virtualUtil.getUserName() );
            if ( Objects.nonNull( xpraServerData ) )
            {
                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 xpraServerData.getAddress(), xpraServerData.getSshPort() );
                session.connect();

                // create .xpra folder
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.CREATE_XPRA_FOLDER, tmpDir ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // create xpra.conf file
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.CREATE_XPRA_FILE, tmpDir ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // init xpra.conf file
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.INIT_XPRA_FILE, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir,
                                                                     tmpDir, tmpDir, tmpDir ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // free port for first instance
                for ( int timer = 0; timer < SSH_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    byte[] tempInputStreamValue = new byte[BYTES_LENGTH];
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.FIND_FREE_XPRA_PORT, xpraServerData.getXpraMinPort() - FREE_PORT_INTERVAL, xpraServerData.getXpraMinPort() ) );
                    InputStream inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL && notificationUtil.isNotStopped(); t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }
                    while ( inputStream.available() > 0 )
                    {
                        int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                        if ( length > 0 )
                        {
                            String value = new String( tempInputStreamValue, 0, length );
                            if ( StringUtils.isNotBlank( value ) )
                            {
                                String[] intValues = StringUtils.split( value, virtualUtil.getLineSeparator() );
                                freeXpraPortForFirstInstance = Integer.parseInt( intValues[0].replaceAll( STRING_REGEX_NUMBER_VALUE, StringUtils.EMPTY ) );
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();
                }

                // remove zombie processes
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_ZOMBIE, virtualUtil.getUserName() ) );
                channel.connect();
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // start XPRA server
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.START_TEST, freeXpraPortForFirstInstance, xpraServerData.getAddress(), freeXpraPortForFirstInstance,
                                                                     applicationLocale.getLocale(), tmpDir,
                                                                     StringUtils.stripToEmpty( StringUtils.join( new String[] { command, commandOptions }, STRING_SPACE ) ) ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // progress
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_40PERCENT );
                    notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_STARTING ) );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                // progress
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_50PERCENT );
                    notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_STARTING ) );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                // get status of started XPRA server
                byte[] temporaryValue = new byte[BYTES_LENGTH];
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += STATUS_SLEEP_INTERVAL )
                {
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                    InputStream inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }
                    while ( inputStream.available() > 0 )
                    {
                        int length = inputStream.read( temporaryValue, 0, BYTES_LENGTH );
                        if ( length > 0 )
                        {
                            String status = new String( temporaryValue, 0, length );
                            if ( StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, freeXpraPortForFirstInstance ) )
                                && getNumberOfWindows( freeXpraPortForFirstInstance ) > NO_WINDOWS )
                            {
                                xpraFirstInstanceStarted = true;
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();

                    if ( xpraFirstInstanceStarted )
                    {
                        break;
                    }
                }
                session.disconnect();
            }

            // application started
            if ( xpraFirstInstanceStarted )
            {
                boolean breakOut = false;
                applicationStarted( xpraServerData, freeXpraPortForFirstInstance );
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( virtualUtil.isDrawingStarted() )
                    {
                        TimeUnit.MILLISECONDS.sleep( XPRA_FIRST_INSTANCE_SLEEP_INTERVAL );
                        breakOut = true;
                    }
                    else if ( virtualUtil.isNotConnectionToServerActive() )
                    {
                        breakOut = true;
                    }
                    if ( breakOut )
                    {
                        break;
                    }
                }
            }

            // application not started
            if ( virtualUtil.isNotDrawingStarted() )
            {
                virtualUtil.setApplicationTestingError( true );
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForFirstInstance, virtualUtil );
                    virtualUtil.startRunnerService( stopServerRunner );
                    notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_APPLICATION ) );
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                    notificationUtil.setProgressBarVisible( false );
                    notificationUtil.setStatusVisible( false );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();
                // application testing stopped
                applicationTestingStopped();
                return;
            }

            // fatal error
            if ( virtualUtil.isNotConnectionToServerActive() )
            {
                fatalError();
                // application testing stopped
                applicationTestingStopped();
                return;
            }

            // progress
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_60PERCENT );
                ui.push();
            } );
            ui.getSession()
              .unlock();

            // start the second instance on the XPRA server
            int numberOfWindowsXpraFirstInstance = NO_WINDOWS;
            if ( xpraFirstInstanceStarted && Objects.nonNull( xpraServerData ) )
            {
                TimeUnit.MILLISECONDS.sleep( STATUS_SLEEP_INTERVAL );
                numberOfWindowsXpraFirstInstance = getNumberOfWindows( freeXpraPortForFirstInstance );

                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 xpraServerData.getAddress(), xpraServerData.getSshPort() );
                session.connect();

                // free port for second instance
                for ( int timer = 0; timer < SSH_SLEEP_INTERVAL && notificationUtil.isNotStopped(); timer += SLEEP_INTERVAL )
                {
                    byte[] tempInputStreamValue = new byte[BYTES_LENGTH];
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.FIND_FREE_XPRA_PORT, xpraServerData.getXpraMinPort() - FREE_PORT_INTERVAL, xpraServerData.getXpraMinPort() ) );
                    InputStream inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }
                    while ( inputStream.available() > 0 )
                    {
                        int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                        if ( length > 0 )
                        {
                            String value = new String( tempInputStreamValue, 0, length );
                            if ( StringUtils.isNotBlank( value ) )
                            {
                                String[] intValues = StringUtils.split( value, virtualUtil.getLineSeparator() );
                                freeXpraPortForSecondInstance = Integer.parseInt( intValues[1].replaceAll( STRING_REGEX_NUMBER_VALUE, StringUtils.EMPTY ) );
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();
                }

                // start XPRA server
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.START_TEST, freeXpraPortForSecondInstance, xpraServerData.getAddress(), freeXpraPortForSecondInstance,
                                                                     applicationLocale.getLocale(), tmpDir,
                                                                     StringUtils.stripToEmpty( StringUtils.join( new String[] { command, commandOptions }, STRING_SPACE ) ) ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // fatal error
                if ( virtualUtil.isNotConnectionToServerActive() )
                {
                    fatalError();
                    // application testing stopped
                    applicationTestingStopped();
                    return;
                }

                // progress
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_70PERCENT );
                    notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_STARTING ) );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                // get status of started XPRA server
                byte[] temporaryValue = new byte[BYTES_LENGTH];
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += STATUS_SLEEP_INTERVAL )
                {
                    channel = session.openChannel( Command.SSH.EXECUTE );
                    ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                    InputStream inputStream = channel.getInputStream();
                    channel.connect();
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( channel.isClosed() )
                        {
                            break;
                        }
                    }
                    while ( inputStream.available() > 0 )
                    {
                        int length = inputStream.read( temporaryValue, 0, BYTES_LENGTH );
                        if ( length > 0 )
                        {
                            String status = new String( temporaryValue, 0, length );
                            if ( StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, freeXpraPortForSecondInstance ) ) )
                            {
                                xpraSecondInstanceStarted = true;
                                break;
                            }
                        }
                    }
                    inputStream.close();
                    channel.disconnect();

                    if ( xpraSecondInstanceStarted )
                    {
                        break;
                    }
                }
                session.disconnect();

                // get status of started XPRA server
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += STATUS_SLEEP_INTERVAL )
                {
                    if ( getNumberOfWindows( freeXpraPortForSecondInstance ) > NO_WINDOWS )
                    {
                        break;
                    }
                }
            }

            // fatal error
            if ( virtualUtil.isNotConnectionToServerActive() )
            {
                fatalError();
                // application testing stopped
                applicationTestingStopped();
                return;
            }

            // progress
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_80PERCENT );
                ui.push();
            } );
            ui.getSession()
              .unlock();

            // connect to the XPRA server
            if ( xpraFirstInstanceStarted && xpraSecondInstanceStarted )
            {
                // application not started
                if ( numberOfWindowsXpraFirstInstance == NO_WINDOWS )
                {
                    virtualUtil.setApplicationTestingError( true );
                    ui.getSession()
                      .lock();
                    ui.access( () -> {
                        StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForFirstInstance, virtualUtil );
                        virtualUtil.startRunnerService( stopServerRunner );
                        notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_APPLICATION ) );
                        notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                        notificationUtil.setProgressBarVisible( false );
                        notificationUtil.setStatusVisible( false );
                        ui.push();
                    } );
                    ui.getSession()
                      .unlock();
                    // application testing stopped
                    applicationTestingStopped();
                    return;
                }

                // one instance status
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    // progress
                    int tt = t;
                    ui.getSession()
                      .lock();
                    ui.access( () -> {
                        notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_80PERCENT + (double) tt / MAX_WAIT_INTERVAL );
                        ui.push();
                    } );
                    ui.getSession()
                      .unlock();

                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    int numberOfWindowsFirstInstance = getNumberOfWindows( freeXpraPortForFirstInstance );
                    int numberOfWindowsSecondInstance = getNumberOfWindows( freeXpraPortForSecondInstance );
                    oneInstanceLocked = numberOfWindowsFirstInstance > numberOfWindowsXpraFirstInstance
                        || ( numberOfWindowsFirstInstance == numberOfWindowsXpraFirstInstance && numberOfWindowsSecondInstance == NO_WINDOWS );
                    if ( oneInstanceLocked || numberOfWindowsSecondInstance == numberOfWindowsXpraFirstInstance )
                    {
                        break;
                    }

                    // fatal error
                    if ( virtualUtil.isNotConnectionToServerActive() )
                    {
                        fatalError();
                        // application testing stopped
                        applicationTestingStopped();
                        return;
                    }
                }

                // progress
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_90PERCENT );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                // one instance in the same instance status
                if ( !oneInstanceLocked )
                {
                    int numberOfWindowsXpraSecondInstance = getNumberOfWindows( freeXpraPortForSecondInstance );
                    ChildApplicationRunner childApplicationRunner =
                        new ChildApplicationRunner( virtualUtil.getUserName(), command, xpraServerData.getAddress(), freeXpraPortForSecondInstance, xpraServerData.getSshPort(), null, virtualUtil );
                    virtualUtil.startRunnerService( childApplicationRunner );
                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += ( numberOfWindowsXpraSecondInstance == NO_WINDOWS ? SLEEP_INTERVAL : 5 * SLEEP_INTERVAL ) )
                    {
                        // progress
                        int tt = t;
                        ui.getSession()
                          .lock();
                        ui.access( () -> {
                            notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_90PERCENT + (double) tt / MAX_WAIT_INTERVAL );
                            ui.push();
                        } );
                        ui.getSession()
                          .unlock();

                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                        if ( numberOfWindowsXpraSecondInstance > NO_WINDOWS )
                        {
                            String windowType = getWindowType( freeXpraPortForSecondInstance );
                            int aa = getNumberOfWindows( freeXpraPortForSecondInstance );
                            sameInstance = aa == numberOfWindowsXpraSecondInstance;
                            if ( ( sameInstance && numberOfWindowsXpraSecondInstance == ONLY_ONE_WINDOW && StringUtils.contains( windowType, WindowType.DIALOG.toString() ) ) || !sameInstance )
                            {
                                oneInstanceLocked = sameInstance && numberOfWindowsXpraSecondInstance == ONLY_ONE_WINDOW && StringUtils.contains( windowType, WindowType.DIALOG.toString() );
                                break;
                            }
                        }
                        else
                        {
                            numberOfWindowsXpraSecondInstance = getNumberOfWindows( freeXpraPortForSecondInstance );
                        }

                        // fatal error
                        if ( virtualUtil.isNotConnectionToServerActive() )
                        {
                            fatalError();
                            // application testing stopped
                            applicationTestingStopped();
                            return;
                        }
                    }

                    // application not started completely
                    if ( numberOfWindowsXpraSecondInstance == NO_WINDOWS )
                    {
                        virtualUtil.setApplicationTestingError( true );
                        ui.getSession()
                          .lock();
                        ui.access( () -> {
                            StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForFirstInstance, virtualUtil );
                            virtualUtil.startRunnerService( stopServerRunner );
                            notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_INSTANCE ) );
                            notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                            notificationUtil.setProgressBarVisible( false );
                            notificationUtil.setStatusVisible( false );
                            ui.push();
                        } );
                        ui.getSession()
                          .unlock();
                        // application testing stopped
                        applicationTestingStopped();
                        return;
                    }
                }

                // progress
                ui.getSession()
                  .lock();
                ui.access( () -> {
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_99PERCENT );
                    notificationUtil.setStatus( virtualUtil.getTranslation( Translation.APPLICATION_TEST_STATUS_STARTED ) );
                    ui.push();
                } );
                ui.getSession()
                  .unlock();

                // updating application data
                String iconName = command.replaceAll( STRING_REGEX_STRING_VALUE, STRING_UNDERLINE );
                updatingApplicationData( iconName, oneInstanceLocked, sameInstance );
            }
        }
        catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        // stop the second Xpra server
        if ( xpraSecondInstanceStarted )
        {
            StopServerRunner stopServerRunner = new StopServerRunner( xpraServerData, freeXpraPortForSecondInstance, virtualUtil );
            virtualUtil.startRunnerService( stopServerRunner );
        }

        if ( !xpraFirstInstanceStarted || !xpraSecondInstanceStarted )
        {
            // (first or second or both) instances not started
            virtualUtil.setApplicationTestingError( true );
            ui.getSession()
              .lock();
            ui.access( () -> {
                notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_ERROR_USER ) );
                notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                notificationUtil.setProgressBarVisible( false );
                notificationUtil.setStatusVisible( false );
                ui.push();
            } );
            ui.getSession()
              .unlock();
        }
        else
        {
            // the application test is complete
            ui.getSession()
              .lock();
            ui.access( () -> {
                try
                {
                    TimeUnit.MILLISECONDS.sleep( COMPLETE_TESTING_SLEEP_INTERVAL );
                    notificationUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_TEST_COMPLETE ) );
                    notificationUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                    notificationUtil.setProgressBarVisible( false );
                    notificationUtil.setStatusVisible( false );
                    ui.push();
                }
                catch ( InterruptedException exception )
                {
                    Thread.currentThread()
                          .interrupt();
                }
            } );
            ui.getSession()
              .unlock();
        }

        // application testing stopped
        applicationTestingStopped();
    }

}
