/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.applications;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.LongStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ServersGroupColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.checkbox.Checkbox;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.Grid.SelectionMode;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.binder.Binder;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.validator.StringLengthValidator;
import com.vaadin.flow.data.value.ValueChangeMode;
import com.vaadin.flow.server.StreamResource;

public abstract class ApplicationEditDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = -8620623373355072512L;

    private static final int FORM_COLUMNS = 1;

    private static final int NO_ELEMENTS = 0;

    private static final int NO_LIMITS = Integer.MAX_VALUE;

    private static final int RANDOM_STRING_LENGTH = 30;

    private static final int VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_ICON = "icon";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_DIALOG = "edit-dialog";

    private static final String CLASS_DIALOG_BUTTONS = "edit-dialog-buttons";

    private static final String CLASS_GRID_CELL_ACTIVE = "active";

    private static final String CLASS_GRID_CELL_NOT_ACTIVE = StringUtils.EMPTY;

    private static final String CLASS_GRID_CELL_THEME = "grid-cell-theme";

    private static final String CLASS_GRID_LABEL = "grid-label";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String STYLE_BUTTON_BACKGROUND_COLOR = "background-color";

    private static final String STYLE_BUTTON_BACKGROUND_COLOR_VALUE = "var(--lumo-contrast-10pct)";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_GRID_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_HEIGHT_35VH_VALUE = "35vh";

    private static final String STYLE_HEIGHT_ICON_VALUE = "var(--lumo-icon-size-m)";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_TOP = "margin-top";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_MIN_WIDTH = "min-width";

    private static final String STYLE_PADDING_14PX_VALUE = "14px";

    private static final String STYLE_PADDING_BOTTOM = "padding-bottom";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_TOP = "padding-top";

    private static final String STYLE_PADDING_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_VERTICAL_ALIGN = "vertical-align";

    private static final String STYLE_VERTICAL_ALIGN_VALUE = "middle";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_50PX_VALUE = "50px";

    private static final String STYLE_WIDTH_ICON_VALUE = "var(--lumo-icon-size-m)";

    private static final VaadinIcon DEFAULT_ICON = VaadinIcon.AREA_SELECT;

    private int oneInstance = ApplicationData.DEFAULT_ONE_INSTANCE;

    private transient List<ServersGroupData> selectedServersGroupDatas = null;

    private transient Map<ServersGroupColumns, String> filterMap = null;

    private byte[] buttonByteIcon = null;

    private Button iconButton = null;

    private Checkbox enabledCheckbox = null;

    private Checkbox oneInstanceCheckbox = null;

    private Checkbox selectAllCheckbox = null;

    private Checkbox serverAutoSelectCheckbox = null;

    private DataProvider<ServersGroupData, Void> dataProvider = null;

    private Grid<ServersGroupData> serversGroupGrid = null;

    private Icon buttonIcon = null;

    private Image buttonImage = null;

    private TextField commandTextField = null;

    private TextField commandOptionsTextField = null;

    private TextField nameTextField = null;

    private TextField noteTextField = null;

    private DialogInfoUtil dialogInfoUtil = null;

    @SuppressWarnings( "deprecation" )
    protected ApplicationEditDialog( UI ui, long userId, ApplicationData applicationData, Database database, boolean testingApplicationEnabled, VirtualUtil virtualUtil )
    {
        filterMap = new EnumMap<>( ServersGroupColumns.class );
        selectedServersGroupDatas = Objects.requireNonNullElse( applicationData.getServersGroupDatas(), new ArrayList<>() );
        oneInstance = Objects.requireNonNullElse( applicationData.getOneInstance(), oneInstance );
        dialogInfoUtil = new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_IMPORT_WAIT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );

        // close/save/test buttons
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        Button saveButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SAVE ) );
        saveButton.setEnabled( virtualUtil.isApplicationTesting() && virtualUtil.isNotApplicationTestingError() );
        saveButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        saveButton.getElement()
                  .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        saveButton.getStyle()
                  .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );
        Button testApplicationButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_APPLICATION_TEST ) );
        testApplicationButton.setEnabled( virtualUtil.isNotApplicationTestingError() && testingApplicationEnabled && StringUtils.isNotBlank( applicationData.getCommand() )
            && applicationData.getId() != ApplicationData.NEW_APPLICATION );
        testApplicationButton.getElement()
                             .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        testApplicationButton.addClickListener( event -> {
            if ( database.isApplicationsDatabaseEmpty() || database.isServerGroupsDatabaseEmpty() || database.isServersDatabaseEmpty() || database.isUsersDatabaseEmpty() )
            {
                // setup not completed
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.SETUP_TESTING ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
            else if ( virtualUtil.isSetupUserActive() )
            {
                // setup user cannot run test
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.SETUP_USER ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
            else if ( CollectionUtils.isEmpty( database.getAllUserSessions( userId ) ) )
            {
                // only user with no active sessions can run test
                if ( applicationData.getOneInstance() > ApplicationData.DEFAULT_ONE_INSTANCE )
                {
                    if ( applicationData.getOneInstance() <= ApplicationData.ONE_INSTANCE_ENABLED_LOCKED )
                    {
                        applicationData.setOneInstance( ApplicationData.ONE_INSTANCE_ENABLED_LOCKED );
                    }
                    else
                    {
                        applicationData.setOneInstance( oneInstanceCheckbox.getValue()
                                                                           .booleanValue() ? ApplicationData.ONE_INSTANCE_ENABLED : ApplicationData.ONE_INSTANCE_DISABLED );
                    }
                }
                applicationData.setCommand( commandTextField.getValue() );
                applicationData.setCommandOptions( commandOptionsTextField.getValue() );
                applicationData.setName( nameTextField.getValue() );
                applicationData.setNote( noteTextField.getValue() );
                applicationData.setServerAutoSelect( serverAutoSelectCheckbox.getValue() );
                applicationData.setEnabled( enabledCheckbox.getValue() );
                applicationData.setServersGroupDatas( selectedServersGroupDatas );
                applicationTest( applicationData );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_TEST_DISABLED ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( saveButton, testApplicationButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.addClassName( CLASS_DIALOG_BUTTONS );

        // button icon
        String iconName = DEFAULT_ICON.create()
                                      .getElement()
                                      .getAttribute( ATTRIBUTE_ICON );
        IconData iconData = applicationData.getIconData();
        iconName = Objects.isNull( iconData ) || StringUtils.isEmpty( iconData.getName() ) || !StringUtils.contains( iconName, ICON_NAME_SEPARATOR ) ? iconName : iconData.getName();
        buttonIcon = new Icon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                               iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
        buttonIcon.setId( iconName );

        // button image
        if ( Objects.nonNull( iconData ) && Objects.nonNull( iconData.getIcon() ) )
        {
            StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
            buttonImage = new Image( streamResource, iconData.getName() );
            buttonImage.setId( iconData.getName() );
            buttonImage.getStyle()
                       .set( STYLE_HEIGHT, STYLE_HEIGHT_ICON_VALUE )
                       .set( STYLE_WIDTH, STYLE_WIDTH_ICON_VALUE )
                       .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            buttonByteIcon = iconData.getIcon();
        }

        // icon button
        iconButton = new Button( Objects.requireNonNullElse( buttonImage, buttonIcon ), event -> {
            ApplicationIconsDialog applicationIconsDialog = new ApplicationIconsDialog( buttonIcon.getId()
                                                                                                  .get(),
                                                                                        virtualUtil )
            {
                private static final long serialVersionUID = -6084317688810452554L;

                @Override
                public void importingData( String folder, String fileName )
                {
                    try
                    {
                        String id = RandomStringUtils.randomAlphanumeric( RANDOM_STRING_LENGTH );
                        byte[] fileContent = Files.readAllBytes( Paths.get( folder, fileName ) );
                        StreamResource streamResource = new StreamResource( id, () -> new ByteArrayInputStream( fileContent ) );
                        buttonImage = new Image( streamResource, id );
                        buttonImage.setId( id );
                        buttonImage.getStyle()
                                   .set( STYLE_HEIGHT, STYLE_HEIGHT_ICON_VALUE )
                                   .set( STYLE_WIDTH, STYLE_WIDTH_ICON_VALUE )
                                   .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
                        iconButton.setIcon( buttonImage );
                        iconButton.setId( id );
                        buttonByteIcon = fileContent;
                        saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                        Files.deleteIfExists( Paths.get( folder, fileName ) );
                        close();
                    }
                    catch ( IOException exception )
                    {
                        getUI().ifPresent( ui -> ui.access( () -> {
                            dialogInfoUtil.setButtonsVisible( true );
                            dialogInfoUtil.setMessage( virtualUtil.getTranslation( Translation.DATA_ERROR_ICON, exception.getMessage() ) );
                            dialogInfoUtil.open();
                            ui.push();
                        } ) );
                    }
                }

                @Override
                public void importingDataError( String errorMessage )
                {
                    getUI().ifPresent( ui -> ui.access( () -> {
                        dialogInfoUtil.setButtonsVisible( true );
                        dialogInfoUtil.setMessage( virtualUtil.getTranslation( Translation.DATA_ERROR_ICON, errorMessage ) );
                        dialogInfoUtil.open();
                        ui.push();
                    } ) );
                }
            };
            applicationIconsDialog.getSelectButton()
                                  .addClickListener( e -> {
                                      buttonImage = null;
                                      buttonByteIcon = null;
                                      buttonIcon = applicationIconsDialog.getSelectedIcon();
                                      iconButton.setIcon( buttonIcon );
                                      iconButton.setId( applicationIconsDialog.getSelectedIconName() );
                                      saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                                  } );
            applicationIconsDialog.open();
        } );
        iconButton.setId( iconName );
        iconButton.getStyle()
                  .set( STYLE_BUTTON_BACKGROUND_COLOR, STYLE_BUTTON_BACKGROUND_COLOR_VALUE )
                  .set( STYLE_MARGIN_TOP, STYLE_MARGIN_VALUE )
                  .set( STYLE_MIN_WIDTH, STYLE_WIDTH_50PX_VALUE )
                  .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                  .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        Label iconButtonLabel = new Label( virtualUtil.getTranslation( Translation.FIELD_ICON ) );
        iconButtonLabel.setFor( virtualUtil.getTranslation( Translation.FIELD_SERVERS_GROUP ) );
        iconButtonLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout iconLayout = new VerticalLayout( iconButtonLabel, iconButton );
        iconLayout.setDefaultHorizontalComponentAlignment( Alignment.BASELINE );
        iconLayout.setMargin( false );
        iconLayout.setPadding( false );
        iconLayout.setSpacing( false );
        iconLayout.setSizeUndefined();
        iconLayout.setWidth( STYLE_WIDTH_50PX_VALUE );

        // name
        nameTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NAME ), StringUtils.stripToEmpty( applicationData.getName() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        nameTextField.setAutofocus( true );
        nameTextField.setRequiredIndicatorVisible( true );
        nameTextField.setValueChangeMode( ValueChangeMode.LAZY );
        nameTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameTextField.getStyle()
                     .set( STYLE_PADDING_TOP, STYLE_PADDING_VALUE );

        // server auto select
        serverAutoSelectCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_SERVER_SELECT ),
                                                 applicationData.getId() == ApplicationData.NEW_APPLICATION ? ApplicationData.DEFAULT_SERVER_AUTO_SELECT : applicationData.isServerAutoSelect() );
        serverAutoSelectCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // one instance
        oneInstance = Objects.isNull( applicationData.getOneInstance() ) ? ApplicationData.DEFAULT_ONE_INSTANCE : oneInstance;
        oneInstanceCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_APPLICATION_INSTANCE ), oneInstance < ApplicationData.ONE_INSTANCE_DISABLED );
        oneInstanceCheckbox.addValueChangeListener( event -> {
            if ( oneInstance <= ApplicationData.ONE_INSTANCE_ENABLED_LOCKED )
            {
                if ( !event.getValue()
                           .booleanValue() )
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( oneInstance == ApplicationData.DEFAULT_ONE_INSTANCE ? Translation.APPLICATION_INSTANCE_TEST
                                    : Translation.APPLICATION_INSTANCE_UNLOCK ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                oneInstanceCheckbox.setValue( Boolean.TRUE );
            }
            else
            {
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) );
            }
        } );

        // enabled
        enabledCheckbox = new Checkbox( virtualUtil.getTranslation( Translation.FIELD_ENABLED ),
                                        applicationData.getId() == ApplicationData.NEW_APPLICATION ? ApplicationData.DEFAULT_ENABLED : applicationData.isEnabled() );
        enabledCheckbox.addValueChangeListener( event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !Objects.equals( event.getValue(), event.getOldValue() ) ) );
        // name layout: icon, name, server auto select, maximized, same instance, enabled
        HorizontalLayout nameLayout = new HorizontalLayout( iconLayout, nameTextField, serverAutoSelectCheckbox, oneInstanceCheckbox, enabledCheckbox );
        nameLayout.expand( nameTextField );
        nameLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        nameLayout.setPadding( false );
        nameLayout.setAlignSelf( Alignment.START, iconLayout );
        nameLayout.setSpacing( true );
        nameLayout.setWidth( STYLE_WIDTH_50PX_VALUE );

        // command
        commandTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_COMMAND ), StringUtils.stripToEmpty( applicationData.getCommand() ), event -> {
            testApplicationButton.setEnabled( StringUtils.isNotBlank( commandTextField.getValue() ) && applicationData.getId() != ApplicationData.NEW_APPLICATION );
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !event.getValue()
                                                                          .equals( event.getOldValue() ) );
        } );
        commandTextField.setRequiredIndicatorVisible( true );
        commandTextField.setValueChangeMode( ValueChangeMode.LAZY );
        commandTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        commandTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // command options
        commandOptionsTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_COMMAND_OPTIONS ), StringUtils.stripToEmpty( applicationData.getCommandOptions() ),
                                                 event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        commandOptionsTextField.setValueChangeMode( ValueChangeMode.LAZY );
        commandOptionsTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        commandOptionsTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // command layout: command, command options
        HorizontalLayout commandLayout = new HorizontalLayout( commandTextField, commandOptionsTextField );
        commandLayout.expand( commandTextField, commandOptionsTextField );
        commandLayout.setDefaultVerticalComponentAlignment( Alignment.BASELINE );
        commandLayout.setPadding( false );
        commandLayout.setSpacing( true );
        commandLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );

        // note
        noteTextField = new TextField( virtualUtil.getTranslation( Translation.FIELD_NOTE ), StringUtils.stripToEmpty( applicationData.getNote() ),
                                       event -> saveButton.setEnabled( isAllRequiredValuesNonEmpty() && !StringUtils.equals( event.getValue(), event.getOldValue() ) ) );
        noteTextField.setValueChangeMode( ValueChangeMode.LAZY );
        noteTextField.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteTextField.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        noteTextField.getStyle()
                     .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_14PX_VALUE );

        // servers group data provider
        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();
            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( order.getDirection()
                                 .equals( SortDirection.DESCENDING ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            List<ServersGroupData> datas = new ArrayList<>( database.getServerGroups( offset, limit, filterMap, orders ) );
            return datas.stream();
        }, query -> {
            long[] selectedServersGroupIds = selectedServersGroupDatas.stream()
                                                                      .mapToLong( ServersGroupData::getId )
                                                                      .toArray();
            long filteredSelectedServersGroupSize = database.getServerGroups( 0, NO_LIMITS, filterMap, null )
                                                            .stream()
                                                            .filter( serversGroupData -> LongStream.of( selectedServersGroupIds )
                                                                                                   .anyMatch( selectedServersGroupId -> selectedServersGroupId == serversGroupData.getId() ) )
                                                            .count();
            Long serversGroupRows = database.getServerGroupsRows( filterMap );
            selectAllCheckbox.setValue( filteredSelectedServersGroupSize == serversGroupRows && serversGroupRows > NO_ELEMENTS );
            return serversGroupRows.intValue();
        } );

        // servers group
        serversGroupGrid = new Grid<>();
        serversGroupGrid.setDataProvider( dataProvider );
        serversGroupGrid.setId( CLASS_GRID_CELL_THEME );
        serversGroupGrid.setPageSize( VISIBLE_ROWS );
        serversGroupGrid.setSelectionMode( SelectionMode.NONE );
        serversGroupGrid.getStyle()
                        .set( STYLE_DISPLAY, STYLE_DISPLAY_GRID_VALUE )
                        .set( STYLE_WIDTH, STYLE_WIDTH_100PERCENT_VALUE );

        // servers group: selected
        selectAllCheckbox = new Checkbox();
        selectAllCheckbox.addClickListener( event -> {
            selectedServersGroupDatas.clear();
            if ( Boolean.TRUE.equals( event.getSource()
                                           .getValue() ) )
            {
                selectedServersGroupDatas.addAll( database.getServerGroups( 0, NO_LIMITS, filterMap, null ) );
            }
            saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
            dataProvider.refreshAll();
        } );
        serversGroupGrid.addColumn( new ComponentRenderer<>( serversGroupData -> {
            Checkbox checkbox = new Checkbox();
            checkbox.setValue( selectedServersGroupDatas.stream()
                                                        .anyMatch( selectedServersGroupData -> Objects.equals( selectedServersGroupData.getId(), serversGroupData.getId() ) ) );
            checkbox.addClickListener( event -> {
                if ( Boolean.TRUE.equals( event.getSource()
                                               .getValue() ) )
                {
                    selectedServersGroupDatas.add( serversGroupData );
                }
                else
                {
                    selectedServersGroupDatas.removeIf( selectedServersGroupData -> Objects.equals( selectedServersGroupData.getId(), serversGroupData.getId() ) );
                }
                saveButton.setEnabled( isAllRequiredValuesNonEmpty() );
                dataProvider.refreshAll();
            } );
            return checkbox;
        } ) )
                        .setFlexGrow( 0 )
                        .setFrozen( true )
                        .setKey( ServersGroupColumns.SELECTED.getName() )
                        .setResizable( false )
                        .setSortable( false )
                        .setWidth( STYLE_WIDTH_50PX_VALUE )
                        .setHeader( selectAllCheckbox );

        // servers group: name
        Grid.Column<ServersGroupData> nameColumn = //
            serversGroupGrid.addColumn( ServersGroupData::getName )
                            .setFlexGrow( 1 )
                            .setFrozen( true )
                            .setKey( ServersGroupColumns.NAME.getName() )
                            .setResizable( true )
                            .setSortable( true )
                            .setSortProperty( ServersGroupColumns.NAME.getName() )
                            .setWidth( STYLE_WIDTH_150PX_VALUE )
                            .setHeader( virtualUtil.getTranslation( ServersGroupColumns.NAME.getValueForTranslation() ) );

        // servers group: note
        Grid.Column<ServersGroupData> noteColumn = //
            serversGroupGrid.addColumn( ServersGroupData::getNote )
                            .setFlexGrow( 3 )
                            .setKey( ServersGroupColumns.NOTE.getName() )
                            .setResizable( true )
                            .setSortable( true )
                            .setSortProperty( ServersGroupColumns.NOTE.getName() )
                            .setHeader( virtualUtil.getTranslation( ServersGroupColumns.NOTE.getValueForTranslation() ) );

        // servers group: enabled
        Grid.Column<ServersGroupData> enabledColumn = //
            serversGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, serversGroupData ) -> {
                Icon icon = serversGroupData.isEnabled()
                                            .booleanValue() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                icon.getStyle()
                    .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                    .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
                div.add( icon );
                div.getStyle()
                   .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
            } ) )
                            .setFlexGrow( 0 )
                            .setKey( ServersGroupColumns.ENABLED.getName() )
                            .setResizable( true )
                            .setSortable( true )
                            .setSortProperty( ServersGroupColumns.ENABLED.getName() )
                            .setWidth( STYLE_WIDTH_130PX_VALUE )
                            .setHeader( virtualUtil.getTranslation( ServersGroupColumns.ENABLED.getValueForTranslation() ) );

        Label serversGroupGridLabel = new Label( virtualUtil.getTranslation( Translation.FIELD_SERVERS_GROUP ) );
        serversGroupGridLabel.setFor( virtualUtil.getTranslation( Translation.FIELD_SERVERS_GROUP ) );
        serversGroupGridLabel.addClassName( CLASS_GRID_LABEL );

        VerticalLayout serversGroupGridLayout = new VerticalLayout( serversGroupGridLabel, serversGroupGrid );
        serversGroupGridLayout.setMargin( false );
        serversGroupGridLayout.setPadding( false );
        serversGroupGridLayout.setSpacing( false );
        serversGroupGridLayout.setSizeUndefined();
        serversGroupGridLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        serversGroupGridLayout.setHeight( STYLE_HEIGHT_35VH_VALUE );

        // legend
        Label legendCommand = new Label( virtualUtil.getTranslation( Translation.APPLICATION_COMMAND ) );
        Label legendCommandOptions = new Label( virtualUtil.getTranslation( Translation.APPLICATION_COMMAND_OPTIONS ) );
        Label legendOneInstance = new Label( virtualUtil.getTranslation( Translation.APPLICATION_INSTANCE ) );
        VerticalLayout legendLayout = new VerticalLayout( legendCommand, legendCommandOptions, legendOneInstance );
        legendLayout.setSpacing( false );
        legendLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        legendLayout.getStyle()
                    .set( STYLE_PADDING_BOTTOM, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_LEFT, STYLE_PADDING_VALUE )
                    .set( STYLE_PADDING_RIGHT, STYLE_PADDING_VALUE );

        Binder<ApplicationData> applicationDataBinder = new Binder<>();
        applicationDataBinder.forField( nameTextField )
                             .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                             .bind( ApplicationData::getName, ApplicationData::setName );
        applicationDataBinder.forField( commandTextField )
                             .withValidator( new StringLengthValidator( virtualUtil.getTranslation( Translation.DATA_VALUE ), 1, null ) )
                             .bind( ApplicationData::getCommand, ApplicationData::setCommand );
        applicationDataBinder.bind( commandOptionsTextField, ApplicationData::getCommandOptions, ApplicationData::setCommandOptions );
        applicationDataBinder.bind( serverAutoSelectCheckbox, ApplicationData::isServerAutoSelect, ApplicationData::setServerAutoSelect );
        applicationDataBinder.bind( noteTextField, ApplicationData::getNote, ApplicationData::setNote );
        applicationDataBinder.bind( enabledCheckbox, ApplicationData::isEnabled, ApplicationData::setEnabled );

        closeButton.addClickListener( event -> close() );
        saveButton.addClickListener( event -> {
            ApplicationData data = new ApplicationData();
            IconData d = new IconData( IconData.NEW_ICON, iconButton.getId()
                                                                    .orElse( IconData.NEW_DATA ),
                                       buttonByteIcon );
            data.setIconData( d );
            if ( oneInstance > ApplicationData.DEFAULT_ONE_INSTANCE )
            {
                if ( oneInstance <= ApplicationData.ONE_INSTANCE_ENABLED_LOCKED )
                {
                    oneInstance = ApplicationData.ONE_INSTANCE_ENABLED_LOCKED;
                }
                else
                {
                    oneInstance = oneInstanceCheckbox.getValue()
                                                     .booleanValue() ? ApplicationData.ONE_INSTANCE_ENABLED : ApplicationData.ONE_INSTANCE_DISABLED;
                }
            }
            data.setId( applicationData.getId() );
            data.setCommand( applicationData.getCommand() );
            data.setName( applicationData.getName() );
            data.setNote( applicationData.getNote() );
            data.setOneInstance( oneInstance );
            data.setSameInstance( applicationData.isSameInstance() );
            data.setServerAutoSelect( applicationData.isServerAutoSelect() );
            data.setEnabled( applicationData.isEnabled() );
            data.setServersGroupDatas( selectedServersGroupDatas );
            if ( applicationDataBinder.writeBeanIfValid( data ) )
            {
                savingData( data );
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );

        // filter: name
        HeaderRow filterRow = serversGroupGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ServersGroupColumns.NAME );
            }
            else
            {
                filterMap.put( ServersGroupColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ServersGroupColumns.NOTE );
            }
            else
            {
                filterMap.put( ServersGroupColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // filter: enabled
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ServersGroupColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ServersGroupColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ServersGroupColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );

        // form layout
        setFormLayout( new FormLayout( nameLayout, commandLayout, noteTextField, serversGroupGridLayout, legendLayout, buttonsLayout ) );
        getFormLayout().setClassName( CLASS_DIALOG );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );

        // set active/selected rows
        serversGroupGrid.getColumnByKey( ServersGroupColumns.ENABLED.getName() )
                        .setClassNameGenerator( serversGroupData -> getCellClassName( serversGroupData.getId() ) );
        serversGroupGrid.getColumnByKey( ServersGroupColumns.NAME.getName() )
                        .setClassNameGenerator( serversGroupData -> getCellClassName( serversGroupData.getId() ) );
        serversGroupGrid.getColumnByKey( ServersGroupColumns.NOTE.getName() )
                        .setClassNameGenerator( serversGroupData -> getCellClassName( serversGroupData.getId() ) );
        serversGroupGrid.getColumnByKey( ServersGroupColumns.SELECTED.getName() )
                        .setClassNameGenerator( serversGroupData -> getCellClassName( serversGroupData.getId() ) );
    }

    public abstract void applicationTest( ApplicationData applicationData );

    public abstract void savingData( ApplicationData applicationData );

    private boolean isAllRequiredValuesNonEmpty()
    {
        return ( StringUtils.isNotBlank( nameTextField.getValue() ) && StringUtils.isNotBlank( commandTextField.getValue() ) && CollectionUtils.isNotEmpty( selectedServersGroupDatas ) );
    }

    private String getCellClassName( Long serversGroupDataId )
    {
        if ( selectedServersGroupDatas.stream()
                                      .anyMatch( selectedServersGroupData -> Objects.equals( selectedServersGroupData.getId(), serversGroupDataId ) ) )
        {
            return CLASS_GRID_CELL_ACTIVE;
        }
        return CLASS_GRID_CELL_NOT_ACTIVE;
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        if ( Objects.nonNull( dialogInfoUtil ) )
        {
            dialogInfoUtil.close();
        }
    }

}
