/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.navigation;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.ImportDataRunner;
import org.dussan.vaadin.dvirtual.ui.applicationgroups.ApplicationGroupsUI;
import org.dussan.vaadin.dvirtual.ui.applications.ApplicationsUI;
import org.dussan.vaadin.dvirtual.ui.servergroups.ServerGroupsUI;
import org.dussan.vaadin.dvirtual.ui.servers.ServersUI;
import org.dussan.vaadin.dvirtual.ui.users.UsersUI;
import org.dussan.vaadin.dvirtual.ui.usersessions.UserSessionsUI;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DataImportDialog;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.contextmenu.ContextMenu;
import com.vaadin.flow.component.contextmenu.MenuItem;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Hr;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.html.Nav;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.IronIcon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.server.StreamResource;

public abstract class NavigationBar
    extends Nav
{

    private static final long serialVersionUID = 6406330880002801113L;

    private static final String ATTRIBUTE_APPLICATIONS = "Applications";

    private static final String CLASS_ACTIVE_BUTTON = "active";

    private static final String CLASS_NAVIGATION_BAR = "navigation-bar";

    private static final String CLASS_LOG_OUT = "logout";

    private static final String CLASS_USER = "user";

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String STYLE_FONT_SIZE = "font-size";

    private static final String STYLE_FONT_SIZE_VALUE = "var(--lumo-font-size-xxs)";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_VALUE = "var(--lumo-icon-size-s)";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_LEFT_VALUE = "auto";

    private static final String STYLE_PADDING = "padding";

    private static final String STYLE_PADDING_LEFT = "padding-left";

    private static final String STYLE_PADDING_LEFT_0PX_VALUE = "0px";

    private static final String STYLE_PADDING_LEFT_28PX_VALUE = "28px";

    private static final String STYLE_PADDING_RIGHT = "padding-right";

    private static final String STYLE_PADDING_RIGHT_CALC_VALUE = "calc(var(--lumo-space-l) + var(--lumo-border-radius) / 4)";

    private static final String STYLE_PADDING_RIGHT_VALUE = "0.5em";

    private static final String STYLE_PADDING_VALUE = "0.25em";

    private static final String STYLE_VERTICAL_ALIGN = "vertical-align";

    private static final String STYLE_VERTICAL_ALIGN_VALUE = "middle";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_VALUE = "var(--lumo-icon-size-s)";

    private boolean fullScreen = false;

    private double navigationBarHeight = 0;

    private Button applicationsButton = null;

    private Button newItemButton = null;

    private Button serverGroupsButton = null;

    private Button serversButton = null;

    private Button applicationGroupsButton = null;

    private Button userButton = null;

    private Button usersButton = null;

    private Button userSessionsButton = null;

    private ContextMenu applicationsContextMenu = null;

    private Database database = null;

    private DataImportDialog dataImportDialog = null;

    private DialogInfoUtil dataImportWaitDialogUtil = null;

    private DialogInfoUtil dialogInfoUtil = null;

    private VirtualUtil virtualUtil = null;

    private MenuItem applicationsMenuItem = null;

    private MenuItem importsDataMenuItem = null;

    private MenuItem logOutMenuItem = null;

    private MenuItem serverGroupsMenuItem = null;

    private MenuItem serversMenuItem = null;

    private MenuItem userGroupsMenuItem = null;

    private MenuItem usersMenuItem = null;

    private MenuItem userSessionsMenuItem = null;

    private ApplicationGroupsUI applicationGroupsUI = null;

    private ApplicationsUI applicationsUI = null;

    private ServerGroupsUI serverGroupsUI = null;

    private ServersUI serversUI = null;

    private UserSessionsUI userSessionsUI = null;

    private UsersUI usersUI = null;

    private UserData userData = null;

    protected NavigationBar( UserData userData, Database database, VirtualUtil virtualUtil )
    {
        setSizeUndefined();
        addClassName( CLASS_NAVIGATION_BAR );
        this.userData = userData;
        this.database = database;
        this.virtualUtil = virtualUtil;

        // log out button
        Button logOutButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_LOG_OUT ), new Icon( VaadinIcon.SIGN_OUT ), event -> logOutUser() );
        logOutButton.addClassName( CLASS_LOG_OUT );

        // user button
        userButton = new Button( StringUtils.defaultIfBlank( userData.getFullName(), userData.getUserName() ), new Icon( VaadinIcon.USER ), event -> UI.getCurrent()
                                                                                                                                                       .getPage()
                                                                                                                                                       .executeJs( Command.JS.SHOW_MENU_BAR_ITEMS ) );
        userButton.addClassName( CLASS_USER );

        if ( userData.isAdministrator()
                     .booleanValue() )
        {
            // applications button
            applicationsButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS ), new Icon( VaadinIcon.FILE_O ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive || virtualUtil.isApplicationTesting() )
                {
                    disableMenuItems();
                    applicationsMenuItem.setEnabled( true );
                    virtualUtil.setApplicationTestingError( false );
                    virtualUtil.setApplicationTestingEditDialogOpened( false );
                    applicationsUI = new ApplicationsUI( userData.getId(), database, virtualUtil )
                    {
                        private static final long serialVersionUID = -4305517100785200989L;

                        @Override
                        public void testApplication( ApplicationData applicationData )
                        {
                            virtualUtil.setApplicationTestingError( false );
                            testingApplication( applicationData );
                        }

                    };
                    addComponent( applicationsUI );

                    Button button = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS_NEW ), new Icon( VaadinIcon.FILE_O ), e -> {
                        if ( !virtualUtil.isApplicationTesting() )
                        {
                            if ( Objects.nonNull( database ) && Objects.nonNull( applicationsUI ) )
                            {
                                applicationsUI.addNewApplication();
                                UI.getCurrent()
                                  .getPage()
                                  .executeJs( Command.JS.FULL_SCREEN_FINAL );
                            }
                            else
                            {
                                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_APPLICATION ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                            }
                        }
                    } );
                    addNewItemButton( button );
                    UI.getCurrent()
                      .getPage()
                      .executeJs( Command.JS.FULL_SCREEN_FINAL );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( applicationsUI );
                    applicationsUI.removeAll();
                    applicationsUI = null;
                    removeNewItemButton();
                }
            } );
            applicationsButton.getElement()
                              .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS ) );

            // application groups
            applicationGroupsButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS ), new Icon( VaadinIcon.COPY_O ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive )
                {
                    disableMenuItems();
                    userGroupsMenuItem.setEnabled( true );
                    applicationGroupsUI = new ApplicationGroupsUI( database, virtualUtil );
                    addComponent( applicationGroupsUI );

                    Button button = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS_NEW ), new Icon( VaadinIcon.COPY_O ), e -> {
                        if ( Objects.nonNull( database ) && Objects.nonNull( applicationGroupsUI ) )
                        {
                            applicationGroupsUI.addNewApplicationsGroup();
                            UI.getCurrent()
                              .getPage()
                              .executeJs( Command.JS.FULL_SCREEN_FINAL );
                        }
                        else
                        {
                            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_APPLICATIONS_GROUP ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                        }
                    } );
                    addNewItemButton( button );
                    UI.getCurrent()
                      .getPage()
                      .executeJs( Command.JS.FULL_SCREEN_FINAL );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( applicationGroupsUI );
                    applicationGroupsUI.removeAll();
                    applicationGroupsUI = null;
                    removeNewItemButton();
                }
            } );
            applicationGroupsButton.getElement()
                                   .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS ) );

            // server groups button
            serverGroupsButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS ), new Icon( VaadinIcon.GRID_BIG_O ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive )
                {
                    disableMenuItems();
                    serverGroupsMenuItem.setEnabled( true );
                    serverGroupsUI = new ServerGroupsUI( database, virtualUtil );
                    addComponent( serverGroupsUI );

                    Button button = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS_NEW ), new Icon( VaadinIcon.GRID_BIG_O ), e -> {
                        if ( Objects.nonNull( database ) && Objects.nonNull( serverGroupsUI ) )
                        {
                            serverGroupsUI.addNewServersGroup();
                            UI.getCurrent()
                              .getPage()
                              .executeJs( Command.JS.FULL_SCREEN_FINAL );
                        }
                        else
                        {
                            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SERVERS_GROUP ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                        }
                    } );
                    addNewItemButton( button );
                    UI.getCurrent()
                      .getPage()
                      .executeJs( Command.JS.FULL_SCREEN_FINAL );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( serverGroupsUI );
                    serverGroupsUI.removeAll();
                    serverGroupsUI = null;
                    removeNewItemButton();
                }
            } );
            serverGroupsButton.getElement()
                              .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS ) );

            // servers button
            serversButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ), new Icon( VaadinIcon.SERVER ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive )
                {
                    disableMenuItems();
                    serversMenuItem.setEnabled( true );
                    serversUI = new ServersUI( database, virtualUtil );
                    addComponent( serversUI );

                    Button button = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS_NEW ), new Icon( VaadinIcon.SERVER ), e -> {
                        if ( Objects.nonNull( database ) && Objects.nonNull( serversUI ) )
                        {
                            serversUI.addNewServer();
                            UI.getCurrent()
                              .getPage()
                              .executeJs( Command.JS.FULL_SCREEN_FINAL );
                        }
                        else
                        {
                            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SERVER ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                        }
                    } );
                    addNewItemButton( button );
                    UI.getCurrent()
                      .getPage()
                      .executeJs( Command.JS.FULL_SCREEN_FINAL );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( serversUI );
                    serversUI.removeAll();
                    serversUI = null;
                    removeNewItemButton();
                }
            } );
            serversButton.getElement()
                         .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ) );

            // users button
            usersButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS ), new Icon( VaadinIcon.USERS ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive )
                {
                    disableMenuItems();
                    usersMenuItem.setEnabled( true );
                    usersUI = new UsersUI( database, virtualUtil );
                    addComponent( usersUI );

                    Button button = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS_NEW ), new Icon( VaadinIcon.USERS ), e -> {
                        if ( Objects.nonNull( database ) && Objects.nonNull( usersUI ) )
                        {
                            usersUI.addNewUser();
                            UI.getCurrent()
                              .getPage()
                              .executeJs( Command.JS.FULL_SCREEN_FINAL );
                        }
                        else
                        {
                            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_USER ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                        }
                    } );
                    addNewItemButton( button );
                    UI.getCurrent()
                      .getPage()
                      .executeJs( Command.JS.FULL_SCREEN_FINAL );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( usersUI );
                    usersUI.removeAll();
                    usersUI = null;
                    removeNewItemButton();
                }
            } );
            usersButton.getElement()
                       .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS ) );

            // user sessions button
            userSessionsButton = new Button( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SESSION_MANAGEMENT ), new Icon( VaadinIcon.TIMER ), event -> {
                boolean buttonActive = setButtonActive( event.getSource() );
                if ( buttonActive )
                {
                    disableMenuItems();
                    userSessionsMenuItem.setEnabled( true );
                    userSessionsUI = new UserSessionsUI( database, virtualUtil );
                    addComponent( userSessionsUI );
                }
                else
                {
                    enableMenuItems();
                    removeComponent( userSessionsUI );
                    userSessionsUI.removeAll();
                    userSessionsUI = null;
                }
            } );
            userSessionsButton.getElement()
                              .setAttribute( ATTRIBUTE_APPLICATIONS, virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SESSION_MANAGEMENT ) );

            // settings: applications
            Icon applicationsImage = new Icon( VaadinIcon.FILE_O );
            applicationsImage.getStyle()
                             .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                             .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                             .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                             .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label applicationsLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATIONS ) );
            applicationsLabel.getStyle()
                             .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                             .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div applicationsDiv = new Div( applicationsImage, applicationsLabel );

            // settings: application groups
            Icon applicationGroupsImage = new Icon( VaadinIcon.COPY_O );
            applicationGroupsImage.getStyle()
                                  .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                                  .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                                  .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                                  .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label applicationGroupsLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_APPLICATION_GROUPS ) );
            applicationGroupsLabel.getStyle()
                                  .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                                  .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div userGroupsDiv = new Div( applicationGroupsImage, applicationGroupsLabel );

            // settings: imports data
            Icon importsDataImage = new Icon( VaadinIcon.UPLOAD_ALT );
            importsDataImage.getStyle()
                            .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                            .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                            .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                            .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label importsDataLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_IMPORT_DATA ) );
            importsDataLabel.getStyle()
                            .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                            .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div importsDataDiv = new Div( importsDataImage, importsDataLabel );

            // settings: log out
            Icon logOutImage = new Icon( VaadinIcon.SIGN_OUT );
            logOutImage.getStyle()
                       .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                       .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                       .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                       .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label logOutLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_LOG_OUT ) );
            logOutLabel.getStyle()
                       .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                       .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div logOutDiv = new Div( logOutImage, logOutLabel );

            // settings: servers
            Icon serversImage = new Icon( VaadinIcon.SERVER );
            serversImage.getStyle()
                        .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                        .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                        .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                        .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label serversLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVERS ) );
            serversLabel.getStyle()
                        .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                        .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div serversDiv = new Div( serversImage, serversLabel );

            // settings: server groups
            Icon serverGroupsImage = new Icon( VaadinIcon.GRID_BIG_O );
            serverGroupsImage.getStyle()
                             .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                             .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                             .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                             .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label serverGroupsLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SERVER_GROUPS ) );
            serverGroupsLabel.getStyle()
                             .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                             .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div serverGroupsDiv = new Div( serverGroupsImage, serverGroupsLabel );

            // settings: users
            Icon usersImage = new Icon( VaadinIcon.USERS );
            usersImage.getStyle()
                      .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                      .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                      .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                      .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label usersLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_USERS ) );
            usersLabel.getStyle()
                      .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                      .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div usersDiv = new Div( usersImage, usersLabel );

            // settings: user session
            Icon userSessionImage = new Icon( VaadinIcon.TIMER );
            userSessionImage.getStyle()
                            .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                            .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                            .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                            .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
            Label userSessionLabel = new Label( virtualUtil.getTranslation( Translation.BAR_NAVIGATION_SESSION_MANAGEMENT ) );
            userSessionLabel.getStyle()
                            .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE )
                            .set( STYLE_VERTICAL_ALIGN, STYLE_VERTICAL_ALIGN_VALUE );
            Div userSessionDiv = new Div( userSessionImage, userSessionLabel );

            // settings button
            Button settingsButton = new Button( null, new Icon( VaadinIcon.MENU ), event -> UI.getCurrent()
                                                                                              .getPage()
                                                                                              .executeJs( Command.JS.SHOW_MENU_BAR_ITEMS ) );
            settingsButton.getStyle()
                          .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_LEFT_VALUE );

            // settings: main menu
            ContextMenu settingsContextMenu = new ContextMenu( settingsButton );
            settingsContextMenu.setOpenOnClick( true );
            serversMenuItem = settingsContextMenu.addItem( serversDiv, event -> serversButton.click() );
            serverGroupsMenuItem = settingsContextMenu.addItem( serverGroupsDiv, event -> serverGroupsButton.click() );
            applicationsMenuItem = settingsContextMenu.addItem( applicationsDiv, event -> applicationsButton.click() );
            userGroupsMenuItem = settingsContextMenu.addItem( userGroupsDiv, event -> applicationGroupsButton.click() );
            usersMenuItem = settingsContextMenu.addItem( usersDiv, event -> usersButton.click() );
            userSessionsMenuItem = settingsContextMenu.addItem( userSessionDiv, event -> userSessionsButton.click() );
            settingsContextMenu.add( new Hr() );
            importsDataMenuItem = settingsContextMenu.addItem( importsDataDiv, event -> importData() );
            settingsContextMenu.add( new Hr() );
            logOutMenuItem = settingsContextMenu.addItem( logOutDiv, event -> logOutButton.click() );

            // add buttons
            add( settingsButton, userButton );
        }
        else
        {
            add( logOutButton, userButton );
        }

        if ( virtualUtil.isNotApplicationsBarVisible() )
        {
            applicationsContextMenu = new ContextMenu( userButton );
            applicationsContextMenu.setOpenOnClick( true );
            addApplications();
        }
    }

    public abstract void addComponent( Component component );

    public abstract void logOutUser();

    public abstract void removeComponent( Component component );

    public abstract void selectApplication( UserData userData, ApplicationData applicationData );

    public abstract void testingApplication( ApplicationData applicationData );

    private void disableMenuItems()
    {
        applicationsMenuItem.setEnabled( false );
        importsDataMenuItem.setEnabled( false );
        logOutMenuItem.setEnabled( false );
        serverGroupsMenuItem.setEnabled( false );
        serversMenuItem.setEnabled( false );
        userGroupsMenuItem.setEnabled( false );
        usersMenuItem.setEnabled( false );
        userSessionsMenuItem.setEnabled( false );
    }

    private void enableMenuItems()
    {
        applicationsMenuItem.setEnabled( true );
        importsDataMenuItem.setEnabled( true );
        logOutMenuItem.setEnabled( true );
        serverGroupsMenuItem.setEnabled( true );
        serversMenuItem.setEnabled( true );
        userGroupsMenuItem.setEnabled( true );
        usersMenuItem.setEnabled( true );
        userSessionsMenuItem.setEnabled( true );
    }

    private void addNewItemButton( Button button )
    {
        if ( Objects.nonNull( button ) )
        {
            removeNewItemButton();
            newItemButton = button;
            newItemButton.addClassName( CLASS_ACTIVE_BUTTON );
            add( newItemButton );
        }
    }

    private void removeNewItemButton()
    {
        if ( Objects.nonNull( newItemButton ) )
        {
            remove( newItemButton );
            newItemButton = null;
        }
    }

    private boolean setButtonActive( Button button )
    {
        boolean buttonActive = button.hasClassName( CLASS_ACTIVE_BUTTON );
        if ( !buttonActive )
        {
            button.setEnabled( true );
            button.addClassName( CLASS_ACTIVE_BUTTON );
            add( button );
        }
        else
        {
            button.removeClassName( CLASS_ACTIVE_BUTTON );
            remove( button );
        }
        return !buttonActive;
    }

    private void importData()
    {
        dataImportDialog = new DataImportDialog( virtualUtil )
        {
            private static final long serialVersionUID = -3515321580843859613L;

            @Override
            public void prepareImportingData()
            {
                getUI().ifPresent( ui -> ui.access( () -> {
                    dataImportWaitDialogUtil = new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_IMPORT_WAIT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
                    dataImportWaitDialogUtil.setButtonsVisible( false );
                    dataImportWaitDialogUtil.open();
                    ui.push();
                } ) );
            }

            @Override
            public void importingData( String folder, String fileName, String fileType )
            {
                if ( Objects.nonNull( database ) && Objects.nonNull( virtualUtil ) )
                {
                    ImportDataRunner importDataRunner = new ImportDataRunner( database, virtualUtil, folder, fileName, fileType )
                    {
                        @Override
                        public void dataImported( String message )
                        {
                            getUI().ifPresent( ui -> ui.access( () -> {
                                dialogInfoUtil = new DialogInfoUtil( message, virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
                                dialogInfoUtil.setButtonsVisible( true );
                                dialogInfoUtil.open();
                                close();
                                if ( Objects.nonNull( dataImportWaitDialogUtil ) )
                                {
                                    dataImportWaitDialogUtil.close();
                                }
                                ui.push();
                            } ) );
                        }
                    };
                    virtualUtil.startImportDataService( importDataRunner );
                }
            }

            @Override
            public void importingDataError( String errorMessage )
            {
                getUI().ifPresent( ui -> ui.access( () -> {
                    dialogInfoUtil = new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_IMPORT, errorMessage ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
                    dialogInfoUtil.setButtonsVisible( true );
                    dialogInfoUtil.open();
                    close();
                    if ( Objects.nonNull( dataImportWaitDialogUtil ) )
                    {
                        dataImportWaitDialogUtil.close();
                    }
                    ui.push();
                } ) );
                close();
            }
        };
        dataImportDialog.open();
    }

    private void addApplications()
    {
        if ( Objects.nonNull( applicationsContextMenu ) && Objects.nonNull( userData ) )
        {
            List<ApplicationData> applicationDatas = new ArrayList<>();
            database.getUserApplications( userData )
                    .stream()
                    .filter( Objects::nonNull )
                    .forEach( userApplicationData -> applicationDatas.add( userApplicationData.getApplicationData() ) );
            database.getUserApplicationGroups( userData )
                    .stream()
                    .filter( Objects::nonNull )
                    .forEach( userApplicationsGroupData -> applicationDatas.addAll( userApplicationsGroupData.getApplicationsGroupData()
                                                                                                             .getApplicationDatas() ) );
            applicationDatas.stream()
                            .distinct()
                            .sorted( Comparator.comparing( ApplicationData::getName ) )
                            .forEach( applicationData -> {
                                HorizontalLayout horizontalLayout = new HorizontalLayout();
                                horizontalLayout.setAlignItems( Alignment.CENTER );
                                horizontalLayout.setMargin( false );
                                horizontalLayout.setPadding( false );
                                horizontalLayout.setSpacing( false );
                                horizontalLayout.setSizeUndefined();

                                IconData iconData = applicationData.getIconData();
                                if ( Objects.nonNull( iconData ) )
                                {
                                    if ( Objects.isNull( iconData.getIcon() ) )
                                    {
                                        String iconName = iconData.getName();
                                        IronIcon ironIcon = new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                                                          iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
                                        ironIcon.getStyle()
                                                .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                                                .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                                                .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                                                .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
                                        horizontalLayout.add( ironIcon );
                                    }
                                    else
                                    {
                                        StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                                        Image image = new Image( streamResource, iconData.getName() );
                                        image.getStyle()
                                             .set( STYLE_WIDTH, STYLE_WIDTH_VALUE )
                                             .set( STYLE_HEIGHT, STYLE_HEIGHT_VALUE )
                                             .set( STYLE_PADDING, STYLE_PADDING_VALUE )
                                             .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_VALUE );
                                        horizontalLayout.add( image );
                                    }
                                }
                                VerticalLayout verticalLayout = new VerticalLayout();
                                verticalLayout.setMargin( false );
                                verticalLayout.setPadding( false );
                                verticalLayout.setSpacing( false );
                                verticalLayout.setSizeUndefined();
                                verticalLayout.getStyle()
                                              .set( STYLE_PADDING_LEFT, fullScreen ? STYLE_PADDING_LEFT_28PX_VALUE : STYLE_PADDING_LEFT_0PX_VALUE )
                                              .set( STYLE_PADDING_RIGHT, STYLE_PADDING_RIGHT_CALC_VALUE );
                                verticalLayout.add( new Label( applicationData.getName() ) );
                                if ( StringUtils.isNotBlank( applicationData.getNote() ) )
                                {
                                    Label label = new Label( applicationData.getNote() );
                                    label.getStyle()
                                         .set( STYLE_FONT_SIZE, STYLE_FONT_SIZE_VALUE );
                                    verticalLayout.add( label );
                                }
                                horizontalLayout.add( verticalLayout );
                                applicationsContextMenu.addItem( horizontalLayout, event -> selectApplication( userData, applicationData ) );
                            } );
        }
    }

    private void removeApplications()
    {
        if ( Objects.nonNull( applicationsContextMenu ) )
        {
            applicationsContextMenu.removeAll();
        }
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        if ( Objects.nonNull( dataImportWaitDialogUtil ) )
        {
            dataImportWaitDialogUtil.close();
        }
        if ( Objects.nonNull( dialogInfoUtil ) )
        {
            dialogInfoUtil.close();
        }
        if ( Objects.nonNull( dataImportDialog ) )
        {
            dataImportDialog.close();
        }
    }

    public void setUserButtonEnabled( boolean enabled )
    {
        if ( virtualUtil.isNotApplicationsBarVisible() )
        {
            userButton.setEnabled( enabled );
        }
    }

    public double getNavigationBarHeight()
    {
        return ( virtualUtil.isApplicationsBarVisible() ? 2 * navigationBarHeight : navigationBarHeight );
    }

    public void setFullScreen( boolean fullScreen )
    {
        navigationBarHeight = 0;
        if ( fullScreen )
        {
            getElement().executeJs( "return $0.clientHeight;", getElement() )
                        .then( height -> navigationBarHeight = height.asNumber() );
        }
        this.fullScreen = fullScreen;
        refresh();
    }

    public void refresh()
    {
        removeApplications();
        addApplications();
    }

    public void applicationsButtonClick()
    {
        if ( Objects.nonNull( applicationsButton ) )
        {
            setButtonActive( applicationsButton );
            applicationsButton.click();
        }
    }

    public void applicationSelect( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationsUI ) && Objects.nonNull( applicationData ) )
        {
            applicationsUI.selectApplication( applicationData );
        }
    }

    public void setupApplications()
    {
        if ( Objects.nonNull( applicationsButton ) )
        {
            applicationsButton.click();
            newItemButton.click();
        }
    }

    public void setupServerGroups()
    {
        if ( Objects.nonNull( serverGroupsButton ) )
        {
            serverGroupsButton.click();
            newItemButton.click();
        }
    }

    public void setupServers()
    {
        if ( Objects.nonNull( serversButton ) )
        {
            serversButton.click();
            newItemButton.click();
        }
    }

    public void setupUsers()
    {
        if ( Objects.nonNull( usersButton ) )
        {
            usersButton.click();
            newItemButton.click();
        }
    }

}
