/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.options;

import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.icons.TypIcon;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;

public abstract class OptionsBar
    extends HorizontalLayout
{

    private static final long serialVersionUID = -76260764529370727L;

    private static final String CLASS_BUTTON_UNLOCKED = "unlocked";

    private static final String CLASS_FULL_SCREEN = "fullscreen";

    private static final String CLASS_OPTIONS_BAR = "options-bar";

    private boolean fullScreen = false;

    private Button arrangeWindowsButton = null;

    private Button downloadButton = null;

    private Button fullScreenButton = null;

    private Button screenshotButton = null;

    private Button uploadButton = null;

    protected OptionsBar( VirtualUtil virtualUtil )
    {
        setMargin( false );
        setPadding( false );
        setSpacing( true );
        setSizeUndefined();
        addClassName( CLASS_OPTIONS_BAR );

        Icon arrangeWindowsdButtonIcon = TypIcon.TH_LARGE_OUTLINE.create();
        arrangeWindowsButton = new Button( null, arrangeWindowsdButtonIcon, event -> arrangeWindows() );
        arrangeWindowsButton.addClassName( CLASS_BUTTON_UNLOCKED );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( arrangeWindowsButton, virtualUtil.getTranslation( Translation.BUTTON_WINDOWS_ARRANGE ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon downloadButtonIcon = TypIcon.DOWNLOAD_OUTLINE.create();
        downloadButton = new Button( null, downloadButtonIcon, event -> downloadFile() );
        downloadButton.addClassName( CLASS_BUTTON_UNLOCKED );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( downloadButton, virtualUtil.getTranslation( Translation.BUTTON_DOWNLOAD ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon fullScreenButtonIcon = TypIcon.ARROW_MAXIMISE.create();
        fullScreenButton = new Button( null, fullScreenButtonIcon, event -> {
            if ( fullScreenButton.hasClassName( CLASS_BUTTON_UNLOCKED ) )
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.BAR_OPTIONS_SCREEN ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        } );
        fullScreenButton.addClassName( CLASS_FULL_SCREEN );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( fullScreenButton, virtualUtil.getTranslation( Translation.BUTTON_FULL_SCREEN ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon screenshotButtonIcon = TypIcon.CAMERA_OUTLINE.create();
        screenshotButton = new Button( null, screenshotButtonIcon, event -> {
            if ( isScreenshotEnabled() )
            {
                takeScreenshot();
            }
        } );
        screenshotButton.addClassName( CLASS_BUTTON_UNLOCKED );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( screenshotButton, virtualUtil.getTranslation( Translation.BUTTON_SCREENSHOT ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon uploadButtonIcon = TypIcon.UPLOAD_OUTLINE.create();
        uploadButton = new Button( null, uploadButtonIcon, event -> uploadFile() );
        uploadButton.addClassName( CLASS_BUTTON_UNLOCKED );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( uploadButton, virtualUtil.getTranslation( Translation.BUTTON_UPLOAD_FILE ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        add( arrangeWindowsButton, screenshotButton, uploadButton, downloadButton, fullScreenButton );
    }

    public abstract void arrangeWindows();

    public abstract void downloadFile();

    public abstract void takeScreenshot();

    public abstract void uploadFile();

    public boolean isFullScreen()
    {
        return fullScreen;
    }

    public void setFullScreen( boolean fullScreen )
    {
        this.fullScreen = !fullScreenButton.hasClassName( CLASS_BUTTON_UNLOCKED ) && fullScreen;
        fullScreenButton.setIcon( fullScreen ? TypIcon.ARROW_MINIMISE.create() : TypIcon.ARROW_MAXIMISE.create() );
    }

    public void setFullScreenStatus( boolean fullScreenStatus )
    {
        if ( !fullScreenStatus )
        {
            fullScreenButton.addClassName( CLASS_BUTTON_UNLOCKED );
        }
    }

    public void setFullScreenButtonVisible( boolean fullScreenButtonVisible )
    {
        fullScreenButton.setVisible( fullScreenButtonVisible );
    }

    public boolean isArrangeWindowsButtonEnabled()
    {
        return !arrangeWindowsButton.hasClassName( CLASS_BUTTON_UNLOCKED ) && arrangeWindowsButton.isVisible();
    }

    public void setArrangeWindowsButtonEnabled( boolean arrangeWindowsButtonEnabled )
    {
        if ( arrangeWindowsButtonEnabled && arrangeWindowsButton.isVisible() )
        {
            arrangeWindowsButton.removeClassName( CLASS_BUTTON_UNLOCKED );
        }
        else
        {
            arrangeWindowsButton.addClassName( CLASS_BUTTON_UNLOCKED );
        }
    }

    public boolean isDownloadButtonEnabled()
    {
        return !downloadButton.hasClassName( CLASS_BUTTON_UNLOCKED ) && downloadButton.isVisible();
    }

    public void setDownloadButtonEnabled( boolean downloadButtonEnabled )
    {
        if ( downloadButtonEnabled && downloadButton.isVisible() )
        {
            downloadButton.removeClassName( CLASS_BUTTON_UNLOCKED );
        }
        else
        {
            downloadButton.addClassName( CLASS_BUTTON_UNLOCKED );
        }
    }

    public void setDownloadButtonVisible( boolean downloadButtonVisible )
    {
        downloadButton.setVisible( downloadButtonVisible );
    }

    public boolean isUploadButtonEnabled()
    {
        return !uploadButton.hasClassName( CLASS_BUTTON_UNLOCKED ) && uploadButton.isVisible();
    }

    public void setUploadButtonEnabled( boolean uploadButtonEnabled )
    {
        if ( uploadButtonEnabled && uploadButton.isVisible() )
        {
            uploadButton.removeClassName( CLASS_BUTTON_UNLOCKED );
        }
        else
        {
            uploadButton.addClassName( CLASS_BUTTON_UNLOCKED );
        }
    }

    public void setUploadButtonVisible( boolean uploadButtonVisible )
    {
        uploadButton.setVisible( uploadButtonVisible );
    }

    public boolean isScreenshotEnabled()
    {
        return !screenshotButton.hasClassName( CLASS_BUTTON_UNLOCKED );
    }

    public void setScreenshotEnabled( boolean screenshotEnabled )
    {
        if ( screenshotEnabled )
        {
            screenshotButton.removeClassName( CLASS_BUTTON_UNLOCKED );
        }
        else
        {
            screenshotButton.addClassName( CLASS_BUTTON_UNLOCKED );
        }
    }

    public void setScreenshotButtonVisible( boolean screenshotButtonVisible )
    {
        screenshotButton.setVisible( screenshotButtonVisible );
    }

}
