/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.events;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.info.KeyboardCharName;
import org.dussan.vaadin.dvirtual.info.KeyboardCharcodeName;
import org.dussan.vaadin.dvirtual.info.KeyboardKeyName;
import org.dussan.vaadin.dvirtual.info.KeyboardNumpadName;
import org.dussan.vaadin.dvirtual.ui.events.data.KeyboardEventData;

import elemental.json.JsonObject;

public class KeyboardEventProcessor
    implements Serializable
{

    private static final long serialVersionUID = -1268904817614199625L;

    public static final String KEYBOARD_ALT_KEY = "event.altKey";

    public static final String KEYBOARD_CTRL_KEY = "event.ctrlKey";

    public static final String KEYBOARD_KEYDOWN = "keydown";

    public static final String KEYBOARD_KEYUP = "keyup";

    public static final String KEYBOARD_IS_COMPOSING = "event.isComposing";

    public static final String KEYBOARD_KEY_CODE = "event.keyCode";

    public static final String KEYBOARD_KEY_NAME = "event.code";

    public static final String KEYBOARD_KEY_STRING = "event.key";

    public static final String KEYBOARD_LOCATION = "event.location";

    public static final String KEYBOARD_META_KEY = "event.metaKey";

    public static final String KEYBOARD_SHIFT_KEY = "event.shiftKey";

    public static final String KEYBOARD_WHICH = "event.which";

    private static final int DOM_KEY_LOCATION_RIGHT = 0x02;

    private static final int KEY_DOWN_IME = 229;

    private static final int KEY_DOWN_CAPS_LOCK = 20;

    private static final int KEY_DOWN_NUM_LOCK = 144;

    private static final int UNDEFINED_VALUE = 0;

    private static final String ISO_LEVEL3_SHIFT = "ISO_Level3_Shift";

    private static final String MODIFIER_ALT = "Alt";

    private static final String MODIFIER_ALT_R = "Alt_R";

    private static final String MODIFIER_ALTGR = "AltGr";

    private static final String MODIFIER_ALTGRAPH = "AltGraph";

    private static final String MODIFIER_CAPS_LOCK = "CapsLock";

    private static final String MODIFIER_CONTROL = "Control";

    private static final String MODIFIER_CONTROL_L = "Control_L";

    private static final String MODIFIER_CONTROL_R = "Control_R";

    private static final String MODIFIER_LOCK = "Lock";

    private static final String MODIFIER_META = "Meta";

    private static final String MODIFIER_META_L = "Meta_L";

    private static final String MODIFIER_META_R = "Meta_R";

    private static final String MODIFIER_NUM_LOCK = "NumLock";

    private static final String MODIFIER_SHIFT = "Shift";

    private static final String MODIFIERS_SPLITTER = ",";

    private Boolean capsLock = null;

    private Boolean numLock = null;

    private KeyboardCharcodeName keyboardCharcodeName = null;

    private KeyboardCharName keyboardCharName = null;

    private KeyboardKeyName keyboardKeyName = null;

    private KeyboardNumpadName keyboardNumpadName = null;

    public KeyboardEventProcessor()
    {
        keyboardCharcodeName = new KeyboardCharcodeName();
        keyboardCharName = new KeyboardCharName();
        keyboardKeyName = new KeyboardKeyName();
        keyboardNumpadName = new KeyboardNumpadName();
    }

    private List<String> eventModifiers( JsonObject eventData )
    {
        List<String> modifiers = new ArrayList<>();
        boolean altKey = false;
        boolean ctrlKey = false;
        boolean metaKey = false;
        boolean shiftKey = false;

        if ( Objects.nonNull( eventData ) )
        {
            altKey = eventData.hasKey( KEYBOARD_ALT_KEY ) && eventData.getBoolean( KEYBOARD_ALT_KEY );
            ctrlKey = eventData.hasKey( KEYBOARD_CTRL_KEY ) && eventData.getBoolean( KEYBOARD_CTRL_KEY );
            metaKey = eventData.hasKey( KEYBOARD_META_KEY ) && eventData.getBoolean( KEYBOARD_META_KEY );
            shiftKey = eventData.hasKey( KEYBOARD_SHIFT_KEY ) && eventData.getBoolean( KEYBOARD_SHIFT_KEY );
        }

        if ( altKey )
        {
            modifiers.add( MODIFIER_ALT.toLowerCase() );
        }

        if ( ctrlKey )
        {
            modifiers.add( MODIFIER_CONTROL.toLowerCase() );
        }

        if ( metaKey )
        {
            modifiers.add( MODIFIER_META.toLowerCase() );
        }

        if ( shiftKey )
        {
            modifiers.add( MODIFIER_SHIFT.toLowerCase() );
        }

        if ( Objects.nonNull( capsLock ) && capsLock.booleanValue() )
        {
            modifiers.add( MODIFIER_CAPS_LOCK.toLowerCase() );
        }

        if ( Objects.nonNull( numLock ) && numLock.booleanValue() )
        {
            modifiers.add( MODIFIER_NUM_LOCK.toLowerCase() );
        }

        return modifiers;
    }

    private List<String> keyboardModifiers( JsonObject eventData, Map<String, String> modifiersMap, boolean altGrPressed, boolean swapKeys )
    {
        List<String> modifiers = new ArrayList<>();
        String alt = MapUtils.getString( modifiersMap, MODIFIER_ALT );
        String altGr = MapUtils.getString( modifiersMap, MODIFIER_ALTGR );
        String ctrl = MapUtils.getString( modifiersMap, MODIFIER_CONTROL, StringUtils.lowerCase( MODIFIER_CONTROL ) );
        String meta = MapUtils.getString( modifiersMap, MODIFIER_META );
        boolean altKey = false;
        boolean ctrlKey = false;
        boolean metaKey = false;
        boolean shiftKey = false;

        if ( Objects.nonNull( eventData ) )
        {
            altKey = eventData.hasKey( KEYBOARD_ALT_KEY ) && eventData.getBoolean( KEYBOARD_ALT_KEY );
            ctrlKey = eventData.hasKey( KEYBOARD_CTRL_KEY ) && eventData.getBoolean( KEYBOARD_CTRL_KEY );
            metaKey = eventData.hasKey( KEYBOARD_META_KEY ) && eventData.getBoolean( KEYBOARD_META_KEY );
            shiftKey = eventData.hasKey( KEYBOARD_SHIFT_KEY ) && eventData.getBoolean( KEYBOARD_SHIFT_KEY );
        }

        if ( swapKeys )
        {
            ctrl = MapUtils.getString( modifiersMap, KEYBOARD_META_KEY );
            meta = MapUtils.getString( modifiersMap, KEYBOARD_CTRL_KEY );
        }

        if ( altKey && StringUtils.isNotBlank( alt ) )
        {
            modifiers.add( alt );
        }

        if ( ctrlKey && StringUtils.isNotBlank( ctrl ) )
        {
            modifiers.add( ctrl );
        }

        if ( metaKey && StringUtils.isNotBlank( meta ) )
        {
            modifiers.add( meta );
        }

        if ( shiftKey )
        {
            modifiers.add( StringUtils.lowerCase( MODIFIER_SHIFT ) );
        }

        if ( Objects.nonNull( numLock ) && numLock.booleanValue() && StringUtils.isNotBlank( MapUtils.getString( modifiersMap, MODIFIER_NUM_LOCK ) ) )
        {
            modifiers.add( MapUtils.getString( modifiersMap, MODIFIER_NUM_LOCK ) );
        }

        if ( Objects.nonNull( capsLock ) && capsLock.booleanValue() )
        {
            modifiers.add( MODIFIER_LOCK.toLowerCase() );
        }

        if ( altGrPressed && StringUtils.isNotBlank( altGr ) )
        {
            modifiers.add( altGr );
            modifiers.remove( alt );
            modifiers.remove( ctrl );
        }

        return modifiers;
    }

    public KeyboardEventData process( JsonObject eventData, Map<String, String> modifiersMap, boolean keyPressed, boolean windowsOS, boolean macOS )
    {
        boolean swapKeys = macOS;
        boolean isComposing = false;
        int keyCode = UNDEFINED_VALUE;
        int location = UNDEFINED_VALUE;
        String keyName = StringUtils.EMPTY;
        String keyString = StringUtils.EMPTY;
        if ( Objects.nonNull( eventData ) )
        {
            isComposing = eventData.hasKey( KEYBOARD_IS_COMPOSING ) && eventData.getBoolean( KEYBOARD_IS_COMPOSING );
            keyCode = eventData.hasKey( KEYBOARD_WHICH ) ? (int) eventData.getNumber( KEYBOARD_WHICH ) : keyCode;
            keyCode = keyCode == UNDEFINED_VALUE && eventData.hasKey( KEYBOARD_KEY_CODE ) ? (int) eventData.getNumber( KEYBOARD_KEY_CODE ) : keyCode;
            location = eventData.hasKey( KEYBOARD_LOCATION ) ? (int) eventData.getNumber( KEYBOARD_LOCATION ) : location;
            keyName = eventData.hasKey( KEYBOARD_KEY_NAME ) ? StringUtils.defaultIfBlank( eventData.getString( KEYBOARD_KEY_NAME ), StringUtils.EMPTY ) : keyName;
            keyString = eventData.hasKey( KEYBOARD_KEY_STRING ) ? StringUtils.defaultIfBlank( eventData.getString( KEYBOARD_KEY_STRING ), StringUtils.EMPTY ) : keyString;
        }

        // key code relating to an event that has been processed by an IME
        // https://developer.mozilla.org/en-US/docs/Web/Events/keydown
        if ( isComposing || keyCode == KEY_DOWN_IME )
        {
            return null;
        }

        // capsLock pressed
        if ( keyPressed && Objects.nonNull( capsLock ) && keyCode == KEY_DOWN_CAPS_LOCK )
        {
            capsLock = !capsLock;
        }

        // numLock pressed
        if ( keyPressed && Objects.nonNull( numLock ) && keyCode == KEY_DOWN_NUM_LOCK )
        {
            numLock = !numLock;
        }

        if ( !StringUtils.equals( keyName, keyString ) && keyboardNumpadName.contains( keyString ) )
        {
            keyName = keyboardNumpadName.get( keyString );
        }
        else if ( keyboardKeyName.contains( keyName ) )
        {
            keyName = keyboardKeyName.get( keyName );
        }
        else if ( keyboardCharName.contains( keyString ) )
        {
            keyName = keyboardCharName.get( keyString );
        }
        else if ( keyboardCharcodeName.contains( keyCode ) )
        {
            keyName = keyboardCharcodeName.get( keyCode );
        }

        if ( keyName.matches( "_L$" ) && location == DOM_KEY_LOCATION_RIGHT )
            keyName = StringUtils.replace( keyName, "_L", "_R" );

        // check if AltGr pressed
        boolean altGrPressed = false;
        if ( StringUtils.equals( keyString, MODIFIER_ALTGRAPH ) || ( StringUtils.equals( keyName, MODIFIER_ALT_R ) && ( macOS || windowsOS ) ) )
        {
            altGrPressed = keyPressed;
            keyName = ISO_LEVEL3_SHIFT;
            keyString = MODIFIER_ALTGRAPH;
        }

        String eventModifiers = StringUtils.join( eventModifiers( eventData ), MODIFIERS_SPLITTER );
        List<String> modifiers = keyboardModifiers( eventData, modifiersMap, altGrPressed, swapKeys );
        int keyValue = keyCode;

        boolean shiftPressed = CollectionUtils.containsAny( modifiers, MODIFIER_SHIFT );
        boolean capslockPressed = CollectionUtils.containsAny( modifiers, MODIFIER_CAPS_LOCK );
        if ( ( capslockPressed && shiftPressed ) || ( !capslockPressed && !shiftPressed ) )
        {
            keyString = StringUtils.lowerCase( keyString );
        }

        String oldKeyString = keyString;
        if ( swapKeys )
        {
            if ( StringUtils.equals( keyName, MODIFIER_CONTROL_L ) )
            {
                keyName = MODIFIER_META_L;
                keyString = MODIFIER_META.toLowerCase();
            }
            else if ( StringUtils.equals( keyName, MODIFIER_META_L ) )
            {
                keyName = MODIFIER_CONTROL_L;
                keyString = MODIFIER_CONTROL.toLowerCase();
            }
            else if ( StringUtils.equals( keyName, MODIFIER_CONTROL_R ) )
            {
                keyName = MODIFIER_META_R;
                keyString = MODIFIER_META.toLowerCase();
            }
            else if ( StringUtils.equals( keyName, MODIFIER_META_R ) )
            {
                keyName = MODIFIER_CONTROL_R;
                keyString = MODIFIER_CONTROL.toLowerCase();
            }
        }

        if ( keyPressed && macOS && StringUtils.containsIgnoreCase( eventModifiers, MODIFIER_META ) && !StringUtils.equals( oldKeyString, MODIFIER_META ) )
        {
            return ( new KeyboardEventData( keyCode, keyValue, keyName, keyString, modifiers, keyPressed, false ) );
        }
        else
        {
            return ( new KeyboardEventData( keyCode, keyValue, keyName, keyString, modifiers, keyPressed, null ) );
        }
    }

    public Boolean getCapsLock()
    {
        return capsLock;
    }

    public void setCapsLock( Boolean capsLock )
    {
        this.capsLock = capsLock;
    }

    public Boolean getNumLock()
    {
        return numLock;
    }

    public void setNumLock( Boolean numLock )
    {
        this.numLock = numLock;
    }

}
