/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.events;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.VirtualClient;
import org.dussan.vaadin.dvirtual.ui.events.data.MouseClickData;
import org.dussan.vaadin.dvirtual.ui.events.data.MouseMoveData;
import org.dussan.vaadin.dvirtual.ui.events.data.MouseWheelData;
import org.dussan.vaadin.dvirtual.xpra.window.Window;

import elemental.json.JsonObject;

public class MouseEventProcessor
    implements Serializable
{

    private static final long serialVersionUID = -8283909745649001342L;

    public static final String ELEMENT_ID = "element.id";

    public static final String KEYBOARD_ALT_KEY = "event.altKey";

    public static final String KEYBOARD_CTRL_KEY = "event.ctrlKey";

    public static final String KEYBOARD_META_KEY = "event.metaKey";

    public static final String KEYBOARD_SHIFT_KEY = "event.shiftKey";

    public static final String KEYBOARD_WHICH = "event.which";

    public static final String MOUSE_AXIS = "event.axis";

    public static final String MOUSE_BUTTON = "event.button";

    public static final String MOUSE_CONTEXTMENU = "contextmenu";

    public static final String MOUSE_DELTA_MODE = "event.deltaMode";

    public static final String MOUSE_DELTA_X = "event.deltaX";

    public static final String MOUSE_DELTA_Y = "event.deltaY";

    public static final String MOUSE_DETAIL = "event.detail";

    public static final String MOUSE_DOWN = "mousedown";

    public static final String MOUSE_MOVE = "mousemove";

    public static final String MOUSE_PAGE_X = "event.pageX";

    public static final String MOUSE_PAGE_Y = "event.pageY";

    public static final String MOUSE_UP = "mouseup";

    public static final String MOUSE_WHEEL_DELTA = "event.wheelDelta";

    public static final String MOUSE_WHEEL_DELTA_X = "event.wheelDeltaX";

    public static final String MOUSE_WHEEL_DELTA_Y = "event.wheelDeltaY";

    private static final int HORIZONTAL_AXIS = 0x01;

    private static final int MOUSE_WHEEL_DELTA_VALUE = 120;

    private static final String MODIFIER_ALT = "Alt";

    private static final String MODIFIER_CONTROL = "Control";

    private static final String MODIFIER_LOCK = "Lock";

    private static final String MODIFIER_META = "Meta";

    private static final String MODIFIER_NUM_LOCK = "NumLock";

    private static final String MODIFIER_SHIFT = "Shift";

    private Boolean capsLock = null;

    private Boolean numLock = null;

    private String mouseWheelEventId = null;

    public MouseEventProcessor()
    {
        // nothing to do
    }

    public class NormalizedMouseWheelData
    {

        private double deltaMode = 0;

        private double pixelX = 0;

        private double pixelY = 0;

        private double spinX = 0;

        private double spinY = 0;

        public double getSpinX()
        {
            return spinX;
        }

        public void setSpinX( double spinX )
        {
            this.spinX = spinX;
        }

        public double getSpinY()
        {
            return spinY;
        }

        public void setSpinY( double spinY )
        {
            this.spinY = spinY;
        }

        public double getPixelX()
        {
            return pixelX;
        }

        public void setPixelX( double pixelX )
        {
            this.pixelX = pixelX;
        }

        public double getPixelY()
        {
            return pixelY;
        }

        public void setPixelY( double pixelY )
        {
            this.pixelY = pixelY;
        }

        public double getDeltaMode()
        {
            return deltaMode;
        }

        public void setDeltaMode( double deltaMode )
        {
            this.deltaMode = deltaMode;
        }

    }

    private int eventButton( JsonObject eventData )
    {
        int button = 0;
        if ( eventData.hasKey( KEYBOARD_WHICH ) )
        {
            // Gecko (Firefox), WebKit (Safari/Chrome) & Opera
            button = Math.max( 0, (int) eventData.getNumber( KEYBOARD_WHICH ) );
        }
        else if ( eventData.hasKey( MOUSE_BUTTON ) )
        {
            // IE, Opera (zero based)
            button = Math.max( 0, (int) eventData.getNumber( MOUSE_BUTTON ) ) + 1;
        }
        if ( button == 4 )
        {
            button = 8;
        }
        else if ( button == 5 )
        {
            button = 9;
        }
        return button;
    }

    private List<String> keyboardModifiers( JsonObject eventData, Map<String, String> modifiersMap, boolean swapKeys )
    {
        List<String> modifiers = new ArrayList<>();
        String alt = MapUtils.getString( modifiersMap, MODIFIER_ALT );
        String ctrl = MapUtils.getString( modifiersMap, MODIFIER_CONTROL, StringUtils.lowerCase( MODIFIER_CONTROL ) );
        String meta = MapUtils.getString( modifiersMap, MODIFIER_META );
        boolean altKey = eventData.getBoolean( KEYBOARD_ALT_KEY );
        boolean ctrlKey = eventData.getBoolean( KEYBOARD_CTRL_KEY );
        boolean metaKey = eventData.getBoolean( KEYBOARD_META_KEY );
        boolean shiftKey = eventData.getBoolean( KEYBOARD_SHIFT_KEY );

        if ( swapKeys )
        {
            ctrl = MapUtils.getString( modifiersMap, KEYBOARD_META_KEY );
            meta = MapUtils.getString( modifiersMap, KEYBOARD_CTRL_KEY );
        }

        if ( altKey && StringUtils.isNotBlank( alt ) )
        {
            modifiers.add( alt );
        }

        if ( ctrlKey && StringUtils.isNotBlank( ctrl ) )
        {
            modifiers.add( ctrl );
        }

        if ( metaKey && StringUtils.isNotBlank( meta ) )
        {
            modifiers.add( meta );
        }

        if ( shiftKey )
        {
            modifiers.add( StringUtils.lowerCase( MODIFIER_SHIFT ) );
        }

        if ( Objects.nonNull( numLock ) && numLock.booleanValue() && StringUtils.isNotBlank( MapUtils.getString( modifiersMap, MODIFIER_NUM_LOCK ) ) )
        {
            modifiers.add( MapUtils.getString( modifiersMap, MODIFIER_NUM_LOCK ) );
        }

        if ( Objects.nonNull( capsLock ) && capsLock.booleanValue() )
        {
            modifiers.add( MODIFIER_LOCK.toLowerCase() );
        }

        return modifiers;
    }

    private NormalizedMouseWheelData normalizeMouseWheel( JsonObject eventData )
    {
        // defaults
        int lineHeight = 40;
        int pageHeight = 800;
        int pixelStep = 10;
        double sX = 0;
        double sY = 0;
        double pX = 0;
        double pY = 0;

        // legacy
        if ( eventData.hasKey( MOUSE_DETAIL ) )
        {
            sY = eventData.getNumber( MOUSE_DETAIL );
        }
        if ( eventData.hasKey( MOUSE_WHEEL_DELTA ) )
        {
            sY = -eventData.getNumber( MOUSE_WHEEL_DELTA ) / MOUSE_WHEEL_DELTA_VALUE;
        }
        if ( eventData.hasKey( MOUSE_WHEEL_DELTA_Y ) )
        {
            sY = -eventData.getNumber( MOUSE_WHEEL_DELTA_Y ) / MOUSE_WHEEL_DELTA_VALUE;
        }
        if ( eventData.hasKey( MOUSE_WHEEL_DELTA_X ) )
        {
            sX = -eventData.getNumber( MOUSE_WHEEL_DELTA_X ) / MOUSE_WHEEL_DELTA_VALUE;
        }

        // side scrolling on FF with DOMMouseScroll
        if ( eventData.hasKey( MOUSE_AXIS ) && ( (int) eventData.getNumber( MOUSE_WHEEL_DELTA_X ) ) == HORIZONTAL_AXIS )
        {
            sX = sY;
            sY = 0;
        }

        pX = sX * pixelStep;
        pY = sY * pixelStep;
        if ( eventData.hasKey( MOUSE_DELTA_X ) )
        {
            pX = eventData.getNumber( MOUSE_DELTA_X );
        }
        if ( eventData.hasKey( MOUSE_DELTA_Y ) )
        {
            pY = eventData.getNumber( MOUSE_DELTA_Y );
        }

        if ( ( Math.abs( pX ) != 0 || Math.abs( pY ) != 0 ) && eventData.hasKey( MOUSE_DELTA_MODE ) && eventData.getNumber( MOUSE_DELTA_MODE ) != 0 )
        {
            if ( eventData.getNumber( MOUSE_DELTA_MODE ) == 1 )
            {
                // delta in LINE units
                pX *= lineHeight;
                pY *= lineHeight;
            }
            else
            {
                // delta in PAGE units
                pX *= pageHeight;
                pY *= pageHeight;
            }
        }

        // fall-back if spin cannot be determined
        if ( Math.abs( pX ) != 0 && Math.abs( sX ) == 0 )
        {
            sX = pX < 1 ? -1 : 1;
        }
        if ( Math.abs( pY ) != 0 && Math.abs( sY ) == 0 )
        {
            sY = pY < 1 ? -1 : 1;
        }

        NormalizedMouseWheelData normalizedMouseWheel = new NormalizedMouseWheelData();
        normalizedMouseWheel.setDeltaMode( eventData.hasKey( MOUSE_DELTA_MODE ) ? eventData.getNumber( MOUSE_DELTA_MODE ) : 0 );
        normalizedMouseWheel.setPixelX( pX );
        normalizedMouseWheel.setPixelY( pY );
        normalizedMouseWheel.setSpinX( sX );
        normalizedMouseWheel.setSpinY( sY );
        return normalizedMouseWheel;
    }

    public String getMouseWheelEventId()
    {
        return mouseWheelEventId;
    }

    public void setMouseWheelEventId( String mouseWheelEventId )
    {
        this.mouseWheelEventId = mouseWheelEventId;
    }

    public MouseMoveData processMouseMove( VirtualClient virtualClient, JsonObject eventData, boolean macOS )
    {
        MouseMoveData mouseMoveData = null;
        String id = StringUtils.defaultIfBlank( eventData.getString( ELEMENT_ID ), StringUtils.EMPTY );
        if ( StringUtils.isNotBlank( id ) && StringUtils.isNumeric( id ) )
        {
            Window window = virtualClient.getWindow( Integer.parseInt( id ) );
            if ( Objects.nonNull( window ) )
            {
                int pageX = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_X ) ) ? (int) eventData.getNumber( MOUSE_PAGE_X ) : 0;
                int pageY = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_Y ) ) ? (int) eventData.getNumber( MOUSE_PAGE_Y ) : 0;
                int left = (int) ( pageX - window.getOffsetLeft() - window.getBorderWidth() - window.getDx() );
                int top = (int) ( pageY - window.getOffsetTop() - window.getBorderWidth() - window.getDy() - window.getTitleHeight() );
                Map<String, String> modifiersMap = window.getModifiersMap();
                List<String> modifiers = keyboardModifiers( eventData, modifiersMap, macOS );
                mouseMoveData = new MouseMoveData( Integer.parseInt( id ), left, top, modifiers );
            }
        }
        return mouseMoveData;
    }

    public MouseClickData processMouseClick( VirtualClient virtualClient, JsonObject eventData, boolean butonPressed, boolean macOS )
    {
        MouseClickData mouseClickData = null;
        String id = StringUtils.defaultIfBlank( eventData.getString( ELEMENT_ID ), StringUtils.EMPTY );
        if ( StringUtils.isNotBlank( id ) && StringUtils.isNumeric( id ) )
        {
            Window window = virtualClient.getWindow( Integer.parseInt( id ) );
            if ( Objects.nonNull( window ) )
            {
                int button = eventButton( eventData );
                int pageX = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_X ) ) ? (int) eventData.getNumber( MOUSE_PAGE_X ) : 0;
                int pageY = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_Y ) ) ? (int) eventData.getNumber( MOUSE_PAGE_Y ) : 0;
                int left = (int) ( pageX - window.getOffsetLeft() - window.getBorderWidth() - window.getDx() );
                int top = (int) ( pageY - window.getOffsetTop() - window.getBorderWidth() - window.getDy() - window.getTitleHeight() );
                Map<String, String> modifiersMap = window.getModifiersMap();
                List<String> modifiers = keyboardModifiers( eventData, modifiersMap, macOS );
                mouseClickData = new MouseClickData( Integer.parseInt( id ), left, top, button, modifiers, butonPressed );
            }
        }
        return mouseClickData;
    }

    public List<Object> processMouseWheel( VirtualClient virtualClient, JsonObject eventData, boolean macOS )
    {
        List<Object> mouseDatas = new ArrayList<>();
        String id = StringUtils.defaultIfBlank( eventData.getString( ELEMENT_ID ), StringUtils.EMPTY );
        if ( StringUtils.isNotBlank( id ) && StringUtils.isNumeric( id ) )
        {
            Window window = virtualClient.getWindow( Integer.parseInt( id ) );
            if ( Objects.nonNull( window ) )
            {
                int pageX = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_X ) ) ? (int) eventData.getNumber( MOUSE_PAGE_X ) : 0;
                int pageY = Objects.nonNull( eventData.getNumber( MOUSE_PAGE_Y ) ) ? (int) eventData.getNumber( MOUSE_PAGE_Y ) : 0;
                int left = (int) ( pageX - window.getOffsetLeft() - window.getBorderWidth() - window.getDx() );
                int top = (int) ( pageY - window.getOffsetTop() - window.getBorderWidth() - window.getDy() - window.getTitleHeight() );
                Map<String, String> modifiersMap = window.getModifiersMap();
                List<String> modifiers = keyboardModifiers( eventData, modifiersMap, macOS );
                NormalizedMouseWheelData normalizedMouseWheelData = normalizeMouseWheel( eventData );

                double px = Math.min( 1200, normalizedMouseWheelData.getPixelX() );
                double pxAbsolute = Math.abs( px );
                double py = Math.min( 1200, normalizedMouseWheelData.getPixelY() );
                double pyAbsolute = Math.abs( py );
                int xButton = 0;
                int xDistance = 0;
                int yButton = 0;
                int yDistance = 0;
                if ( window.isMousePreciseWheel() )
                {
                    if ( pxAbsolute > 0 )
                    {
                        xButton = px >= 0 ? 6 : 7;
                        xDistance = (int) Math.round( px * 1000 / MOUSE_WHEEL_DELTA_VALUE );
                        mouseDatas.add( new MouseWheelData( Integer.parseInt( id ), left, top, xButton, xDistance, modifiers ) );
                    }
                    if ( pyAbsolute > 0 )
                    {
                        yButton = py >= 0 ? 5 : 4;
                        yDistance = (int) Math.round( py * 1000 / MOUSE_WHEEL_DELTA_VALUE );
                        mouseDatas.add( new MouseWheelData( Integer.parseInt( id ), left, top, yButton, yDistance, modifiers ) );
                    }
                }
                else
                {
                    // generate a single event if we can, or add to accumulators
                    if ( pxAbsolute >= 40 && pxAbsolute <= 160 )
                    {
                        window.setMouseWheelDeltaX( px > 0 ? MOUSE_WHEEL_DELTA_VALUE : -MOUSE_WHEEL_DELTA_VALUE );
                    }
                    else
                    {
                        window.setMouseWheelDeltaX( window.getMouseWheelDeltaX() + px );
                    }
                    if ( pyAbsolute >= 40 && pyAbsolute <= 160 )
                    {
                        window.setMouseWheelDeltaY( py > 0 ? MOUSE_WHEEL_DELTA_VALUE : -MOUSE_WHEEL_DELTA_VALUE );
                    }
                    else
                    {
                        window.setMouseWheelDeltaY( window.getMouseWheelDeltaY() + py );
                    }

                    // send synthetic click/release as many times as needed
                    xButton = window.getMouseWheelDeltaX() >= 0 ? 6 : 7;
                    yButton = window.getMouseWheelDeltaY() >= 0 ? 5 : 4;
                    double xMouseWheel = Math.abs( window.getMouseWheelDeltaX() );
                    double yMouseWheel = Math.abs( window.getMouseWheelDeltaY() );
                    if ( xMouseWheel >= MOUSE_WHEEL_DELTA_VALUE )
                    {
                        do
                        {
                            mouseDatas.add( new MouseClickData( Integer.parseInt( id ), left, top, xButton, modifiers, true ) );
                            xMouseWheel -= MOUSE_WHEEL_DELTA_VALUE;
                            window.setMouseWheelDeltaX( window.getMouseWheelDeltaX() >= 0 ? xMouseWheel : -xMouseWheel );
                        }
                        while ( xMouseWheel >= MOUSE_WHEEL_DELTA_VALUE );
                    }
                    if ( yMouseWheel >= MOUSE_WHEEL_DELTA_VALUE )
                    {
                        do
                        {
                            mouseDatas.add( new MouseClickData( Integer.parseInt( id ), left, top, yButton, modifiers, true ) );
                            yMouseWheel -= MOUSE_WHEEL_DELTA_VALUE;
                            window.setMouseWheelDeltaY( window.getMouseWheelDeltaY() >= 0 ? yMouseWheel : -yMouseWheel );
                        }
                        while ( yMouseWheel >= MOUSE_WHEEL_DELTA_VALUE );
                    }

                    // store left overs
                    window.setMouseWheelDeltaX( window.getMouseWheelDeltaX() >= 0 ? xMouseWheel : -xMouseWheel );
                    window.setMouseWheelDeltaY( window.getMouseWheelDeltaY() >= 0 ? yMouseWheel : -yMouseWheel );
                }
            }
        }
        return mouseDatas;
    }

    public void setCapsLock( Boolean capsLock )
    {
        this.capsLock = capsLock;
    }

    public void setNumLock( Boolean numLock )
    {
        this.numLock = numLock;
    }

}
