/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.users;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.UserColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.value.ValueChangeMode;

public class UsersUI
    extends VerticalLayout
{

    private static final long serialVersionUID = -1808090692508134917L;

    private static final int FILTER_VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final String CLASS_SETTINGS = "settings";

    private static final String GRID_GROUPS = "grid-groups";

    private static final String GRID_GROUPS_ICON = "grid-groups-icon";

    private static final String GRID_GROUPS_LABEL = "grid-groups-label";

    private static final String NAME_WITH_NOTES = "%s (%s)";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_RED_VALUE = "red";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "1px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_200PX_VALUE = "200px";

    private static final String STYLE_WIDTH_250PX_VALUE = "250px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private transient List<UserData> userDatas = null;

    private transient Map<UserColumns, String> filterMap = null;

    private Database database = null;

    private DataProvider<UserData, Void> dataProvider = null;

    private Grid<UserData> userGrid = null;

    private Grid.Column<UserData> editDeleteColumn = null;

    private VirtualUtil virtualUtil = null;

    public UsersUI()
    {
        setMargin( false );
        setSpacing( false );
        setSizeFull();
        addClassName( CLASS_SETTINGS );
        filterMap = new EnumMap<>( UserColumns.class );
    }

    public UsersUI( Database database, VirtualUtil virtualUtil )
    {
        this();
        this.database = database;
        this.virtualUtil = virtualUtil;

        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( SortDirection.DESCENDING.equals( order.getDirection() ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( this.database.getUsers( offset, limit, filterMap, orders ) ).stream();
        }, query -> database.getUsersRows( filterMap )
                            .intValue() );

        userDatas = new ArrayList<>( database.getAllUsers() );
        userGrid = new Grid<>();
        userGrid.setDataProvider( dataProvider );
        userGrid.setPageSize( VISIBLE_ROWS );
        userGrid.getStyle()
                .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE );
        userGrid.addItemDoubleClickListener( event -> {
            userGrid.select( event.getItem() );
            editUserData( event.getItem() ).open();
        } );
        add( userGrid );

        // user name
        Grid.Column<UserData> userNameColumn = userGrid.addColumn( UserData::getUserName )
                                                       .setFlexGrow( 1 )
                                                       .setResizable( true )
                                                       .setSortable( true )
                                                       .setSortProperty( UserColumns.USERNAME.getName() )
                                                       .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                       .setHeader( virtualUtil.getTranslation( UserColumns.USERNAME.getValueForTranslation() ) );
        // full name
        Grid.Column<UserData> firstNameColumn = userGrid.addColumn( UserData::getFullName )
                                                        .setFlexGrow( 1 )
                                                        .setResizable( true )
                                                        .setSortable( true )
                                                        .setSortProperty( UserColumns.FULLNAME.getName() )
                                                        .setWidth( STYLE_WIDTH_200PX_VALUE )
                                                        .setHeader( virtualUtil.getTranslation( UserColumns.FULLNAME.getValueForTranslation() ) );
        // applications group
        Grid.Column<UserData> applicationsGroupColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( userDiv, userData ) -> {
            database.getUserApplicationGroups( userData )
                    .stream()
                    .filter( Objects::nonNull )
                    .sorted( ( data1, data2 ) -> data1.getApplicationsGroupData()
                                                      .getName()
                                                      .compareTo( data2.getApplicationsGroupData()
                                                                       .getName() ) )
                    .forEach( userApplicationsGroupData -> {
                        ApplicationsGroupData applicationsGroupData = userApplicationsGroupData.getApplicationsGroupData();
                        Label label = new Label( StringUtils.isBlank( applicationsGroupData.getNote() ) ? applicationsGroupData.getName()
                                        : String.format( NAME_WITH_NOTES, applicationsGroupData.getName(), applicationsGroupData.getNote() ) );
                        label.addClassName( GRID_GROUPS_LABEL );
                        Icon icon = applicationsGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                        icon.addClassName( GRID_GROUPS_ICON );

                        Div applicationsGroupDiv = new Div();
                        applicationsGroupDiv.addClassName( GRID_GROUPS );
                        applicationsGroupDiv.add( icon );
                        applicationsGroupDiv.add( label );
                        userDiv.add( applicationsGroupDiv );
                    } );
        } ) )
                                                                .setFlexGrow( 1 )
                                                                .setResizable( true )
                                                                .setWidth( STYLE_WIDTH_250PX_VALUE )
                                                                .setHeader( virtualUtil.getTranslation( UserColumns.APPLICATIONS_GROUP.getValueForTranslation() ) );
        // application
        Grid.Column<UserData> applicationColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( applicationsDiv, userData ) -> {
            database.getUserApplications( userData )
                    .stream()
                    .filter( Objects::nonNull )
                    .sorted( ( data1, data2 ) -> data1.getApplicationData()
                                                      .getName()
                                                      .compareTo( data2.getApplicationData()
                                                                       .getName() ) )
                    .forEach( userApplicationData -> {
                        ApplicationData applicationData = userApplicationData.getApplicationData();
                        Label label = new Label( StringUtils.isBlank( applicationData.getNote() ) ? applicationData.getName()
                                        : String.format( NAME_WITH_NOTES, applicationData.getName(), applicationData.getNote() ) );
                        label.addClassName( GRID_GROUPS_LABEL );
                        Icon icon = applicationData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                        icon.addClassName( GRID_GROUPS_ICON );

                        Div applicationDiv = new Div();
                        applicationDiv.addClassName( GRID_GROUPS );
                        applicationDiv.add( icon );
                        applicationDiv.add( label );
                        applicationsDiv.add( applicationDiv );
                    } );
        } ) )
                                                          .setFlexGrow( 1 )
                                                          .setResizable( true )
                                                          .setWidth( STYLE_WIDTH_250PX_VALUE )
                                                          .setHeader( virtualUtil.getTranslation( Translation.FIELD_APPLICATION ) );
        // maximum applications
        Grid.Column<UserData> maxApplicationsColumn = userGrid.addColumn( UserData::getMaxApplications )
                                                              .setFlexGrow( 0 )
                                                              .setResizable( true )
                                                              .setSortable( true )
                                                              .setSortProperty( UserColumns.APPLICATION_MAX.getName() )
                                                              .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                              .setHeader( virtualUtil.getTranslation( UserColumns.APPLICATION_MAX.getValueForTranslation() ) );
        // maximum file size
        Grid.Column<UserData> maxFileSizeColumn = userGrid.addColumn( UserData::getMaxFileSize )
                                                          .setFlexGrow( 0 )
                                                          .setResizable( true )
                                                          .setSortable( true )
                                                          .setSortProperty( UserColumns.FILE_SIZE.getName() )
                                                          .setWidth( STYLE_WIDTH_250PX_VALUE )
                                                          .setHeader( virtualUtil.getTranslation( UserColumns.FILE_SIZE.getValueForTranslation() ) );
        // administrator
        Grid.Column<UserData> administratorColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isAdministrator() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                            .setFlexGrow( 0 )
                                                            .setResizable( true )
                                                            .setSortable( true )
                                                            .setSortProperty( UserColumns.ADMINISTRATOR.getName() )
                                                            .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                            .setHeader( virtualUtil.getTranslation( UserColumns.ADMINISTRATOR.getValueForTranslation() ) );
        // screenshot
        Grid.Column<UserData> screenshotColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isScreenshot() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                         .setFlexGrow( 0 )
                                                         .setResizable( true )
                                                         .setSortable( true )
                                                         .setSortProperty( UserColumns.UPLOAD.getName() )
                                                         .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                         .setHeader( virtualUtil.getTranslation( UserColumns.SCREENSHOT.getValueForTranslation() ) );
        // upload
        Grid.Column<UserData> uploadColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isUpload() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                     .setFlexGrow( 0 )
                                                     .setResizable( true )
                                                     .setSortable( true )
                                                     .setSortProperty( UserColumns.UPLOAD.getName() )
                                                     .setWidth( STYLE_WIDTH_200PX_VALUE )
                                                     .setHeader( virtualUtil.getTranslation( UserColumns.UPLOAD.getValueForTranslation() ) );
        // download
        Grid.Column<UserData> downloadColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isDownload() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                       .setFlexGrow( 0 )
                                                       .setResizable( true )
                                                       .setSortable( true )
                                                       .setSortProperty( UserColumns.DOWNLOAD.getName() )
                                                       .setWidth( STYLE_WIDTH_200PX_VALUE )
                                                       .setHeader( virtualUtil.getTranslation( UserColumns.DOWNLOAD.getValueForTranslation() ) );
        // full screen
        Grid.Column<UserData> fullScreenColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isFullScreen() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                         .setFlexGrow( 0 )
                                                         .setResizable( true )
                                                         .setSortable( true )
                                                         .setSortProperty( UserColumns.UPLOAD.getName() )
                                                         .setWidth( STYLE_WIDTH_200PX_VALUE )
                                                         .setHeader( virtualUtil.getTranslation( UserColumns.FULL_SCREEN.getValueForTranslation() ) );
        // enabled
        Grid.Column<UserData> enabledColumn = userGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, userData ) -> {
            Icon icon = userData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                      .setFlexGrow( 0 )
                                                      .setResizable( true )
                                                      .setSortable( true )
                                                      .setSortProperty( UserColumns.ENABLED.getName() )
                                                      .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                      .setHeader( virtualUtil.getTranslation( UserColumns.ENABLED.getValueForTranslation() ) );
        // edit/delete data
        editDeleteColumn = userGrid.addColumn( new ComponentRenderer<>( userData -> {
            // delete
            Button deleteButton = new Button( new Icon( VaadinIcon.TRASH ), event -> {
                UserRemoveDialog dialog = new UserRemoveDialog( userData, virtualUtil )
                {
                    private static final long serialVersionUID = 2628864348848236761L;

                    @Override
                    public void savingData( UserData data )
                    {
                        if ( Objects.nonNull( data ) )
                        {
                            removeUser( userData );
                        }
                        else
                        {
                            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                        }
                    }
                };
                dialog.open();
                userGrid.select( userData );
            } );
            deleteButton.getStyle()
                        .set( STYLE_COLOR, STYLE_COLOR_RED_VALUE );

            // edit
            Button editButton = new Button( new Icon( VaadinIcon.EDIT ), event -> {
                userGrid.select( userData );
                editUserData( userData ).open();
            } );

            if ( virtualUtil.isTooltipsVisible() )
            {
                TooltipRunner tooltipRunner1 = new TooltipRunner( deleteButton, virtualUtil.getTranslation( Translation.BUTTON_DELETE ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner1 );
                TooltipRunner tooltipRunner2 = new TooltipRunner( editButton, virtualUtil.getTranslation( Translation.BUTTON_EDIT ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner2 );
            }

            HorizontalLayout layout = new HorizontalLayout( editButton, deleteButton );
            layout.setMargin( false );
            layout.setPadding( false );
            layout.setSpacing( false );
            layout.setSizeUndefined();
            return layout;
        } ) )
                                   .setFlexGrow( 0 )
                                   .setWidth( STYLE_WIDTH_100PX_VALUE )
                                   .setResizable( false );
        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( userDatas ) );

        // filter: user name
        HeaderRow filterRow = userGrid.appendHeaderRow();
        TextField userNameFilter = new TextField();
        userNameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        userNameFilter.setSizeFull();
        userNameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        userNameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        userNameFilter.addValueChangeListener( event -> {
            if ( userNameFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.USERNAME );
            }
            else
            {
                filterMap.put( UserColumns.USERNAME, userNameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( userNameColumn )
                 .setComponent( userNameFilter );

        // filter: full name
        TextField fullNameFilter = new TextField();
        fullNameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        fullNameFilter.setSizeFull();
        fullNameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        fullNameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        fullNameFilter.addValueChangeListener( event -> {
            if ( fullNameFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.FULLNAME );
            }
            else
            {
                filterMap.put( UserColumns.FULLNAME, fullNameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( firstNameColumn )
                 .setComponent( fullNameFilter );

        // filter: user group
        TextField applicationsGroupFilter = new TextField();
        applicationsGroupFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        applicationsGroupFilter.setSizeFull();
        applicationsGroupFilter.setValueChangeMode( ValueChangeMode.LAZY );
        applicationsGroupFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        applicationsGroupFilter.addValueChangeListener( event -> {
            if ( applicationsGroupFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.APPLICATIONS_GROUP );
            }
            else
            {
                filterMap.put( UserColumns.APPLICATIONS_GROUP, applicationsGroupFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( applicationsGroupColumn )
                 .setComponent( applicationsGroupFilter );

        // filter: application
        TextField applicationFilter = new TextField();
        applicationFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        applicationFilter.setSizeFull();
        applicationFilter.setValueChangeMode( ValueChangeMode.LAZY );
        applicationFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        applicationFilter.addValueChangeListener( event -> {
            if ( applicationFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.APPLICATION );
            }
            else
            {
                filterMap.put( UserColumns.APPLICATION, applicationFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( applicationColumn )
                 .setComponent( applicationFilter );

        // filter: maximum applications
        TextField maxApplicationsFilter = new TextField();
        maxApplicationsFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        maxApplicationsFilter.setSizeFull();
        maxApplicationsFilter.setValueChangeMode( ValueChangeMode.LAZY );
        maxApplicationsFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        maxApplicationsFilter.addValueChangeListener( event -> {
            if ( maxApplicationsFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.APPLICATION_MAX );
            }
            else
            {
                filterMap.put( UserColumns.APPLICATION_MAX, maxApplicationsFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( maxApplicationsColumn )
                 .setComponent( maxApplicationsFilter );

        // filter: maximum file size
        TextField maxFileSizeFilter = new TextField();
        maxFileSizeFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        maxFileSizeFilter.setSizeFull();
        maxFileSizeFilter.setValueChangeMode( ValueChangeMode.LAZY );
        maxFileSizeFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        maxFileSizeFilter.addValueChangeListener( event -> {
            if ( maxFileSizeFilter.isEmpty() )
            {
                filterMap.remove( UserColumns.FILE_SIZE );
            }
            else
            {
                filterMap.put( UserColumns.FILE_SIZE, maxFileSizeFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( maxFileSizeColumn )
                 .setComponent( maxFileSizeFilter );

        // filter: administrator
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> administratorFilter = new ComboBox<>();
        administratorFilter.setAllowCustomValue( false );
        administratorFilter.setClassName( STYLE_GRID_COMBOBOX );
        administratorFilter.setItems( filterAll, filterYes, filterNo );
        administratorFilter.setValue( filterAll );
        administratorFilter.setSizeFull();
        administratorFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( administratorFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.ADMINISTRATOR, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( administratorFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.ADMINISTRATOR, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.ADMINISTRATOR );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( administratorColumn )
                 .setComponent( administratorFilter );

        // filter: screenshot
        ComboBox<String> screenshotFilter = new ComboBox<>();
        screenshotFilter.setAllowCustomValue( false );
        screenshotFilter.setClassName( STYLE_GRID_COMBOBOX );
        screenshotFilter.setItems( filterAll, filterYes, filterNo );
        screenshotFilter.setValue( filterAll );
        screenshotFilter.setSizeFull();
        screenshotFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( screenshotFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.SCREENSHOT, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( screenshotFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.SCREENSHOT, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.SCREENSHOT );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( screenshotColumn )
                 .setComponent( screenshotFilter );

        // filter: upload
        ComboBox<String> uploadFilter = new ComboBox<>();
        uploadFilter.setAllowCustomValue( false );
        uploadFilter.setClassName( STYLE_GRID_COMBOBOX );
        uploadFilter.setItems( filterAll, filterYes, filterNo );
        uploadFilter.setValue( filterAll );
        uploadFilter.setSizeFull();
        uploadFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( uploadFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.UPLOAD, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( uploadFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.UPLOAD, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.UPLOAD );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( uploadColumn )
                 .setComponent( uploadFilter );

        // filter: download
        ComboBox<String> downloadFilter = new ComboBox<>();
        downloadFilter.setAllowCustomValue( false );
        downloadFilter.setClassName( STYLE_GRID_COMBOBOX );
        downloadFilter.setItems( filterAll, filterYes, filterNo );
        downloadFilter.setValue( filterAll );
        downloadFilter.setSizeFull();
        downloadFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( downloadFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.DOWNLOAD, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( downloadFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.DOWNLOAD, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.DOWNLOAD );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( downloadColumn )
                 .setComponent( downloadFilter );

        // filter: full screen
        ComboBox<String> fullScreenFilter = new ComboBox<>();
        fullScreenFilter.setAllowCustomValue( false );
        fullScreenFilter.setClassName( STYLE_GRID_COMBOBOX );
        fullScreenFilter.setItems( filterAll, filterYes, filterNo );
        fullScreenFilter.setValue( filterAll );
        fullScreenFilter.setSizeFull();
        fullScreenFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( fullScreenFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.FULL_SCREEN, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( fullScreenFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.FULL_SCREEN, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.FULL_SCREEN );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( fullScreenColumn )
                 .setComponent( fullScreenFilter );

        // filter: enabled
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );
    }

    private UserEditDialog editUserData( UserData userData )
    {
        return ( new UserEditDialog( userData.getId(), userData, database, virtualUtil )
        {
            private static final long serialVersionUID = 601871212063757874L;

            @Override
            public void savingData( UserData data, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
            {
                if ( Objects.nonNull( data ) )
                {
                    boolean closeDialog = true;
                    if ( Objects.nonNull( database.getUser( data.getId() ) ) )
                    {
                        closeDialog = updateUser( data, applicationsGroupDatas, applicationDatas );
                    }
                    else
                    {
                        closeDialog = addUser( data, applicationsGroupDatas, applicationDatas );
                    }

                    if ( closeDialog )
                    {
                        close();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        } );
    }

    public boolean addUser( UserData userData, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
    {
        if ( Objects.nonNull( userData ) )
        {
            userData.setId( database.addUser( userData ) );
            if ( !Objects.equals( userData.getId(), UserData.UNDEFINED_USER ) )
            {
                database.getUserApplicationGroups( userData )
                        .forEach( userApplicationsGroupData -> database.removeUserApplicationsGroup( userApplicationsGroupData ) );
                applicationsGroupDatas.forEach( applicationsGroupData -> database.addUserApplicationsGroup( userData, applicationsGroupData ) );
                database.getUserApplications( userData )
                        .forEach( userApplication -> database.removeUserApplication( userApplication ) );
                applicationDatas.forEach( application -> database.addUserApplication( userData, application ) );

                userDatas.add( userData );
                dataProvider.refreshAll();
                userGrid.select( userData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( userDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.USER_ERROR_ADD ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean addUser( long id, String userName, String fullName, int maxApplications, int maxFileSize, boolean administrator, boolean screenshot, boolean upload, boolean download,
                            boolean fullScreen, boolean enabled, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
    {
        return addUser( new UserData( id, userName, fullName, maxApplications, maxFileSize, administrator, screenshot, upload, download, fullScreen, enabled ), applicationsGroupDatas,
                        applicationDatas );
    }

    public boolean removeUser( UserData userData )
    {
        if ( Objects.nonNull( userData ) )
        {
            database.getUserApplications( userData )
                    .forEach( userApplication -> database.removeUserApplication( userApplication ) );
            database.getUserApplicationGroups( userData )
                    .forEach( userApplicationsGroupData -> database.removeUserApplicationsGroup( userApplicationsGroupData ) );
            if ( database.removeUser( userData ) )
            {
                userGrid.deselectAll();
                userDatas.remove( userData );
                dataProvider.refreshAll();
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( userDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.USER_ERROR_REMOVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean removeUser( int userId )
    {
        return removeUser( database.getUser( userId ) );
    }

    public boolean updateUser( UserData userData, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
    {
        if ( Objects.nonNull( userData ) )
        {
            if ( database.updateUser( userData ) )
            {
                database.getUserApplicationGroups( userData )
                        .forEach( userApplicationsGroupData -> database.removeUserApplicationsGroup( userApplicationsGroupData ) );
                applicationsGroupDatas.forEach( applicationsGroupData -> database.addUserApplicationsGroup( userData, applicationsGroupData ) );
                database.getUserApplications( userData )
                        .forEach( userApplication -> database.removeUserApplication( userApplication ) );
                applicationDatas.forEach( application -> database.addUserApplication( userData, application ) );
                userDatas.removeIf( data -> Objects.equals( data.getId(), userData.getId() ) );
                userDatas.add( userData );
                dataProvider.refreshAll();
                userGrid.select( userData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( userDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE, userData.getUserName() ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean updateUser( long id, String userName, String fullName, int maxApplications, int maxFileSize, boolean administrator, boolean screenshot, boolean upload, boolean download,
                               boolean fullScreen, boolean enabled, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
    {
        return updateUser( new UserData( id, userName, fullName, maxApplications, maxFileSize, administrator, screenshot, upload, download, fullScreen, enabled ), applicationsGroupDatas,
                           applicationDatas );
    }

    public void addNewUser()
    {
        UserEditDialog dialog = new UserEditDialog( UserData.NEW_USER, new UserData(), database, virtualUtil )
        {
            private static final long serialVersionUID = 4381057068266036061L;

            @Override
            public void savingData( UserData userData, List<ApplicationsGroupData> applicationsGroupDatas, List<ApplicationData> applicationDatas )
            {
                if ( Objects.nonNull( userData ) )
                {
                    userData.setId( database.addUser( userData ) );
                    if ( !Objects.equals( userData.getId(), UserData.UNDEFINED_USER ) )
                    {
                        applicationsGroupDatas.forEach( applicationsGroupData -> database.addUserApplicationsGroup( userData, applicationsGroupData ) );
                        applicationDatas.forEach( application -> database.addUserApplication( userData, application ) );
                        userDatas.add( userData );
                        dataProvider.refreshAll();
                        userGrid.select( userData );
                        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( userDatas ) );
                        close();
                    }
                    else
                    {
                        new DialogInfoUtil( virtualUtil.getTranslation( Translation.USER_ERROR_EXISTS, userData.getUserName() ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        };
        dialog.open();
    }

}
