/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.usersessions;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.UserSessionColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.grid.ColumnTextAlign;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.IronIcon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.value.ValueChangeMode;
import com.vaadin.flow.server.StreamResource;

public class UserSessionsUI
    extends VerticalLayout
{

    private static final long serialVersionUID = -5291848215346722410L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int FILTER_VALUE_CHANGE_TIMEOUT = 500;

    private static final int MAX_WAIT_INTERVAL = 10000; // milliseconds

    private static final int NO_COMPONENTS = 0;

    private static final int NO_STREAM_VALUE = -1;

    private static final int SLEEP_INTERVAL = 100; // milliseconds

    private static final int VISIBLE_ROWS = 50;

    private static final String CLASS_ICON = "session-icon";

    private static final String CLASS_LABEL = "label";

    private static final String CLASS_SETTINGS = "settings";

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String SERVER_STOPPED = "The server is stopped.";

    private static final String SERVER_STOPPING = "Stopping the server: {}:{}";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_RED_VALUE = "red";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "1px";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_250PX_VALUE = "250px";

    private static final String STYLE_WIDTH_80PX_VALUE = "80px";

    private transient List<UserSessionData> userSessionDatas = null;

    private transient Map<UserSessionColumns, String> filterMap = null;

    private Database database = null;

    private DataProvider<String, Void> dataProvider = null;

    private Grid<String> userSessionGrid = null;

    private VirtualUtil virtualUtil = null;

    private UserSessionsUI()
    {
        setMargin( false );
        setSpacing( true );
        setSizeFull();
        addClassName( CLASS_SETTINGS );
        filterMap = new EnumMap<>( UserSessionColumns.class );
    }

    public UserSessionsUI( Database database, VirtualUtil virtualUtil )
    {
        this();
        this.database = database;
        this.virtualUtil = virtualUtil;

        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> canvasIds = new ArrayList<>( this.database.getAllUserSessionCanvasIds( offset, limit, filterMap ) );
            userSessionDatas = this.database.getAllUserSessions( filterMap, canvasIds.toArray( String[]::new ) );

            // add child applications
            for ( UserSessionData userSessionData : new ArrayList<>( userSessionDatas ) )
            {
                database.getUserSessionApplications( userSessionData )
                        .stream()
                        .forEach( userSessionApplicationData -> {
                            ApplicationData applicationData = database.getApplication( userSessionApplicationData.getApplicationId() );
                            UserSessionData sessionData =
                                new UserSessionData( userSessionData.getId(), userSessionData.getUserData(), userSessionData.getServerData(), userSessionData.getPort(), applicationData,
                                                     userSessionData.getStatus(), userSessionData.getCanvasId(), userSessionData.getDesktopWidth(), userSessionData.getDesktopHeight(),
                                                     userSessionApplicationData.getWindowId(), userSessionData.isLocked(), userSessionData.getUpdated() );
                            userSessionDatas.add( sessionData );
                        } );
            }
            userSessionDatas.removeIf( userSessionData -> Objects.isNull( userSessionData.getWindowId() ) );
            userSessionDatas.sort( ( UserSessionData data1, UserSessionData data2 ) -> data1.getApplicationData()
                                                                                            .getName()
                                                                                            .compareTo( data2.getApplicationData()
                                                                                                             .getName() ) );
            return canvasIds.stream();
        }, query -> database.getAllUserSessionCanvasIdsRows( filterMap ) );

        userSessionGrid = new Grid<>();
        userSessionGrid.setDataProvider( dataProvider );
        userSessionGrid.setPageSize( VISIBLE_ROWS );
        userSessionGrid.getStyle()
                       .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                       .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                       .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE );
        add( userSessionGrid );

        // full name
        Grid.Column<String> fullNameColumn = userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                            .forEach( userSessionData -> {
                                if ( divs.getChildren()
                                         .count() == NO_COMPONENTS )
                                {
                                    Div div = new Div();
                                    Label label = new Label( userSessionData.getUserData()
                                                                            .getFullName() );
                                    div.add( label );
                                    divs.add( div );
                                }
                            } );
        } ) )
                                                            .setFlexGrow( 1 )
                                                            .setResizable( true )
                                                            .setSortable( false )
                                                            // .setSortProperty( UserSessionColumns.NAME_FULL.getName() )
                                                            .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                            .setHeader( virtualUtil.getTranslation( UserSessionColumns.NAME_FULL.getValueForTranslation() ) );
        // application
        Grid.Column<String> applicationColumn = userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                            .forEach( userSessionData -> {
                                Div userSessionDiv = new Div();
                                IconData iconData = userSessionData.getApplicationData()
                                                                   .getIconData();
                                if ( Objects.isNull( iconData.getIcon() ) )
                                {
                                    String iconName = iconData.getName();
                                    IronIcon ironIcon = new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                                                      iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
                                    ironIcon.addClassName( CLASS_ICON );
                                    userSessionDiv.add( ironIcon );
                                }
                                else
                                {
                                    StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                                    Image image = new Image( streamResource, iconData.getName() );
                                    image.addClassName( CLASS_ICON );
                                    userSessionDiv.add( image );
                                }

                                Label label = new Label( userSessionData.getApplicationData()
                                                                        .getName() );
                                label.addClassName( CLASS_LABEL );

                                userSessionDiv.add( label );
                                divs.add( userSessionDiv );
                            } );
        } ) )
                                                               .setFlexGrow( 1 )
                                                               .setResizable( true )
                                                               .setSortable( false )
                                                               .setWidth( STYLE_WIDTH_250PX_VALUE )
                                                               .setHeader( virtualUtil.getTranslation( UserSessionColumns.APPLICATION.getValueForTranslation() ) );
        // server
        Grid.Column<String> serverColumn = userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                            .forEach( userSessionData -> {
                                if ( divs.getChildren()
                                         .count() == 0 )
                                {
                                    Div userSessionDiv = new Div();
                                    Label label = new Label( userSessionData.getServerData()
                                                                            .getName() );
                                    label.addClassName( CLASS_LABEL );

                                    userSessionDiv.add( label );
                                    divs.add( userSessionDiv );
                                }
                            } );
        } ) )
                                                          .setFlexGrow( 1 )
                                                          .setResizable( true )
                                                          .setSortable( false )
                                                          .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                          .setHeader( virtualUtil.getTranslation( Translation.FIELD_SERVER ) );
        // occupied
        Grid.Column<String> activeColumn = userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) && userSessionData.isLocked() )
                            .forEach( userSessionData -> {
                                if ( divs.getChildren()
                                         .count() == 0 )
                                {
                                    Icon icon = new Icon( VaadinIcon.CHECK );
                                    icon.getStyle()
                                        .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                                        .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
                                    divs.add( icon );
                                }
                            } );
        } ) )
                                                          .setFlexGrow( 0 )
                                                          .setResizable( true )
                                                          .setSortable( false )
                                                          // .setSortProperty( UserSessionColumns.ACTIVE.getName() )
                                                          .setTextAlign( ColumnTextAlign.CENTER )
                                                          .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                          .setHeader( virtualUtil.getTranslation( UserSessionColumns.ACTIVE.getValueForTranslation() ) );
        // updated
        userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                            .forEach( userSessionData -> {
                                if ( divs.getChildren()
                                         .count() == 0 )
                                {
                                    Div userSessionDiv = new Div();
                                    Label label = new Label();
                                    try
                                    {
                                        SimpleDateFormat simpleDateFormat = new SimpleDateFormat( virtualUtil.getTranslation( Translation.FIELD_UPDATED_FORMAT ) );
                                        label = new Label( simpleDateFormat.format( userSessionData.getUpdated() ) );
                                    }
                                    catch ( IllegalArgumentException | NullPointerException exception )
                                    {
                                        label = new Label( userSessionData.getUpdated()
                                                                          .toString() );
                                    }
                                    userSessionDiv.add( label );
                                    divs.add( userSessionDiv );
                                }
                            } );
        } ) )
                       .setFlexGrow( 0 )
                       .setResizable( false )
                       .setSortable( false )
                       // .setSortProperty( UserSessionColumns.UPDATED.getName() )
                       .setWidth( STYLE_WIDTH_150PX_VALUE )
                       .setHeader( virtualUtil.getTranslation( UserSessionColumns.UPDATED.getValueForTranslation() ) );

        // delete session
        userSessionGrid.addColumn( new ComponentRenderer<>( Div::new, ( divs, canvasId ) -> {
            userSessionDatas.stream()
                            .filter( Objects::nonNull )
                            .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                            .forEach( userSessionData -> {
                                if ( divs.getChildren()
                                         .count() == 0 )
                                {
                                    Button deleteButton = new Button( new Icon( VaadinIcon.TRASH ), event -> {
                                        UserSessionRemoveDialog userSessionRemoveDialog = new UserSessionRemoveDialog( virtualUtil );
                                        userSessionRemoveDialog.setDeleteButtonClickListener( e -> {
                                            stopServer( canvasId );
                                            dataProvider.refreshAll();
                                            userSessionRemoveDialog.close();
                                        } );
                                        userSessionGrid.select( canvasId );
                                        userSessionRemoveDialog.open();
                                    } );
                                    deleteButton.getStyle()
                                                .set( STYLE_COLOR, STYLE_COLOR_RED_VALUE );
                                    divs.add( deleteButton );

                                    if ( virtualUtil.isTooltipsVisible() )
                                    {
                                        TooltipRunner tooltipRunner = new TooltipRunner( deleteButton, virtualUtil.getTranslation( Translation.BUTTON_DELETE ), UI.getCurrent() );
                                        virtualUtil.startRunnerService( tooltipRunner );
                                    }
                                }
                            } );
        } ) )
                       .setFlexGrow( 0 )
                       .setTextAlign( ColumnTextAlign.CENTER )
                       .setResizable( false )
                       .setWidth( STYLE_WIDTH_80PX_VALUE );

        // filter: full name
        HeaderRow filterRow = userSessionGrid.appendHeaderRow();
        TextField fullNameFilter = new TextField();
        fullNameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        fullNameFilter.setSizeFull();
        fullNameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        fullNameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        fullNameFilter.addValueChangeListener( event -> {
            if ( fullNameFilter.isEmpty() )
            {
                filterMap.remove( UserSessionColumns.NAME_FULL );
            }
            else
            {
                filterMap.put( UserSessionColumns.NAME_FULL, fullNameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( fullNameColumn )
                 .setComponent( fullNameFilter );

        // filter: applications
        TextField applicationFilter = new TextField();
        applicationFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        applicationFilter.setSizeFull();
        applicationFilter.setValueChangeMode( ValueChangeMode.LAZY );
        applicationFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        applicationFilter.addValueChangeListener( event -> {
            if ( applicationFilter.isEmpty() )
            {
                filterMap.remove( UserSessionColumns.APPLICATION );
            }
            else
            {
                filterMap.put( UserSessionColumns.APPLICATION, applicationFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( applicationColumn )
                 .setComponent( applicationFilter );

        // filter: applications
        TextField serverFilter = new TextField();
        serverFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        serverFilter.setSizeFull();
        serverFilter.setValueChangeMode( ValueChangeMode.LAZY );
        serverFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        serverFilter.addValueChangeListener( event -> {
            if ( serverFilter.isEmpty() )
            {
                filterMap.remove( UserSessionColumns.SERVER );
            }
            else
            {
                filterMap.put( UserSessionColumns.SERVER, serverFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( serverColumn )
                 .setComponent( serverFilter );

        // filter: active
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> activeFilter = new ComboBox<>();
        activeFilter.setAllowCustomValue( false );
        activeFilter.setClassName( STYLE_GRID_COMBOBOX );
        activeFilter.setItems( filterAll, filterYes, filterNo );
        activeFilter.setValue( filterAll );
        activeFilter.setSizeFull();
        activeFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( activeFilter.getValue(), filterYes ) )
            {
                filterMap.put( UserSessionColumns.ACTIVE, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( activeFilter.getValue(), filterNo ) )
            {
                filterMap.put( UserSessionColumns.ACTIVE, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( UserSessionColumns.ACTIVE );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( activeColumn )
                 .setComponent( activeFilter );
    }

    private void stopServer( String canvasId )
    {
        DialogInfoUtil dialogInfoUtil = new DialogInfoUtil( virtualUtil.getTranslation( Translation.USERS_SESSION_WAIT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        dialogInfoUtil.setButtonsVisible( false );
        dialogInfoUtil.open();
        UI.getCurrent()
          .push();

        userSessionDatas.stream()
                        .filter( Objects::nonNull )
                        .filter( userSessionData -> StringUtils.equals( userSessionData.getCanvasId(), canvasId ) )
                        .forEach( userSessionData -> {
                            Channel channel = null;
                            Session session = null;
                            ServerData serverData = userSessionData.getServerData();
                            String status = null;
                            int length = NO_STREAM_VALUE;
                            boolean xpraServerRunning = false;
                            byte[] temporaryValue = new byte[BYTES_LENGTH];

                            try
                            {
                                // prepare SSH session
                                session = SshUtil.createSession( virtualUtil.getUserName(),
                                                                 EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                                 serverData.getAddress(), serverData.getSshPort() );
                                session.connect();

                                // get status of XPRA server
                                channel = session.openChannel( Command.SSH.EXECUTE );
                                ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                                InputStream inputStream = channel.getInputStream();
                                channel.connect();
                                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                                {
                                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                    if ( channel.isClosed() )
                                    {
                                        break;
                                    }
                                }
                                while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                                {
                                    if ( length > 0 )
                                    {
                                        status = new String( temporaryValue, 0, length );
                                        if ( StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, userSessionData.getPort() ) ) )
                                        {
                                            xpraServerRunning = true;
                                            break;
                                        }
                                    }
                                }
                                inputStream.close();
                                channel.disconnect();

                                if ( xpraServerRunning )
                                {
                                    // stop XPRA server
                                    LOGGER.info( SERVER_STOPPING, serverData.getAddress(), userSessionData.getPort() );
                                    channel = session.openChannel( Command.SSH.EXECUTE );
                                    ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.STOP, userSessionData.getPort() ) );
                                    channel.connect();
                                    for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                                    {
                                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                        if ( channel.isClosed() )
                                        {
                                            break;
                                        }
                                    }
                                    channel.disconnect();

                                    // get status of stopped XPRA server
                                    boolean serverStopped = false;
                                    temporaryValue = new byte[BYTES_LENGTH];
                                    for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                                    {
                                        channel = session.openChannel( Command.SSH.EXECUTE );
                                        ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                                        inputStream = channel.getInputStream();
                                        channel.connect();
                                        for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                                        {
                                            TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                            if ( channel.isClosed() )
                                            {
                                                break;
                                            }
                                        }

                                        length = NO_STREAM_VALUE;
                                        while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                                        {
                                            if ( length > 0 )
                                            {
                                                status = new String( temporaryValue, 0, length );
                                                if ( !StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, userSessionData.getPort() ) ) )
                                                {
                                                    serverStopped = true;
                                                    break;
                                                }
                                            }
                                        }
                                        inputStream.close();
                                        channel.disconnect();

                                        if ( serverStopped )
                                        {
                                            break;
                                        }
                                    }

                                    // stop Xpra pulseaudio
                                    channel = session.openChannel( Command.SSH.EXECUTE );
                                    ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_PULSEAUDIO, virtualUtil.getUserName() ) );
                                    channel.connect();
                                    for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                                    {
                                        TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                        if ( channel.isClosed() )
                                        {
                                            break;
                                        }
                                    }
                                    channel.disconnect();
                                }
                                session.disconnect();
                            }
                            catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
                            {
                                Thread.currentThread()
                                      .interrupt();
                            }
                            finally
                            {
                                if ( Objects.nonNull( channel ) && channel.isConnected() )
                                {
                                    channel.disconnect();
                                }

                                if ( Objects.nonNull( session ) && session.isConnected() )
                                {
                                    session.disconnect();
                                }
                            }

                            // remove user session data from database
                            LOGGER.info( SERVER_STOPPED );
                            if ( Objects.nonNull( database ) )
                            {
                                database.removeUserSession( userSessionData.getId() );
                                database.removeUserSessionApplication( userSessionData );
                            }
                        } );

        dialogInfoUtil.close();
    }

}
