/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.utils;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.crypto.spec.SecretKeySpec;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.i18n.TranslationProvider;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.info.KeyboardLayout;
import org.dussan.vaadin.dvirtual.info.TranslationLocale;
import org.dussan.vaadin.dvirtual.runners.CheckBrowserHealthRunner;
import org.dussan.vaadin.dvirtual.runners.FileDownloadRunner;
import org.dussan.vaadin.dvirtual.runners.FileUploadRunner;
import org.dussan.vaadin.dvirtual.runners.ImportDataRunner;
import org.dussan.vaadin.dvirtual.runners.factory.RunnerFactory;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.html.Label;

public abstract class VirtualUtil
    implements Serializable
{

    private static final long serialVersionUID = 5317653720063093339L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int CHECK_BROWSER_DELAY = 5; // seconds

    private static final int CHECK_BROWSER_INITIAL_DELAY = 0; // seconds

    private static final int FILE_DOWNLOAD_DELAY = 250; // milliseconds

    private static final int FILE_UPLOAD_DELAY = 250; // milliseconds

    private static final int IMPORT_DATA_DELAY = 250; // milliseconds

    private static final int MAX_WAIT_INTERVAL = 30000; // milliseconds

    private static final int NUMBER_OF_RUNNERS = 10;

    private static final int SLEEP_INTERVAL = 50; // milliseconds

    private static final int UNDEFINED_CHILD_APPLICATION = 0;

    private static final String ERROR_LOGIN = "Wrong username/password or server not running on address %s!";

    private static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

    private static final String RUNNING_APPLICATIONS = "<b>%s:</b>";

    private static final String STRING_EQUAL = "=";

    private static final String STRING_NEW_LINE = "<br/>";

    private static final String WINDOW_TITLE = "windows.%s.title=";

    private boolean allRunningInSameInstance = false;

    private boolean applicationsBarVisible = true;

    private boolean applicationTesting = false;

    private boolean applicationTestingEditDialogOpened = false;

    private boolean applicationTestingError = false;

    private boolean browserOpened = true;

    private boolean childApplicationCanStarted = false;

    private boolean connectionToServerActive = false;

    private boolean drawingStarted = false;

    private boolean logOutInProcess = false;

    private boolean logOutUserAfterLogOutProcessEnded = false;

    private boolean recreateNewWindowEnabled = false;

    private boolean recreateNewWindowAgainEnabled = false;

    private boolean runningApplicationClosed = true;

    private boolean runningApplicationInClosingProcess = false;

    private boolean serverConnected = false;

    private boolean setupFileDeleted = true;

    private boolean setupUserActive = false;

    private boolean startupNotificationsVisible = true;

    private boolean tooltipsVisible = true;

    private boolean userLogOut = false;

    private byte[] secureIV = null;

    private int desktopWidth = 0;

    private int desktopHeight = 0;

    private int mouseMoveThrottlePeriod = 0;

    private long childApplicationId = UNDEFINED_CHILD_APPLICATION;

    private String canvasId = null;

    private String userName = null;

    private String userPassword = null;

    private String windowsFocus = null;

    private String windowsStatus = null;

    private ApplicationLocale applicationLocale = null;

    private transient ExecutorService connectorService = null;

    private transient ExecutorService runnerService = null;

    private KeyboardLayout keyboardLayout = null;

    private transient ScheduledExecutorService checkBrowserHealthService = null;

    private transient ScheduledExecutorService fileDownloadFileService = null;

    private transient ScheduledExecutorService fileUploadFileService = null;

    private transient ScheduledExecutorService importDataService = null;

    private SecretKeySpec secretKeySpec = null;

    private TranslationProvider translationProvider = null;

    protected VirtualUtil()
    {
        translationProvider = new TranslationProvider();
        setApplicationLocale( ApplicationLocale.DEFAULT_LOCALE );
        setKeyboardLayout( KeyboardLayout.DEFAULT_LAYOUT );
    }

    public abstract void logOutUser();

    public byte[] getSecureIV()
    {
        return secureIV;
    }

    public void setSecureIV( String saltIV )
    {
        this.secureIV = EncryptionUtil.getSecureIV( saltIV );
    }

    public SecretKeySpec getSecretKeySpec()
    {
        return secretKeySpec;
    }

    public void setSecretKeySpec( String salt, String saltKey )
    {
        this.secretKeySpec = EncryptionUtil.getSecretKeySpecSpec( salt, saltKey );
    }

    public String getCanvasId()
    {
        return canvasId;
    }

    public void setCanvasId( String canvasId )
    {
        this.canvasId = canvasId;
    }

    public ApplicationLocale getApplicationLocale()
    {
        return applicationLocale;
    }

    public void setApplicationLocale( ApplicationLocale applicationLocale )
    {
        if ( Objects.nonNull( applicationLocale ) )
        {
            this.applicationLocale = applicationLocale;
        }
    }

    public KeyboardLayout getKeyboardLayout()
    {
        return keyboardLayout;
    }

    public void setKeyboardLayout( KeyboardLayout keyboardLayout )
    {
        if ( Objects.nonNull( keyboardLayout ) )
        {
            this.keyboardLayout = keyboardLayout;
        }
    }

    public void stopAllStartedServices()
    {
        stopCheckBrowserHealthService();
        stopConnectorService();
        stopRunnerService();
        stopFileDownloadService();
        stopFileUploadService();
        stopImportDataService();
    }

    public boolean isCheckBrowserHealthServiceRunning()
    {
        return ( Objects.nonNull( checkBrowserHealthService ) );
    }

    public void startCheckBrowserHealthService( CheckBrowserHealthRunner command )
    {
        try
        {
            if ( Objects.isNull( checkBrowserHealthService ) )
            {
                checkBrowserHealthService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.BROWSER, getCanvasId() ) );
            }
            checkBrowserHealthService.scheduleWithFixedDelay( command, CHECK_BROWSER_INITIAL_DELAY, CHECK_BROWSER_DELAY, TimeUnit.SECONDS );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetCheckBrowserHealthService()
    {
        stopCheckBrowserHealthService();
        checkBrowserHealthService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.BROWSER, getCanvasId() ) );
    }

    public void stopCheckBrowserHealthService()
    {
        if ( Objects.nonNull( checkBrowserHealthService ) )
        {
            checkBrowserHealthService.shutdown();
            checkBrowserHealthService = null;
        }
    }

    public void waitToConnectorServiceIsDown()
    {
        while ( isConnectorServiceUp() )
        {
            try
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
            }
            catch ( InterruptedException e )
            {
                Thread.currentThread()
                      .interrupt();
            }
        }
    }

    public boolean isConnectorServiceUp()
    {
        return ( Objects.nonNull( connectorService ) && !connectorService.isShutdown() && !connectorService.isTerminated() );
    }

    public void startConnectorService( Runnable command )
    {
        try
        {
            if ( Objects.isNull( connectorService ) )
            {
                connectorService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.CONNECTOR, getCanvasId() ) );
            }
            connectorService.execute( command );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetConnectorService()
    {
        stopConnectorService();
        connectorService = Executors.newSingleThreadExecutor( new RunnerFactory( RunnerFactory.CONNECTOR, getCanvasId() ) );
    }

    public void stopConnectorService()
    {
        if ( Objects.nonNull( connectorService ) )
        {
            connectorService.shutdown();
            connectorService = null;
        }
    }

    public void startFileDownloadService( FileDownloadRunner command )
    {
        try
        {
            if ( Objects.isNull( fileDownloadFileService ) )
            {
                fileDownloadFileService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.DOWNLOAD, getCanvasId() ) );
            }
            ScheduledFuture<?> scheduledFuture = fileDownloadFileService.schedule( command, FILE_DOWNLOAD_DELAY, TimeUnit.MILLISECONDS );
            command.setScheduledFuture( scheduledFuture );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetFileDownloadService()
    {
        stopFileDownloadService();
        fileDownloadFileService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.DOWNLOAD, getCanvasId() ) );
    }

    public void stopFileDownloadService()
    {
        if ( Objects.nonNull( fileDownloadFileService ) )
        {
            fileDownloadFileService.shutdown();
            fileDownloadFileService = null;
        }
    }

    public void startFileUploadService( FileUploadRunner command )
    {
        try
        {
            if ( Objects.isNull( fileUploadFileService ) )
            {
                fileUploadFileService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.UPLOAD, getCanvasId() ) );
            }
            ScheduledFuture<?> scheduledFuture = fileUploadFileService.schedule( command, FILE_UPLOAD_DELAY, TimeUnit.MILLISECONDS );
            command.setScheduledFuture( scheduledFuture );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetFileUploadService()
    {
        stopFileUploadService();
        fileUploadFileService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.UPLOAD, getCanvasId() ) );
    }

    public void stopFileUploadService()
    {
        if ( Objects.nonNull( fileUploadFileService ) )
        {
            fileUploadFileService.shutdown();
            fileUploadFileService = null;
        }
    }

    public void startImportDataService( ImportDataRunner command )
    {
        try
        {
            if ( Objects.isNull( importDataService ) )
            {
                importDataService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.IMPORT, getCanvasId() ) );
            }
            ScheduledFuture<?> scheduledFuture = importDataService.schedule( command, IMPORT_DATA_DELAY, TimeUnit.MILLISECONDS );
            command.setScheduledFuture( scheduledFuture );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetImportDataService()
    {
        stopImportDataService();
        importDataService = Executors.newSingleThreadScheduledExecutor( new RunnerFactory( RunnerFactory.IMPORT, getCanvasId() ) );
    }

    public void stopImportDataService()
    {
        if ( Objects.nonNull( importDataService ) )
        {
            importDataService.shutdown();
        }
    }

    public void startRunnerService( Runnable command )
    {
        try
        {
            if ( Objects.isNull( runnerService ) )
            {
                runnerService = Executors.newFixedThreadPool( NUMBER_OF_RUNNERS, new RunnerFactory( getCanvasId(), true ) );
            }
            runnerService.execute( command );
        }
        catch ( RejectedExecutionException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
    }

    public void resetRunnerService()
    {
        stopRunnerService();
        runnerService = Executors.newFixedThreadPool( NUMBER_OF_RUNNERS, new RunnerFactory( getCanvasId(), true ) );
    }

    public void stopRunnerService()
    {
        if ( Objects.nonNull( runnerService ) )
        {
            runnerService.shutdown();
            runnerService = null;
        }
    }

    public String getUserName()
    {
        return userName;
    }

    public void setUserName( String userName )
    {
        this.userName = userName;
    }

    public String getUserPassword()
    {
        return userPassword;
    }

    public void setUserPassword( String userPassword )
    {
        this.userPassword = userPassword;
    }

    public String getWindowsFocus()
    {
        return StringUtils.defaultIfBlank( windowsFocus, StringUtils.EMPTY );
    }

    public void setWindowsFocus( String windowsFocus )
    {
        this.windowsFocus = windowsFocus;
    }

    public String getWindowsStatus()
    {
        return windowsStatus;
    }

    public void setWindowsStatus( String windowsStatus )
    {
        this.windowsStatus = windowsStatus;
    }

    public int getDesktopWidth()
    {
        return desktopWidth;
    }

    public void setDesktopWidth( int desktopWidth )
    {
        this.desktopWidth = desktopWidth;
    }

    public int getDesktopHeight()
    {
        return desktopHeight;
    }

    public void setDesktopHeight( int desktopHeight )
    {
        this.desktopHeight = desktopHeight;
    }

    public int getMouseMoveThrottlePeriod()
    {
        return mouseMoveThrottlePeriod;
    }

    public void setMouseMoveThrottlePeriod( int mouseMoveThrottlePeriod )
    {
        this.mouseMoveThrottlePeriod = mouseMoveThrottlePeriod;
    }

    public boolean isSetupFileDeleted()
    {
        return setupFileDeleted;
    }

    public void setSetupFileDeleted( boolean setupFileDeleted )
    {
        this.setupFileDeleted = setupFileDeleted;
    }

    public boolean isSetupUserActive()
    {
        return setupUserActive;
    }

    public void setSetupUserActive( boolean setupUserActive )
    {
        this.setupUserActive = setupUserActive;
    }

    public boolean canUserLogin( String userName, String userPassword, ServerData serverData )
    {
        boolean canLogin = false;
        if ( StringUtils.isNotBlank( userName ) && StringUtils.isNotBlank( userPassword ) && Objects.nonNull( serverData ) )
        {
            Session session = null;
            try
            {
                session = SshUtil.createSession( userName, userPassword, serverData.getAddress(), serverData.getSshPort() );
                session.connect();
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                session.disconnect();
                canLogin = true;
            }
            catch ( Exception exception )
            {
                LOGGER.error( String.format( ERROR_LOGIN, serverData.getAddress() ) );
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }
        return canLogin;
    }

    public String getWindowsTilte( String userName, ServerData serverData, int xpraPort, int[] windowIds )
    {
        Channel channel = null;
        Session session = null;
        String inputStreamValue = StringUtils.EMPTY;
        try
        {
            session = SshUtil.createSession( userName, EncryptionUtil.decrypt( userPassword, getSecretKeySpec(), getSecureIV() ), serverData.getAddress(), serverData.getSshPort() );
            session.connect();
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.WINDOWS_TITLE, xpraPort ) );
            InputStream inputStream = channel.getInputStream();
            channel.connect();

            byte[] tempInputStreamValue = new byte[BYTES_LENGTH];
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            while ( inputStream.available() > 0 )
            {
                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                if ( length > 0 )
                {
                    inputStreamValue = inputStreamValue.concat( new String( tempInputStreamValue, 0, length ) );
                }
            }
            inputStream.close();
            channel.disconnect();
            session.disconnect();
        }
        catch ( InterruptedException | IllegalThreadStateException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        StringBuilder windowsTitleBuilder = new StringBuilder();
        List<String> windowsTitle = Arrays.asList( StringUtils.split( inputStreamValue, LINE_SEPARATOR ) );
        windowsTitle.removeIf( StringUtils::isBlank );
        List<String> titles = new ArrayList<>();
        for ( String windowTitle : windowsTitle )
        {
            for ( int windowId : windowIds )
            {
                if ( windowTitle.startsWith( String.format( WINDOW_TITLE, windowId ) ) )
                {
                    String[] title = StringUtils.split( windowTitle, STRING_EQUAL );
                    if ( title.length > 1 && StringUtils.isNotBlank( title[1] ) )
                    {
                        titles.add( title[1] );
                    }
                }
            }
        }
        Collections.sort( titles );
        if ( CollectionUtils.isNotEmpty( titles ) )
        {
            windowsTitleBuilder.append( String.format( RUNNING_APPLICATIONS, getTranslation( Translation.APPLICATIONS_RUNNING ) ) );
        }
        for ( String title : titles )
        {
            windowsTitleBuilder.append( STRING_NEW_LINE );
            Label label = new Label( title );
            windowsTitleBuilder.append( label.getElement()
                                             .getOuterHTML() );
        }
        return windowsTitleBuilder.toString();
    }

    public void setDefaultValues()
    {
        drawingStarted = false;
        logOutInProcess = false;
        logOutUserAfterLogOutProcessEnded = false;
        recreateNewWindowEnabled = false;
        recreateNewWindowAgainEnabled = false;
        runningApplicationClosed = true;
        runningApplicationInClosingProcess = false;
        userLogOut = false;
    }

    public String getLineSeparator()
    {
        return LINE_SEPARATOR;
    }

    public void detachLockedServer( String userName, ServerData serverData, Integer port )
    {
        if ( StringUtils.isNotBlank( userName ) && StringUtils.isNotBlank( userPassword ) && Objects.nonNull( serverData ) && Objects.nonNull( port ) )
        {
            detachLockedServer( userName, serverData.getAddress(), serverData.getSshPort(), port );
        }
    }

    public void detachLockedServer( String userName, String host, Integer sshPort, Integer port )
    {
        if ( StringUtils.isNotBlank( userName ) && StringUtils.isNotBlank( userPassword ) && StringUtils.isNotBlank( host ) && Objects.nonNull( sshPort ) && Objects.nonNull( port ) )
        {
            Channel channel = null;
            Session session = null;
            try
            {
                session = SshUtil.createSession( userName, EncryptionUtil.decrypt( userPassword, getSecretKeySpec(), getSecureIV() ), host, sshPort );
                session.connect();
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.DETACH, port ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | InterruptedException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }
    }

    public boolean isBrowserOpened()
    {
        return browserOpened;
    }

    public boolean isNotBrowserOpened()
    {
        return !browserOpened;
    }

    public void setBrowserOpened( boolean browserOpened )
    {
        this.browserOpened = browserOpened;
    }

    public boolean isConnectionToServerActive()
    {
        return connectionToServerActive;
    }

    public boolean isNotConnectionToServerActive()
    {
        return !connectionToServerActive;
    }

    public void setConnectionToServerActive( boolean connectionToServerActive )
    {
        this.connectionToServerActive = connectionToServerActive;
    }

    public boolean isServerConnected()
    {
        return serverConnected;
    }

    public boolean isNotServerConnected()
    {
        return !serverConnected;
    }

    public void setServerConnected( boolean serverConnected )
    {
        this.serverConnected = serverConnected;
    }

    public boolean canChildApplicationStarted()
    {
        return childApplicationCanStarted;
    }

    public boolean canNotChildApplicationStarted()
    {
        return !childApplicationCanStarted;
    }

    public void setChildApplicationCanStarted( boolean childApplicationCanStarted )
    {
        this.childApplicationCanStarted = childApplicationCanStarted;
    }

    public boolean haveToStartChildApplication()
    {
        return ( childApplicationId != UNDEFINED_CHILD_APPLICATION );
    }

    public boolean haveNotToStartChildApplication()
    {
        return ( childApplicationId == UNDEFINED_CHILD_APPLICATION );
    }

    public long getChildApplicationId()
    {
        return childApplicationId;
    }

    public void resetChildApplicationId()
    {
        setChildApplicationCanStarted( false );
        setChildApplicationId( UNDEFINED_CHILD_APPLICATION );
    }

    public void setChildApplicationId( long childApplicationId )
    {
        this.childApplicationId = childApplicationId;
    }

    public boolean isDrawingStarted()
    {
        return drawingStarted;
    }

    public boolean isNotDrawingStarted()
    {
        return !drawingStarted;
    }

    public void setDrawingStarted( boolean drawingStarted )
    {
        this.drawingStarted = drawingStarted;
    }

    public boolean isUserLogOut()
    {
        return userLogOut;
    }

    public void setUserLogOut( boolean userLogOut )
    {
        this.userLogOut = userLogOut;
    }

    public boolean isLogOutInProcess()
    {
        return logOutInProcess;
    }

    public boolean isNotLogOutInProcess()
    {
        return !logOutInProcess;
    }

    public void setLogOutInProcess( boolean logOutInProcess )
    {
        this.logOutInProcess = logOutInProcess;
        if ( logOutInProcess )
        {
            logOutUserAfterLogOutProcessEnded = true;
        }
    }

    public boolean isRunningApplicationClosed()
    {
        return runningApplicationClosed;
    }

    public void setRunningApplicationClosed( boolean runningApplicationClosed )
    {
        this.runningApplicationClosed = runningApplicationClosed;
    }

    public boolean isRunningApplicationInClosingProcess()
    {
        return runningApplicationInClosingProcess;
    }

    public boolean isNotRunningApplicationInClosingProcess()
    {
        return !runningApplicationInClosingProcess;
    }

    public void setRunningApplicationInClosingProcess( boolean runningApplicationInClosingProcess )
    {
        this.runningApplicationInClosingProcess = runningApplicationInClosingProcess;

        // sign out user
        if ( logOutUserAfterLogOutProcessEnded && isNotRunningApplicationInClosingProcess() && isRunningApplicationClosed() && isNotLogOutInProcess() )
        {
            logOutUserAfterLogOutProcessEnded = false;
            logOutUser();
        }
    }

    public boolean isRecreateNewWindowEnabled()
    {
        return recreateNewWindowEnabled;
    }

    public boolean isNotRecreateNewWindowEnabled()
    {
        return !recreateNewWindowEnabled;
    }

    public void setRecreateNewWindowEnabled( boolean recreateNewWindowEnabled )
    {
        this.recreateNewWindowEnabled = recreateNewWindowEnabled;
    }

    public boolean isRecreateNewWindowAgainEnabled()
    {
        return recreateNewWindowAgainEnabled;
    }

    public boolean isNotRecreateNewWindowAgainEnabled()
    {
        return !recreateNewWindowAgainEnabled;
    }

    public void setRecreateNewWindowAgainEnabled( boolean recreateNewWindowAgainEnabled )
    {
        this.recreateNewWindowAgainEnabled = recreateNewWindowAgainEnabled;
    }

    public TranslationLocale getTranslationLocale()
    {
        return translationProvider.getTranslationLocale();
    }

    public void setTranslationLocale( TranslationLocale translationLocale )
    {
        translationProvider.setTranslationLocale( translationLocale );
    }

    public String getTranslation( String key, Object... params )
    {
        return translationProvider.getTranslation( key, params );
    }

    public boolean isAllRunningInSameInstance()
    {
        return allRunningInSameInstance;
    }

    public void setAllRunningInSameInstance( boolean allRunningInSameInstance )
    {
        this.allRunningInSameInstance = allRunningInSameInstance;
    }

    public boolean isApplicationsBarVisible()
    {
        return applicationsBarVisible;
    }

    public boolean isNotApplicationsBarVisible()
    {
        return !applicationsBarVisible;
    }

    public void setApplicationsBarVisible( boolean applicationsBarVisible )
    {
        this.applicationsBarVisible = applicationsBarVisible;
    }

    public boolean isStartupNotificationsVisible()
    {
        return startupNotificationsVisible;
    }

    public void setStartupNotificationsVisible( boolean startupNotificationsVisible )
    {
        this.startupNotificationsVisible = startupNotificationsVisible;
    }

    public boolean isApplicationTesting()
    {
        return applicationTesting;
    }

    public boolean isNotApplicationTesting()
    {
        return !applicationTesting;
    }

    public void setApplicationTesting( boolean applicationTesting )
    {
        this.applicationTesting = applicationTesting;
    }

    public boolean isApplicationTestingEditDialogOpened()
    {
        return applicationTestingEditDialogOpened;
    }

    public void setApplicationTestingEditDialogOpened( boolean applicationTestingEditDialogOpened )
    {
        this.applicationTestingEditDialogOpened = applicationTestingEditDialogOpened;
    }

    public boolean isApplicationTestingError()
    {
        return applicationTestingError;
    }

    public boolean isNotApplicationTestingError()
    {
        return !applicationTestingError;
    }

    public void setApplicationTestingError( boolean applicationTestingError )
    {
        this.applicationTestingError = applicationTestingError;
    }

    public boolean isTooltipsVisible()
    {
        return tooltipsVisible;
    }

    public void setTooltipsVisible( boolean tooltipsVisible )
    {
        this.tooltipsVisible = tooltipsVisible;
    }

}
