/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.connector;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogConnectingUtil;
import org.dussan.vaadin.dvirtual.xpra.client.Client;
import org.dussan.vaadin.dvirtual.xpra.io.PacketReader;
import org.dussan.vaadin.dvirtual.xpra.io.Sender;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientCloseWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientDisconnectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.client.ClientSuspendPacket;
import org.dussan.vaadin.dvirtual.xpra.window.Window;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;

public class Connector
    implements Runnable, Serializable
{

    private static final long serialVersionUID = 2161543333379208686L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int MAX_CLOSE_WAIT_INTERVAL = 10000; // milliseconds

    private static final int MAX_RECONNECTIONS_ALLOWED = 20;

    private static final int MAX_WAIT_INTERVAL = 30000; // milliseconds

    private static final int NO_STREAM_VALUE = -1;

    private static final int SLEEP_ERROR_INTERVAL = 1000; // milliseconds

    private static final int SLEEP_INTERVAL = 50; // milliseconds

    private static final int SLEEP_MAX_INTERVAL = 5000; // milliseconds

    private static final int SLEEP_STEP_INTERVAL = 1000; // milliseconds

    private static final String APPLICATION_STOPPING = "Stoping application: {} on server {}:{}";

    private static final String CONNECTION_CLOSED = "Connection to the server is closed.";

    private static final String CONNECTION_CLOSING = "Closing a connection: {}:{}";

    private static final String CONNECTION_ERROR = "Connection error, trying to recconect!";

    private static final String CONNECTION_OPENING = "Opening a connection: {}:{}";

    private static final String SECURITY_ERROR = "Security error!";

    private static final String SERVER_STARTING = "Starting the server...";

    private static final String SERVER_STOPPED = "The server is stopped.";

    private static final String SERVER_STOPPING = "Stopping the server: {}:{}";

    private static final String THREAD_ERROR = "Thread error!";

    public static final int SSH_DEFAULT_PORT = 22;

    private boolean stopServer = false;

    private int reconnectCounter = 0;

    private int sshPort = ServerData.DEFAULT_SSH_PORT;

    private int xpraPort = UserSessionData.NEW_PORT;

    private long userSessionId = UserSessionData.NEW_USER_SESSION;

    private String command = null;

    private String errorMessage = null;

    private String host = null;

    private Client client = null;

    private Database database = null;

    private DialogConnectingUtil dialogConnectingUtil = null;

    private VirtualUtil virtualUtil = null;

    public Connector( Client client, String host, int xpraPort, VirtualUtil virtualUtil )
    {
        this.client = client;
        this.host = host;
        this.xpraPort = xpraPort;
        this.virtualUtil = virtualUtil;

        // window message
        dialogConnectingUtil = new DialogConnectingUtil( virtualUtil.getTranslation( Translation.APPLICATION_CONNECT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        dialogConnectingUtil.setButtonsVisible( false );
    }

    public Connector( Client client, String host, int xpraPort, DialogConnectingUtil dialogConnectingUtil, VirtualUtil virtualUtil )
    {
        if ( Objects.isNull( dialogConnectingUtil ) || !StringUtils.equals( dialogConnectingUtil.getMessage(), virtualUtil.getTranslation( Translation.APPLICATION_CONNECT ) ) )
        {
            LOGGER.info( SERVER_STARTING );
        }

        this.client = client;
        this.host = host;
        this.xpraPort = xpraPort;
        this.dialogConnectingUtil = dialogConnectingUtil;
        this.virtualUtil = virtualUtil;

        // window message
        if ( Objects.isNull( this.dialogConnectingUtil ) )
        {
            this.dialogConnectingUtil = new DialogConnectingUtil( virtualUtil.getTranslation( Translation.APPLICATION_CONNECT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
            this.dialogConnectingUtil.setButtonsVisible( false );
            this.dialogConnectingUtil.setProgressBarVisible( false );
        }
    }

    private boolean applicationRunning()
    {
        boolean applicationRunning = false;

        Channel channel = null;
        Session session = null;
        try
        {
            session =
                SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ), host, sshPort );
            session.connect();

            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
            InputStream inputStream = channel.getInputStream();
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }

            int length = NO_STREAM_VALUE;
            byte[] temporaryValue = new byte[BYTES_LENGTH];
            while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
            {
                if ( length > 0 )
                {
                    String status = new String( temporaryValue, 0, length );
                    if ( StringUtils.isNotBlank( status ) && status.contains( String.format( Command.XPRA.LIST_STATUS, xpraPort ) ) )
                    {
                        applicationRunning = true;
                        break;
                    }
                }
            }
            inputStream.close();
            channel.disconnect();
            session.disconnect();
        }
        catch ( InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        return applicationRunning;
    }

    protected void onPacketReceived( List<Object> packetList )
    {
        getClient().onPacketReceived( packetList );
    }

    public void setSessionCommandData( String command, int sshPort )
    {
        this.command = command;
        this.sshPort = sshPort;
    }

    public void setUserSessionData( Database database, UserSessionData userSessionData )
    {
        this.database = database;
        this.userSessionId = userSessionData.getId();
        setSessionCommandData( userSessionData.getApplicationData()
                                              .getCommand(),
                               userSessionData.getServerData()
                                              .getSshPort() );
    }

    public DialogConnectingUtil getDialogConnectingUtil()
    {
        return dialogConnectingUtil;
    }

    public String getErrorMessage()
    {
        return errorMessage;
    }

    public void setErrorMessage( String errorMessage )
    {
        this.errorMessage = errorMessage;
    }

    public void resetErrorMessage()
    {
        errorMessage = null;
    }

    public Client getClient()
    {
        return client;
    }

    public boolean isRunning()
    {
        return virtualUtil.isConnectorServiceUp();
    }

    public synchronized boolean connect()
    {
        try
        {
            virtualUtil.resetConnectorService();
            virtualUtil.startConnectorService( this );
            return true;
        }
        catch ( RejectedExecutionException exception )
        {
            setErrorMessage( exception.getMessage() );
            LOGGER.error( THREAD_ERROR, exception );
        }
        return false;
    }

    public synchronized void disconnect()
    {
        if ( ( virtualUtil.isConnectorServiceUp() || virtualUtil.isNotBrowserOpened() ) && Objects.nonNull( getClient() ) && Objects.nonNull( getClient().getSender() ) )
        {
            getClient().getSender()
                       .onSend( new ClientDisconnectPacket() );
        }

        if ( virtualUtil.isBrowserOpened() )
        {
            virtualUtil.stopConnectorService();
            virtualUtil.stopFileDownloadService();
            virtualUtil.stopFileUploadService();
            virtualUtil.stopImportDataService();
        }
    }

    public void stopApplication()
    {
        LOGGER.info( APPLICATION_STOPPING, command, host, xpraPort );

        Channel channel = null;
        Session session = null;
        try
        {
            // close all windows/applications
            List<Window> windows = new ArrayList<>( getClient().getWindows() );
            for ( Window window : windows )
            {
                getClient().getSender()
                           .onSend( new ClientCloseWindowPacket( window.getId() ) );
            }
            TimeUnit.MILLISECONDS.sleep( SLEEP_MAX_INTERVAL );
            LOGGER.info( SERVER_STOPPING, host, xpraPort );

            // disconnect client from server
            getClient().onDisconnect();
            TimeUnit.MILLISECONDS.sleep( SLEEP_STEP_INTERVAL );

            // prepare SSH session
            session =
                SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ), host, sshPort );
            session.connect();

            // get status of running XPRA server
            boolean serverRunning = true;
            byte[] temporaryValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
            InputStream inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            int length = NO_STREAM_VALUE;
            while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
            {
                if ( length > 0 )
                {
                    String status = new String( temporaryValue, 0, length );
                    serverRunning = StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, xpraPort ) );
                    break;
                }
            }
            inputStream.close();
            channel.disconnect();

            // stop XPRA server
            if ( serverRunning )
            {
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.STOP, xpraPort ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_CLOSE_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();
            }
            session.disconnect();

            LOGGER.info( SERVER_STOPPED );
        }
        catch ( IllegalThreadStateException | InterruptedException | IOException | JSchException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        // remove user session data from database
        if ( Objects.nonNull( database ) )
        {
            database.removeUserSession( userSessionId );
            database.removeUserSessionApplication( database.getUserSession( userSessionId ) );
        }
    }

    public void stopServer()
    {
        stopServer = true;
    }

    public boolean haveToStopServer()
    {
        return stopServer;
    }

    public void resetReconnectCounter()
    {
        reconnectCounter = 0;
    }

    @Override
    public void run()
    {
        LOGGER.info( CONNECTION_OPENING, host, xpraPort );
        boolean detachedServerLocked = false;

        if ( !applicationRunning() )
        {
            getClient().setDisconnectedByServerError( true );
        }
        else
        {
            for ( ;; )
            {
                if ( detachedServerLocked )
                {
                    virtualUtil.detachLockedServer( virtualUtil.getUserName(), host, sshPort, xpraPort );
                    detachedServerLocked = false;
                }

                boolean reconnect = false;
                try (Socket socket = new Socket( host, xpraPort ))
                {
                    socket.setKeepAlive( true );
                    InputStream inputStream = socket.getInputStream();
                    OutputStream outputStream = socket.getOutputStream();
                    getClient().onConnect( new Sender( outputStream ) );

                    int socketPort = socket.getPort();
                    String hostName = socket.getInetAddress()
                                            .getHostName();
                    String hostAddress = socket.getInetAddress()
                                               .getHostAddress();

                    PacketReader reader = new PacketReader( inputStream );
                    while ( !Thread.interrupted() && !getClient().isDisconnectedByServer() )
                    {
                        List<Object> packetData = reader.readList();
                        if ( CollectionUtils.isNotEmpty( packetData ) && ( StringUtils.equalsIgnoreCase( hostName, host ) || StringUtils.equalsIgnoreCase( hostAddress, host ) )
                            && socketPort == xpraPort )
                        {
                            onPacketReceived( packetData );
                        }
                        else if ( CollectionUtils.isEmpty( packetData ) )
                        {
                            // DON'T CHANGE
                            TimeUnit.MILLISECONDS.sleep( SLEEP_ERROR_INTERVAL );
                            getClient().getSender()
                                       .onSend( new ClientSuspendPacket() );
                            getClient().getSender()
                                       .send( new ClientDisconnectPacket() );
                            getClient().setForceUpdateEnabled( true );
                            reconnect = true;
                            break;
                        }
                    }

                    // close streams
                    socket.shutdownInput();
                    socket.shutdownOutput();
                }
                catch ( SecurityException exception )
                {
                    LOGGER.warn( SECURITY_ERROR, exception );
                    setErrorMessage( exception.getMessage() );
                    getClient().setDisconnectedBySecurityError( true );
                    getDialogConnectingUtil().setShowDialog( true );
                    stopServer();
                }
                catch ( NullPointerException exception )
                {
                    LOGGER.warn( CONNECTION_ERROR );
                    setErrorMessage( exception.getMessage() );
                    getClient().setDisconnectedByServerError( true );
                    getDialogConnectingUtil().setShowDialog( false );
                    reconnect = false;
                }
                catch ( Exception exception )
                {
                    // catch all other exceptions and try to reconnect to the socket again
                    LOGGER.warn( CONNECTION_ERROR );
                    reconnectCounter += 1;
                    reconnect = true;
                }
                finally
                {
                    detachedServerLocked = true;
                    if ( Objects.nonNull( getClient() ) && Objects.nonNull( getClient().getSender() ) )
                    {
                        if ( reconnect )
                        {
                            getClient().setForceUpdateEnabled( true );
                            detachedServerLocked = false;
                        }
                        else
                        {
                            getClient().getSender()
                                       .close();
                            reconnect = false;
                        }
                    }
                }

                // exit from loop because reconnect is not required or too many reconnection attempts
                if ( !reconnect || reconnectCounter > MAX_RECONNECTIONS_ALLOWED )
                {
                    break;
                }
            }
        }

        if ( getClient().isDisconnectedByServerError() )
        {
            stopServer();
        }
        LOGGER.info( CONNECTION_CLOSING, host, xpraPort );
        if ( getClient().getVirtualUtil()
                        .isNotBrowserOpened() )
        {
            LOGGER.info( CONNECTION_CLOSED );
        }
        virtualUtil.setServerConnected( false );

        // save error message due to too many reconnection attempt
        if ( reconnectCounter > MAX_RECONNECTIONS_ALLOWED )
        {
            LOGGER.info( CONNECTION_CLOSED );
            setErrorMessage( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECTION ) );
            virtualUtil.stopConnectorService();
        }
    }

}
