/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.io;

import java.io.Serializable;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;

import org.apache.commons.collections4.CollectionUtils;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.packets.ServerPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerClipboardRequestPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerClipboardTokenPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerConfigureOverrideRedirectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerCursorPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDesktopSizePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDisconnectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerDrawPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerEncodingsPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerEosPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerHelloPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerLostWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerNewOverrideRedirectPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerNewWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerPingEchoPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerPingPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerRaiseWindowPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerSetClipboardEnabledPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerSetDeflatePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerStartupCompletePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowIconPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowMetadataPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowMoveResizePacket;
import org.dussan.vaadin.dvirtual.xpra.packets.server.ServerWindowResizedPacket;
import org.dussan.vaadin.dvirtual.xpra.processors.ServerPacketProcessors;
import org.dussan.vaadin.dvirtual.xpra.processors.base.ServerPacketProcessor;
import org.slf4j.Logger;

public class Receiver
    implements Serializable
{

    private static final long serialVersionUID = -4059378862142610037L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String PACKET_NOT_SUPPORTED = "Server packet is not supported, packet type: {}.";

    private static final String PACKET_TO_SMALL = "Server packet is too small.";

    private ServerPacketProcessors serverPacketProcessors = null;

    public Receiver()
    {
        serverPacketProcessors = new ServerPacketProcessors();
    }

    public void addProcessor( PacketType.Server type, ServerPacketProcessor<?> processor )
    {
        serverPacketProcessors.addProcessor( type, processor );
    }

    @SuppressWarnings( "unchecked" )
    public synchronized void onReceive( List<Object> packetDatas )
    {
        if ( CollectionUtils.isEmpty( packetDatas ) )
        {
            LOGGER.info( PACKET_TO_SMALL );
        }
        else
        {
            ServerPacket serverPacket = null;
            Iterator<Object> packetData = packetDatas.iterator();
            String serverPacketType = ServerPacket.getString( packetData.next() );
            PacketType.Server packetType = PacketType.Server.getType( serverPacketType );
            if ( Objects.nonNull( packetType ) )
            {
                switch ( packetType )
                {
                    case CLIPBOARD_REQUEST:
                        serverPacket = new ServerClipboardRequestPacket();
                        break;
                    case CLIPBOARD_TOKEN:
                        serverPacket = new ServerClipboardTokenPacket();
                        break;
                    case CONFIGURE_OVERRIDE_REDIRECT:
                        serverPacket = new ServerConfigureOverrideRedirectPacket();
                        break;
                    case CURSOR:
                        serverPacket = new ServerCursorPacket();
                        break;
                    case DESKTOP_SIZE:
                        serverPacket = new ServerDesktopSizePacket();
                        break;
                    case DISCONNECT:
                        serverPacket = new ServerDisconnectPacket();
                        break;
                    case DRAW:
                        serverPacket = new ServerDrawPacket();
                        break;
                    case ENCODINGS:
                        serverPacket = new ServerEncodingsPacket();
                        break;
                    case EOS:
                        serverPacket = new ServerEosPacket();
                        break;
                    case HELLO:
                        serverPacket = new ServerHelloPacket();
                        break;
                    case LOST_WINDOW:
                        serverPacket = new ServerLostWindowPacket();
                        break;
                    case NEW_OVERRIDE_REDIRECT:
                        serverPacket = new ServerNewOverrideRedirectPacket();
                        break;
                    case NEW_WINDOW:
                        serverPacket = new ServerNewWindowPacket();
                        break;
                    case PING:
                        serverPacket = new ServerPingPacket();
                        break;
                    case PING_ECHO:
                        serverPacket = new ServerPingEchoPacket();
                        break;
                    case RAISE_WINDOW:
                        serverPacket = new ServerRaiseWindowPacket();
                        break;
                    case SET_CLIPBOARD_ENABLED:
                        serverPacket = new ServerSetClipboardEnabledPacket();
                        break;
                    case SET_DEFLATE:
                        serverPacket = new ServerSetDeflatePacket();
                        break;
                    case STARTUP_COMPLETE:
                        serverPacket = new ServerStartupCompletePacket();
                        break;
                    case WINDOW_ICON:
                        serverPacket = new ServerWindowIconPacket();
                        break;
                    case WINDOW_METADATA:
                        serverPacket = new ServerWindowMetadataPacket();
                        break;
                    case WINDOW_MOVE_RESIZE:
                        serverPacket = new ServerWindowMoveResizePacket();
                        break;
                    case WINDOW_RESIZED:
                        serverPacket = new ServerWindowResizedPacket();
                        break;
                    default:
                        // nothing to do
                        break;
                }
            }

            if ( Objects.nonNull( serverPacket ) )
            {
                // LOGGER.info( "==========================RECEIVE: {}", serverPacketType );
                ServerPacketProcessor<ServerPacket> serverPacketProcessor = (ServerPacketProcessor<ServerPacket>) serverPacketProcessors.getProcessor( packetType );
                serverPacket.unpackPacket( packetData );
                serverPacketProcessor.process( serverPacket );
            }
            else
            {
                LOGGER.warn( PACKET_NOT_SUPPORTED, serverPacketType );
            }
        }
    }

}
