/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.xpra.packets.server;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.packets.ServerPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClientCapabilities;
import org.dussan.vaadin.dvirtual.xpra.packets.base.Packet;

public class ServerHelloPacket
    extends ServerPacket
{

    private static final long serialVersionUID = -5078950481269523034L;

    private static final String ALT_MODIFIER = "Alt";

    private static final String ALTGR_MODIFIER = "AltGr";

    private static final String CONTROL_MODIFIER = "Control";

    private static final String META_MODIFIER = "Meta";

    private static final String MODIFIER_KEYCODES = "modifier_keycodes";

    private static final String MODIFIER_ALT_L = "Alt_L";

    private static final String MODIFIER_CONTROL_L = "Control_L";

    private static final String MODIFIER_ISO_LEVEL3_SHIFT = "ISO_Level3_Shift";

    private static final String MODIFIER_META_L = "Meta_L";

    private static final String MODIFIER_MODE_SWITCH = "Mode_switch";

    private static final String MODIFIER_NUM_LOCK = "Num_Lock";

    private static final String MOUSE_PRECISE_WHEEL = "wheel.precise";

    private static final String NUM_LOCK_MODIFIER = "NumLock";

    private transient Map<String, Object> capabilities = null;

    private transient Map<String, String> modifiers = null;

    public ServerHelloPacket()
    {
        super( PacketType.Server.HELLO.toString() );
    }

    public Map<String, Object> getCapabilities()
    {
        return capabilities;
    }

    public boolean isPrintingEnabled()
    {
        return getBoolean( capabilities.get( ClientCapabilities.PRINTING ) );
    }

    public boolean isRencodeEnabled()
    {
        return getBoolean( capabilities.get( ClientCapabilities.RENCODE ) );
    }

    public List<String> getCapability( String key )
    {
        List<String> list = new ArrayList<>();
        Object capability = capabilities.get( key );
        if ( capability instanceof Collection )
        {
            Collection<?> collection = (Collection<?>) capability;
            for ( Object value : collection )
            {
                list.add( getString( value ) );
            }
        }
        return list;
    }

    public List<String> getModifiers()
    {
        List<String> mods = new ArrayList<>();
        modifiers.forEach( ( key, value ) -> {
            mods.add( key );
            mods.add( value );
        } );
        return mods;
    }

    public boolean getMousePreciseWheel()
    {
        return (boolean) getCapabilities().get( MOUSE_PRECISE_WHEEL );
    }

    @SuppressWarnings( "unchecked" )
    @Override
    public void unpackPacket( Iterator<Object> packetData )
    {
        super.unpackPacket( packetData );
        modifiers = new HashMap<>();
        capabilities = (Map<String, Object>) packetData.next();
        modifiers.put( CONTROL_MODIFIER, CONTROL_MODIFIER.toLowerCase() );

        if ( capabilities.containsKey( MODIFIER_KEYCODES ) )
        {
            Map<String, Object> modifierKeycodes = (Map<String, Object>) capabilities.get( MODIFIER_KEYCODES );
            List<String> keys = new ArrayList<>( modifierKeycodes.keySet() );
            for ( String key : keys )
            {
                Object modifierKeycode = modifierKeycodes.get( key );
                if ( modifierKeycode instanceof List )
                {
                    for ( Object modifier : (ArrayList<Object>) modifierKeycode )
                    {
                        if ( modifier instanceof List )
                        {
                            List<Object> mods = (ArrayList<Object>) modifier;
                            for ( Object mod : mods )
                            {
                                String modValue = StringUtils.stripToEmpty( Packet.getString( mod ) );
                                switch ( modValue )
                                {
                                    case MODIFIER_ALT_L:
                                        modifiers.put( ALT_MODIFIER, key );
                                        break;
                                    case MODIFIER_ISO_LEVEL3_SHIFT:
                                    case MODIFIER_MODE_SWITCH:
                                        modifiers.put( ALTGR_MODIFIER, key );
                                        break;
                                    case MODIFIER_CONTROL_L:
                                        modifiers.put( CONTROL_MODIFIER, key );
                                        break;
                                    case MODIFIER_META_L:
                                        modifiers.put( META_MODIFIER, key );
                                        break;
                                    case MODIFIER_NUM_LOCK:
                                        modifiers.put( NUM_LOCK_MODIFIER, key );
                                        break;
                                    default:
                                        // nothing to do
                                        break;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

}
